"""
fbxExportTool.py

UI tool for fbx export

"""

import os.path
import maya.cmds as mc
import maya.mel as mm

from ..utils import name
from ..utils import transform
from ..utils import filereference
from ..utils import skinCluster
from ..utils import visibility
from ..utils import shape
from ..utils import connect
from . import presets


class fbxExport_window:
    
    def __init__(self):
        
        self.window = 'fbxExport_window'
        self.title = 'fbxExport'
        self.size = (300,530)

        # check if window is already opened
        if mc.window( self.window, exists = True ):
            mc.deleteUI( self.window, window = True )
        
        # window creation
        self.window = mc.window( self.window, t = self.title, widthHeight = self.size )
        
        mc.columnLayout( adjustableColumn=True )

        # text hint for user
        mc.text( 'Select something from the rig and hit Export FBX', h = 20)
        
        mc.text( 'Define export file path', h = 20)
        self.FilePath = mc.textField()
        
        mc.button( label = 'Open file browser ...', command = self.openSaveWindow )
        

        # save fbx
        mc.separator( h = 10 )
        mc.text( 'Export FBX of selected rig', h = 20 )
        mc.button( label = 'Export FBX', width = 50, command = self.exportFbx, backgroundColor= [0.3,0.5,0.8] )
        
        
        # selection of the rig
        mc.separator( h = 10 )
        self.selectedRigTxf = mc.textField( en = 0 )
        mc.button( label = 'Optionaly: Save rig object selection', command = self.saveSelect )
        mc.separator( h = 5, st = 'none' ) 
        mc.button( label = 'Clear saved selection', command = self.clearSelection )

        # options 1
        mc.separator( h = 10 )
        mc.rowLayout()
        self.exportOptions1Cbg = mc.checkBoxGrp(columnWidth3=[70,60,60], numberOfCheckBoxes=2, label='Options 1: ', label1='model', label2='export timeline only', v1=True, v2=True)
        mc.setParent('..')

        # options 2
        mc.separator( h = 10, style= 'none' )
        mc.rowLayout()
        self.exportOptions2Cbg = mc.checkBoxGrp(columnWidth3=[70,120,60], numberOfCheckBoxes=2, label='Options 2: ', label1='remove namespace', label2='subdivide', v1=True, v2= False)
        mc.setParent('..')
        mc.columnLayout()
        self.exportOptions2Rbg = mc.radioButtonGrp( columnWidth4= [70,50,50,50], label='subdiv level: ', labelArray3=['1', '2', '3'], numberOfRadioButtons=3, select=1 )
        mc.text( '  subdiv UV boundary smoothing:', h = 20 )
        self.exportOptions3Rbg = mc.radioButtonGrp( columnWidth5= [0,50,100,50,50], label='', labelArray4=['none', 'edges+corners', 'edges', 'Maya Cm-Cl'], numberOfRadioButtons=4, select=2,
                                                   an1= 'None', an2= 'Preserve Edges and Corners', an3= 'Preserve Edges', an4= 'Maya Catmull-Clark' )
        mc.text( '  UV map to copy subdiv weights:', h = 20 )
        self.exportOptions4Rbg = mc.radioButtonGrp( columnWidth3= [50,100,100], label='', labelArray2=['use model', 'make generic'], numberOfRadioButtons=2, select=1,
                                                   an1= 'use model UV map to copy skin weights to subdiv model - make sure all faces have UVs and non-overlapping', an2= 'make generic UV map to copy skin weights to subdiv model' )
        mc.setParent('..')

        # selection of non-subdivide geos
        self.selectNonSubdivTxf = mc.textField( en = 0 )
        mc.button( label = 'Optionaly: Select non-subdivide mesh objects', command = self.selectNonSubdiv )
        mc.separator( h = 5, st = 'none' ) 
        mc.button( label = 'Clear saved selection', command = self.clearNonSubdiv )

        # options 3
        mc.separator( h = 10, style= 'none' )
        mc.rowLayout()
        self.exportOptions3Cbg = mc.checkBoxGrp(columnWidth2=[70,160], numberOfCheckBoxes=1, label='Options 3: ', label1='export only visible geometry', v1=True)
        mc.setParent('..')

        # options 4
        mc.separator( h = 10, style= 'none' )
        mc.rowLayout()
        self.exportOptions4Cbg = mc.checkBoxGrp(columnWidth2=[70,160], numberOfCheckBoxes=1, label='Options 4: ', label1='joints and model under one group', v1=True)
        mc.setParent('..')

    
        # display window
        mc.showWindow(self.window)
        mc.window( self.window, e=True, widthHeight = self.size )

    def openSaveWindow(self, *args):
        singleFilter = 'FBX File (*.fbx)'
        SelectedPath = mc.fileDialog2( fileFilter = singleFilter, ds = 2, fileMode = 0 )[0]
        mc.textField(self.FilePath, e = True, tx = SelectedPath)

    def saveSelect(self, *args):
        selection = mc.ls(sl=1)[0]
        mc.textField(self.selectedRigTxf, e = True, tx = selection)

    def clearSelection(self, *args):
        mc.textField(self.selectedRigTxf, e = True, tx = '')
    
    def selectNonSubdiv(self, *args):
        sel = mc.ls(sl=1)
        selTx = ' '.join(sel)
        mc.textField(self.selectNonSubdivTxf, e = True, tx = selTx)
    
    def clearNonSubdiv(self, *args):
        mc.textField(self.selectNonSubdivTxf, e = True, tx = '')

    def exportFbx(self, *args):

        # check scene selection
        sel = mc.ls(sl=1)
        txtFldSel = mc.textField(self.selectedRigTxf, q= 1, tx= 1)
        if not sel and not txtFldSel:
            mc.confirmDialog(t= 'rigitAnimTools fbxExport', m='please make a selection of anything from the rig')
            return

        yesStr = 'yes'
        sceneSavedRes = mc.confirmDialog(title= '          fbxExport', message= 'Is your scene saved?', button=['yes','no'], 
                                         defaultButton=yesStr, cancelButton='No', dismissString='No')
        
        if sceneSavedRes == yesStr:
            self.saveFBX()
        else:
            return

    def saveFBX(self, *args):
                      
            turnOffUndo = True

            # check UI export options
            exportModelCheck = mc.checkBoxGrp( self.exportOptions1Cbg, q=True, v1=True )
            exportTimelineOnly = mc.checkBoxGrp( self.exportOptions1Cbg, q=True, v2=True )
            removeNamespace = mc.checkBoxGrp( self.exportOptions2Cbg, q=True, v1=True )
            subdivideModel = mc.checkBoxGrp( self.exportOptions2Cbg, q=True, v2=True )
            subdivLevelUi = mc.radioButtonGrp( self.exportOptions2Rbg, q=True, select=True )
            subdUvBoundarySmooth = mc.radioButtonGrp( self.exportOptions3Rbg, q=True, select=True )
            subdivUvMapMode = mc.radioButtonGrp( self.exportOptions4Rbg, q=True, select=True )
            modelExpVisOnly = mc.checkBoxGrp( self.exportOptions3Cbg, q=True, v1=True )
            oneExportGrp = mc.checkBoxGrp( self.exportOptions4Cbg, q=True, v1=True )
            fileLocation = mc.textField(self.FilePath, q= 1, tx= 1)
            selRigObj = mc.textField(self.selectedRigTxf, q= 1, tx= 1)
            selNonSubdivStr = mc.textField(self.selectNonSubdivTxf, q=1, tx= 1)
            subdivBakeFrame = mc.currentTime( q=1 )

            if not selRigObj:
                selRigObj = mc.ls(sl=1)[0]
            
            # get non-subdivide objects
            nonSubdivObjs = []
            if selNonSubdivStr:
                objs = selNonSubdivStr.split()
                nonSubdivObjs = objs[:]

                if removeNamespace:
                    nonSubdivObjs = [name.removeNamespace(o) for o in objs]
            
            # get rig info
            rigTopGrp = transform.getTopNode( selRigObj, returnShortNames = True )

            origRigTopGrp = rigTopGrp
            nameSpace = name.getNamespace(origRigTopGrp)
            isReference = filereference.isReference(origRigTopGrp)
            
            # turn off Undo and delete Undo queue
            if turnOffUndo:

                mc.flushUndo()
                mc.undoInfo( state=False )

            # import reference and optionally remove namespace
            if nameSpace or isReference:

                if nameSpace == '' and isReference:
                    nameSpace = ':'

                try:
                    filereference.importReferenceToScene( namespaceName = nameSpace, removeNamespace = removeNamespace, verbose = True )
                except:
                    pass
                origRigTopGrp = name.removeNamespace(origRigTopGrp)
                
                if removeNamespace:
                    nameSpace = ''
                    rigTopGrp = name.removeNamespace( rigTopGrp )
                else:
                    nameSpace = nameSpace + ':'
                
            # sort export group names
            assetName = mc.getAttr('{}.rigitName'.format(rigTopGrp))
            exportPrefix = 'export_'
            jointsGrp = nameSpace + exportPrefix + 'joints_grp'
            self.sceneCheck_exportJntGrp(jointsGrp)
            jointsGrp = mc.rename(jointsGrp, assetName + '_' + jointsGrp)
            assetModelGrp = nameSpace + 'assetModel_grp'
            assetModelGrp = mc.rename(assetModelGrp, assetName + '_' + nameSpace + exportPrefix + 'assetModel_grp')
            jointsGrp = transform.getLongNameInHierarchy( jointsGrp, rigTopGrp )[0]
            assetModelGrp = transform.getLongNameInHierarchy( assetModelGrp, rigTopGrp )[0]

            exportTopJoints = mc.listRelatives(jointsGrp, c=True, type= 'joint')
            # commented for now - renaming Top Joint with asset name prefix
            # for topj in exportTopJoints:
            #     mc.rename( topj,  assetName + '_' + topj)

            # make FBX export groups and delete rig groups
            fbxExportGrp = mc.group( n = nameSpace + assetName + '_rig_export_grp', em = True )
            visCtrl = nameSpace + 'vis_ctl'
            visCtrl = transform.getLongNameInHierarchy( visCtrl, rigTopGrp )[0]
            mc.setAttr( visCtrl + '.geometryDispType', 0)

            mc.parent( jointsGrp, fbxExportGrp )
            jointsGrp = transform.getLongNameInHierarchy( jointsGrp, fbxExportGrp )[0]

            if exportModelCheck:
                mc.parent( assetModelGrp, fbxExportGrp )
                assetModelGrp = transform.getLongNameInHierarchy( assetModelGrp, fbxExportGrp )[0]
            
            all = mc.listRelatives( origRigTopGrp, ad = True, fullPath= True )
            for item in all:
                mc.lockNode( item, l= False )
            
            # bake animation
            # joints + blendshapes (morphs)
            exportJoints = mc.listRelatives(jointsGrp, ad = True, type= 'joint', fullPath= True)
            bslist = mc.ls(type= 'blendShape')
            exportJoints.extend(bslist)
            
            minFrame = mc.findKeyframe('*', w = 'first')
            maxFrame = mc.findKeyframe('*', w = 'last')

            if exportTimelineOnly:
                minFrame = mc.playbackOptions(q=True, min=True)
                maxFrame = mc.playbackOptions(q=True, max=True)
            
            # self._disconnectVis(assetModelGrp, exportModelCheck) # make sure visibility will not be changed after bake process
            # - NOTE: not used, disconnect visibility would delete potential animations on this channel
            mc.bakeResults(exportJoints, t = (minFrame, maxFrame), shape=True, disableImplicitControl= True, simulation = True)
            mc.delete( origRigTopGrp )
            mc.showHidden( jointsGrp )

            # normalize skinCluster in case some joints were not in Export Joints
            forcedGenericUvObjs = []
            if exportModelCheck:

                mc.currentTime(subdivBakeFrame) # go back to original frame
                
                # get all geos under asset model group and filter them
                allGeos = self.getAllGeos(assetModelGrp, modelExpVisOnly)
                mc.sets(allGeos)
                
                # prepare non-subdiv object names
                nonSubdivObjsLong = []
                if nonSubdivObjs:
                    for o in nonSubdivObjs:
                        
                        res = transform.getLongNameInHierarchy(o, assetModelGrp)
                        if res:
                            nonSubdivObjsLong.append(res[0])

                for g in allGeos:

                    # normalize skin weights
                    gLong = transform.getLongNameInHierarchy( g, assetModelGrp )[0]
                    sc = skinCluster.getRelated(gLong)
                    if sc:
                        mc.skinPercent(sc, gLong, normalize=True)
                    
                    # subdivide mesh
                    if subdivideModel:
                        
                        if gLong in nonSubdivObjsLong:
                            continue
                        
                        # hiresMesh, forcedGenericUVs = skinCluster.copySubdiv_delnondf( gLong, subdivLevel= subdivLevelUi, subdUvBoundary= subdUvBoundarySmooth)
                        hiresMesh, forcedGenericUVs = skinCluster.copySubdiv_uvCopy( gLong, subdivLevel= subdivLevelUi, subdUvBoundary= subdUvBoundarySmooth, subdivUvMapMode= subdivUvMapMode )
                        
                        if forcedGenericUVs:
                            forcedGenericUvObjs.append(g)

            # remove top export group
            fbxExportSelection = [fbxExportGrp]
            if not oneExportGrp:
                jointsGrp = mc.parent(jointsGrp, w=True)[0]
                exportGrpObjs = mc.listRelatives(fbxExportGrp, c=True, type= 'transform')
                exportTopJoints = mc.listRelatives(jointsGrp, c=True, type= 'joint')
                exportGrpObjs = mc.parent(exportGrpObjs, w=True)
                exportTopJoints = mc.parent(exportTopJoints, w=True)
                exportGrpObjs.extend(exportTopJoints)
                mc.delete(fbxExportGrp,jointsGrp)
                fbxExportSelection = exportGrpObjs

            # FBX export
            fbxPlugin = 'fbxmaya'
            if not mc.pluginInfo( fbxPlugin, q= True, loaded= True ):
                mc.loadPlugin( fbxPlugin )
            
            presetsPath = os.path.split(presets.__file__)[0]
            fbxPresetFile = 'rigit_fbx_export_001.fbxexportpreset'
            presetsPath = os.path.join(presetsPath, fbxPresetFile)
            presetsPathMel = presetsPath.replace('\\','/')
            mm.eval('FBXLoadExportPresetFile -f "%s"' % presetsPathMel)
            mc.select( fbxExportSelection )
            mm.eval('FBXExport -f "%s" "-s"' % fileLocation)

            # report
            self.exportReportDialog(assetName, forcedGenericUvObjs)

            # turn Undo back on again
            mc.undoInfo( state=True )
    
    def getAllGeos(self, assetModelGrp, modelExpVisOnly):

        """
        get all geos under model group
        """

        geosUnderModelGrp = shape.getAllGeosUnderObject(assetModelGrp, longNames= True)
        allGeos = []
        for geo in geosUnderModelGrp:

            if not mc.objExists(geo):
                continue

            gAddInList = True
            if modelExpVisOnly:
                
                gVisible = visibility.test(geo)
                if not gVisible:
                    mc.delete(geo)
                    gAddInList = False
            
            if gAddInList:
                allGeos.append( geo )
        
        allGeos = list( set( allGeos ) ) # cleanup duplicates

        # delete all empty hidden model groups
        if modelExpVisOnly:
            
            allObjs = mc.listRelatives( assetModelGrp, ad=True, type= 'transform', fullPath= True )
            for o in allObjs:

                if not mc.objExists(o):
                    continue

                meshUnder = mc.listRelatives( o, ad=True, type= 'mesh' )
                if not meshUnder:
                    visVal = mc.getAttr( o + '.v' )
                    if visVal == 0:
                        mc.delete(o)

        
        return allGeos
    
    def _disconnectVis(self, assetModelGrp, exportModelCheck):

        """
        disconnect visibility connections from model objects - make sure visibility will not be changed after bake process
        """

        if not exportModelCheck:
            return
        
        modelObjs = mc.listRelatives(assetModelGrp, ad=True, f=True)
        for o in modelObjs:

            connect.disconnect(o + '.v')

    def exportReportDialog(self, assetName, forcedGenericUvObjs):

        """
        export report dialog
        """

        if not forcedGenericUvObjs:
            return
        
        genUvsObjsShort = [ name.short(o) for o in forcedGenericUvObjs ]
        
        objslines = '\n'.join(genUvsObjsShort)
        mc.confirmDialog(t= 'rigitAnimTools fbxExport', 
                         m='asset: %s\noverlapping UVs found on listed objects\ngeneric UVs had to be made:\n\n%s'% (assetName,objslines),
                         button= 'close')
        
    
    def sceneCheck_exportJntGrp(self, jointsGrp):

        """
        check if export joints group exists
        """

        if mc.objExists(jointsGrp):
            return
        
        mc.confirmDialog(t= 'rigitAnimTools fbxExport', m='export joints group "%s" not found in the rig.\nPlease make sure it was processed for FBX export.\n\nFor TDs - use rigit.tools.gameTools.skeletonExportSetup()'% jointsGrp)
        raise Exception('rigitAnimTools fbxExport - %s not found in the rig.\nPlease make sure it was processed for FBX export.\nFor TDs - use rigit.tools.gameTools.skeletonExportSetup()'% jointsGrp)

        
def fbxExportUI():
    fbxUI = fbxExport_window()
