""" 
file: ikFkMatch.py
 
Description:
    This tool switches the IK/FK rigs on the arms and legs of any Rigit rig 
      while matching the locations of the appropriate limb for a seamless rig switch.
    Select any icon on the arms or the lets and click 'IKFK Match', the tool will 
    switch/match the rig of the limb the selected icon is part of.
 
How to Run: 

from rigitAnimTools import ikFkMatch
reload(ikFkMatch)
ikFkMatch.ikFkMatchUI()
"""
from __future__ import division
from __future__ import print_function

import maya.OpenMaya as om
import maya.cmds as mc

from . import definitions as df


def checkDrivers ( ns, objSide ):
        
    hipDriverJnt = df.nameDefDefault.jointDriverNames['hipJnt'] 
    
    if mc.objExists( hipDriverJnt ):
        
        df.nameDefDefault.jointNames['hipJnt'] = df.nameDefDefault.jointDriverNames['hipJnt'] 
        df.nameDefDefault.jointNames['kneeJnt'] = df.nameDefDefault.jointDriverNames['kneeJnt'] 
        df.nameDefDefault.jointNames['shoulderJnt'] = df.nameDefDefault.jointDriverNames['shoulderJnt'] 
        df.nameDefDefault.jointNames['elbowJnt'] = df.nameDefDefault.jointDriverNames['elbowJnt'] 

def getCtlType ( name ):

    
    for val in list(df.nameDefDefault.jointNames.values()):
        if val in name:
            return 3

    for val in list(df.nameDefDefault.legNames.values()):
        if val in name:
            return 2

    for val in list(df.nameDefDefault.armNames.values()):
        if val in name:
            return 1

    return 0

def getNameType ( name ):

    if name in list(df.nameDefDefault.jointNames.keys()):
        return 3

    if name in list(df.nameDefDefault.legNames.keys()):
        return 2

    if name in list(df.nameDefDefault.armNames.keys()):
        return 1

    return 0

def getName ( name, ns, side ):
    
    nameType = getNameType ( name )
    sideStr = df.nameDefDefault.sideCtl[side]

    if nameType == 3:
        sideStr = df.nameDefDefault.sideJnt[side]

    prefix = ns + sideStr

    if nameType == 3:
        return prefix + df.nameDefDefault.jointNames[name]

    if nameType == 2:
        return prefix + df.nameDefDefault.legNames[name]

    if nameType == 1:
        return prefix + df.nameDefDefault.armNames[name]



def ikFkMatchUI( *args ):
    
    global dnWindowText, dnSelObjJob, rangeIkFkRadios, startFrameIntF, endFrameIntF

    if mc.window( 'rigSwitchWindow', exists = True ):
        mc.deleteUI( 'rigSwitchWindow' )

    rigWind = mc.window( 'rigSwitchWindow', w = 270, h = 100, s = False, title = 'RIG-IT IK-FK Match Tool' )
    
    main_column = mc.columnLayout( adjustableColumn=True )
    
    switch_row = mc.rowColumnLayout( nc = 3 )
    mc.separator( st = 'none', hr = False, w = 10, h = 30 )
    dnWindowText = mc.text( l = '[Select an arm or leg control]' )
    mc.separator( st = 'none', hr = False, w = 10 )
    mc.separator( st = 'none', hr = False, w = 10 )
    mc.button( w = 260, h = 25, c = dnSwitchRig, label = 'IKFK Match' )
    mc.separator( st = 'none', hr = False, w = 10, h = 30 )
    
    mc.setParent(main_column)
    mc.separator( st = 'none', h = 20, w = 50 )
    rangeText = mc.text( l = 'Range IK/FK' )
    mc.separator( st = 'none', h = 5, w = 50 )

    mc.setParent(main_column)
    rangeButton_row = mc.rowColumnLayout( nc = 4 )
    mc.separator( st = 'none', hr = False, w = 110 )
    rangeIkFkRadios = mc.radioCollection()
    mc.radioButton( label='ik', docTag= 'rangeIkFkRb_ik', select=True )
    mc.radioButton( label='fk', docTag= 'rangeIkFkRb_fk' )
    mc.separator( st = 'none', h = 20, w = 50 )

    mc.setParent(main_column)
    range_row = mc.rowColumnLayout( nc = 4 )
    mc.separator( st = 'none', hr = False, w = 40 )
    startFrameIntF = mc.intField( value= getCurrentFrameRange()[0] )
    mc.separator( st = 'none', hr = False, w = 10 )
    endFrameIntF = mc.intField( value= getCurrentFrameRange()[1] )

    mc.setParent(main_column)
    rangeButton_column = mc.columnLayout( adjustableColumn=True )
    mc.separator( st = 'none', h = 10, w = 40 )
    rangeButton_row = mc.rowColumnLayout( nc = 3 )
    mc.separator( st = 'none', h = 20, w = 40 )
    mc.button( w = 200, h = 25, c = rangeIkFkSwitch, label = 'Switch in Range' )
    mc.separator( st = 'none', h = 20, w = 50 )
    mc.setParent(rangeButton_column)

    mc.setParent(main_column)    

    mc.showWindow(rigWind)
    mc.window( 'rigSwitchWindow', e = True, w = 285, h = 200 )
    dnChangeText()

    dnSelObjJob = mc.scriptJob( e = ["SelectionChanged", dnChangeText], protected = True )
    #mc.scriptJob( uid = ['rigSwitchWindow', dnKillJob(dnSelObjJob)] )

def getCurrentFrameRange():

    rangeMin = mc.playbackOptions(q=1,min=1)
    rangeMax = mc.playbackOptions(q=1,max=1)

    return [rangeMin, rangeMax]

def rangeIkFkSwitch( *args ):

    # get switch mode
    selectedRb = mc.radioCollection(rangeIkFkRadios, q=1, select=True)
    rbDocTag = mc.radioButton(selectedRb, q=1, docTag=True)
    
    # get range from UI
    startFrame =  mc.intField( startFrameIntF, q=True, v=1 )
    endFrame = mc.intField( endFrameIntF, q=True, v=1 )

    ikfkMode = 'ik'
    ikfkBool = 0
    if rbDocTag.endswith( 'fk' ):

        ikfkMode = 'fk'
        ikfkBool = 1
    
    # get toggle control
    res = dnSwitchRig_getInfo()

    if not res:
        mc.confirmDialog( title='RIG-IT IK-FK', message='Please select Arm or Leg control' )
        return
    
    if res:
        
        ns = res[0]
        objSide = res[1]
        legTypeOn = res[2]
        toggleIconName = res[3]
        isHindleg = res[4]
    
    # get IK and FK controls
    ik_switchControls = dnSwitchRigs( ns, objSide, legTypeOn, toggleIconName, isHindleg, onlyReturnControls = True, overrideMode= True, overrideModeIkFk= 0 )
    fk_switchControls = dnSwitchRigs( ns, objSide, legTypeOn, toggleIconName, isHindleg, onlyReturnControls = True, overrideMode= True, overrideModeIkFk= 1 )
    
    # get all ikfk keys in range
    keyframes = []
    newModeCtrlsList = []
    currentCtrlsList = []
    if ikfkMode == 'ik':

        newModeCtrlsList = ik_switchControls[1:]
        currentCtrlsList = fk_switchControls[1:]
    
    else: # fk

        newModeCtrlsList = fk_switchControls[1:]
        currentCtrlsList = ik_switchControls[1:]

    allKeyedFrames = []
    for c in currentCtrlsList:

        keys = listFramesWithKeys( c, startFrame= startFrame, endFrame= endFrame )
        allKeyedFrames.extend( keys )

    allKeyedFrames = list( set(allKeyedFrames) )

    # set autokeyframe off
    mc.autoKeyframe( state= False )

    # keyframe ikfk toggle to current mode
    for frame in allKeyedFrames:

        mc.currentTime(frame)
        mc.setAttr( toggleIconName + '.fkIk', ikfkBool)
        mc.setKeyframe(toggleIconName, t=frame, at= 'fkIk')

    # keyframe control for new mode
    for i, frame in enumerate( allKeyedFrames ):

        mc.currentTime(frame)
        mc.select(currentCtrlsList[-1])
        dnSwitchRig_params( switchToMode = ikfkMode )
        for c in newModeCtrlsList:
            mc.setKeyframe(c, t=frame)
        
        mc.setAttr( toggleIconName + '.fkIk', 1-ikfkBool)
        mc.setKeyframe(toggleIconName, t=frame, at= 'fkIk')
        
    



def listFramesWithKeys( obj, startFrame= None, endFrame= None ):

    """
    list frames with keys. If startFrame and endFrame is not specified, function will return all keys from full timelime

    @param obj: str, name of object
    @param startFrame: int, optional, startFrame
    @param endFrame: int, optional, endFrame
    """

    keyData = []

    if startFrame and endFrame:
        keyData = mc.keyframe(obj, q=1, time=(startFrame,endFrame),valueChange=False, timeChange=True)
    else:
        keyData = mc.keyframe(obj, q=1, time=(),valueChange=False, timeChange=True)
    
    keyframes = []
    if keyData:
        
        # get keyframe times
        keyframes = keyData[0::2] # remove all even items (keep only odd items which have keyframe values)
        keyframes = list( set(keyframes) ) # remove duplicates
    

    return keyframes


def dnChangeText( *args ):

    # get the selected object, it's TP and the character name
    selObjs = mc.ls( sl = True )

    if selObjs:
        obj = selObjs[0]
        if selObjs:

            topParent = obj
            for a in range( 50 ):

                parent = mc.listRelatives ( topParent, p = True )
                if parent:
                    topParent = parent[0]
                else:
                    break  

        if ( mc.ls ( topParent + '.rigitName' )):
            charName = mc.getAttr ( topParent + '.rigitName' )

            # change the text in the UI
            objStr = 'character: {}            {}'.format(charName, obj)
            mc.text( dnWindowText, e = True, l = objStr )
            return
        
    mc.text( dnWindowText, e = True, l = '[Select an arm or leg control]' )


def dnSwitchRig( *args ):
    
    res = dnSwitchRig_getInfo()
    
    if res:
        
        ns = res[0]
        objSide = res[1]
        legTypeOn = res[2]
        toggleIconName = res[3]
        isHindleg = res[4]

        dnSwitchRigs( ns, objSide, legTypeOn, toggleIconName, isHindleg )

def dnSwitchRig_params( switchToMode = 'fk' ):

    res = dnSwitchRig_getInfo()
    
    if res:
        
        ns = res[0]
        objSide = res[1]
        legTypeOn = res[2]
        toggleIconName = res[3]
        isHindleg = res[4]

        switchModeInt = 0 # set to FK for IK switch
        if switchToMode == 'fk':
            switchModeInt = 1 # set to IK for FK switch

        dnSwitchRigs( ns, objSide, legTypeOn, toggleIconName, isHindleg, overrideMode= True, overrideModeIkFk= switchModeInt )



def dnSwitchRig_getInfo():

    selObj = mc.ls( sl = True )

    if selObj:

        # namespace
        df.nameDefDefault.prefix = ''
        
        nameSplit = selObj[0].split( ':' )
        ns = ''
        if len( nameSplit ) > 1:
            ns = nameSplit[0] + ':'
        objStr = nameSplit[-1]

        # get the selected object, figure out which side and limb it is

        legTypeOn = 0
        objSide = ''

        if df.nameDefDefault.sideCtl['l'] in objStr:
            objSide = 'l'

        elif df.nameDefDefault.sideCtl['r']  in objStr:
            objSide = 'r'

        legTypeOn = getCtlType ( objStr )

        
        # get the name of the icon and attr to make the switch and get current setting
        if legTypeOn and objSide:

            if legTypeOn == 1:
                toggleIconName = getName ( 'armToggleCtrl', ns, objSide )
                
            elif legTypeOn == 2:
                toggleIconName = getName ( 'legToggleCtrl', ns, objSide )

            isHindleg = bool ( mc.ls ( getName ('legAnkleIkCtrl', ns, objSide ) ) )
    
            return [ ns, objSide, legTypeOn-1, toggleIconName, isHindleg ]
    
    return


def dnSwitchRigs( ns, objSide, legTypeOn, toggleIconName, isHindleg, onlyReturnControls = False, overrideMode= False, overrideModeIkFk= 0 ):
       
    ikModeOn = mc.getAttr( toggleIconName + '.fkIk' )
    ikModeChanged = False

    # override
    if overrideMode:

        ikModeOn = overrideModeIkFk

    checkDrivers ( ns, objSide )
    switchControls = []

    # arm IK controls
    armIkBaseCtrl = 'armIkBaseCtrl'
    armIk1Ctrl = 'armIk1Ctrl'
    armIk2Ctrl = 'armIk2Ctrl'
    armIkRotCtrl = 'armIkRotCtrl'
    armIkPvCtrl = 'armIkPvCtrl'

    # arm FK controls
    fkShoulderCtrl = 'fkShoulderCtrl'
    fkElbowCtrl = 'fkElbowCtrl'
    fkHandCtrl = 'fkHandCtrl'
    
    # leg IK controls
    legIkBaseCtrl = 'legIkBaseCtrl'
    legIk1Ctrl = 'legIk1Ctrl'
    legIk2Ctrl = 'legIk2Ctrl'
    legBallIkCtrl = 'legBallIkCtrl'
    legIkPoleVtCtrl = 'legIkPoleVtCtrl'
    legBallIkCtrl = 'legBallIkCtrl'
    
    # leg FK controls
    hipFkCtrl = 'hipFkCtrl'
    kneeFkCtrl = 'kneeFkCtrl'
    fkFootCtrl = 'fkFootCtrl'
    toeFkCtrl = 'toeFkCtrl'


    if ikModeOn:

        # IK mode
        if legTypeOn:

            ikDrivenJnt = 'toesJnt'
            if isHindleg:
                ikDrivenJnt = 'toes2Jnt'
            
            hipFkCtl = getName (hipFkCtrl, ns, objSide )
            kneeFkCtl = getName (kneeFkCtrl, ns, objSide )
            fkFootCtl = getName (fkFootCtrl, ns, objSide )
            toeFkCtl = getName (toeFkCtrl, ns, objSide )

            switchControls = [toggleIconName, hipFkCtl, kneeFkCtl, fkFootCtl, toeFkCtl]

            if onlyReturnControls:

                return switchControls

      
            legWtransVals = mc.xform( getName ('hipJnt', ns, objSide ), q = True, ws = True, t = True )
            mc.xform( getName (hipFkCtrl, ns, objSide ), ws = True, t = legWtransVals )
            hipRotVals = getCtlTransform( getName (hipFkCtrl, ns, objSide ), getName ('hipJnt', ns, objSide ), toggleIconName, ikModeOn )
            mc.setAttr( getName (hipFkCtrl, ns, objSide ) + '.rotate', hipRotVals[0], hipRotVals[1], hipRotVals[2] )
            
            kneeRotVals = getCtlTransform( getName (kneeFkCtrl, ns, objSide ), getName ('kneeJnt', ns, objSide ), toggleIconName, ikModeOn )
            mc.setAttr( getName (kneeFkCtrl, ns, objSide ) + '.rotate', kneeRotVals[0], kneeRotVals[1], kneeRotVals[2] )
            
            footRotVals = getCtlTransform( getName (fkFootCtrl, ns, objSide ), getName ('footJnt', ns, objSide ), toggleIconName, ikModeOn )
            mc.setAttr( getName (fkFootCtrl, ns, objSide ) + '.rotate', footRotVals[0], footRotVals[1], footRotVals[2] )
            
            toeRotVals = getCtlTransform( getName (toeFkCtrl, ns, objSide ), getName (ikDrivenJnt, ns, objSide ), toggleIconName, ikModeOn )
            mc.setAttr( getName (toeFkCtrl, ns, objSide ) + '.rotate', toeRotVals[0], toeRotVals[1], toeRotVals[2] )

        else:

            fkShoulderCtl = getName (fkShoulderCtrl, ns, objSide )
            fkElbowCtl = getName (fkElbowCtrl, ns, objSide )
            fkHandCtl = getName (fkHandCtrl, ns, objSide )

            switchControls = [toggleIconName, fkShoulderCtl, fkElbowCtl, fkHandCtl]

            if onlyReturnControls:

                return switchControls
            
            fkShoulderCtrlRoVal = getCtlTransform( getName (fkShoulderCtrl, ns, objSide ), getName ('shoulderJnt', ns, objSide ), toggleIconName, ikModeOn )
            mc.setAttr( getName (fkShoulderCtrl, ns, objSide ) + '.rotate', fkShoulderCtrlRoVal[0], fkShoulderCtrlRoVal[1], fkShoulderCtrlRoVal[2])
            
            fkElbowCtrlRoVal = getCtlTransform( getName (fkElbowCtrl, ns, objSide ), getName ('elbowJnt', ns, objSide ), toggleIconName, ikModeOn )
            for id, axis in enumerate ( ['x', 'y', 'z'] ):

                attr = getName (fkElbowCtrl, ns, objSide ) + '.r' + axis
                if not mc.getAttr ( attr, l = True ):
                    mc.setAttr( attr, fkElbowCtrlRoVal[id] )
            
            fkHandCtrlRoVal = getCtlTransform( getName (fkHandCtrl, ns, objSide ), getName ('handJnt', ns, objSide ), toggleIconName, ikModeOn )
            mc.setAttr( getName (fkHandCtrl, ns, objSide ) + '.rotate', fkHandCtrlRoVal[0], fkHandCtrlRoVal[1], fkHandCtrlRoVal[2] )

            
    else:
        # FK mode
        if legTypeOn:
            
            ikEndJnt = 'footJnt'
            if isHindleg:
                ikEndJnt = 'toes2Jnt'

            legIk1Ctl = getName (legIk1Ctrl, ns, objSide )
            legIk2Ctl = getName (legIk2Ctrl, ns, objSide )
            legBallIkCtl = getName (legBallIkCtrl, ns, objSide )
            legIkBaseCtl = getName (legIkBaseCtrl, ns, objSide )
            legIkPoleVtCtl = getName (legIkPoleVtCtrl, ns, objSide )

            switchControls = [toggleIconName, legIk1Ctl, legIk2Ctl, legBallIkCtl, legIkBaseCtl, legIkPoleVtCtl]

            if onlyReturnControls:

                return switchControls


            legWtransVal = mc.xform( getName ('hipJnt', ns, objSide ), q = True, ws = True, t = True )
            footWtransVal = mc.xform( getName (ikEndJnt, ns, objSide ), q = True, ws = True, t = True )
            footRots = getCtlTransform( getName (legIk1Ctrl, ns, objSide ), getName (ikEndJnt, ns, objSide ), toggleIconName, ikModeOn )
            pvIconPlace = getPoleVectorPosition( getName ('hipJnt', ns, objSide ), getName ('kneeJnt', ns, objSide ), getName ('footJnt', ns, objSide ) )

            if isHindleg:
                reverseLoc = makeReverseLoc ( getName (ikEndJnt, ns, objSide ), getName ('footJnt', ns, objSide ), getName (legIkPoleVtCtrl, ns, objSide ) )

            
            mc.setAttr( toggleIconName + '.fkIk', 1 - ikModeOn)


            if mc.ls ( getName (legBallIkCtrl, ns, objSide ) ):
                mc.setAttr( getName (legBallIkCtrl, ns, objSide ) + '.r', 0, 0, 0 )
            mc.setAttr( getName (legIk2Ctrl, ns, objSide ) + '.r', 0, 0, 0 )
            mc.xform( getName (legIkBaseCtrl, ns, objSide ), ws = True, t = legWtransVal )
            mc.xform( getName (legIk1Ctrl, ns, objSide ), ws = True, t = footWtransVal )
            mc.setAttr( getName (legIk1Ctrl, ns, objSide ) + '.r', footRots[0], footRots[1], footRots[2] )
            mc.xform( getName (legIkPoleVtCtrl, ns, objSide ), ws = True, t = pvIconPlace )

            if isHindleg:
                
                mc.delete ( mc.orientConstraint ( reverseLoc, getName ('legAnkleIkCtrl', ns, objSide ) ) )
                mc.delete ( reverseLoc )
                
            else:

                toeRotVal = getCtlTransform( getName ('legToeIkCtrl', ns, objSide ), getName ('toesJnt', ns, objSide ), toggleIconName, ikModeOn )
                mc.setAttr( getName ('legToeIkCtrl', ns, objSide ) + '.r', toeRotVal[0], toeRotVal[1], toeRotVal[2] )
            
        else:

            armIkBaseCtl = getName (armIkBaseCtrl, ns, objSide )
            armIk1Ctl = getName (armIk1Ctrl, ns, objSide )
            armIk2Ctl = getName (armIk2Ctrl, ns, objSide )
            armIkRotCtl = getName (armIkRotCtrl, ns, objSide )
            armIkPvCtl = getName (armIkPvCtrl, ns, objSide )
            
            switchControls = [toggleIconName, armIkBaseCtl, armIk1Ctl, armIk2Ctl, armIkRotCtl, armIkPvCtl]

            if onlyReturnControls:

                return switchControls

            # arm limb
            #print 'asdfdafs', getName ('clavicleJnt', ns, objSide ) >> what's this? commented out :)
            armIkBaseCtrlRotVals = getCtlTransform( getName (armIkBaseCtrl, ns, objSide ), getName ('clavicleJnt', ns, objSide ), toggleIconName, ikModeOn )

            armWtransVal = getCtlTransform( getName (armIk1Ctrl, ns, objSide ), getName ('handJnt', ns, objSide ), toggleIconName, ikModeOn, returnPosition = True  )
            armIk1CtrlRotVals = getCtlTransform( getName (armIk1Ctrl, ns, objSide ), getName ('handJnt', ns, objSide ), toggleIconName, ikModeOn )
            armIkRotCtrlRotVals = getCtlTransform( getName (armIkRotCtrl, ns, objSide ), getName ('handJnt', ns, objSide ), toggleIconName, ikModeOn )
  
            pvPos = getPoleVectorPosition ( getName ('shoulderJnt', ns, objSide ), getName ( 'elbowJnt', ns, objSide ), getName ('handJnt', ns, objSide ), offsetMulti = 1.0, convex = True, keepWorldY = True )

            # toggle to IK

            mc.setAttr( toggleIconName + '.fkIk', 1 - ikModeOn )
        
            # apply values to controls 
            mc.setAttr( getName (armIk2Ctrl, ns, objSide ) + '.t', 0, 0, 0 )
            mc.setAttr( getName (armIk2Ctrl, ns, objSide ) + '.r', 0, 0, 0 )
            mc.setAttr( getName (armIk1Ctrl, ns, objSide ) + '.t', armWtransVal[0], armWtransVal[1], armWtransVal[2] )

            for id, axis in enumerate ( ['x', 'y', 'z'] ):

                attr = getName (armIk1Ctrl, ns, objSide ) + '.r' + axis
                if not mc.getAttr ( attr, l = True ):
                    mc.setAttr( attr, armIk1CtrlRotVals[id] )


            
            mc.xform( getName (armIkPvCtrl, ns, objSide ), ws = True, t = pvPos )
            mc.setAttr( getName (armIkBaseCtrl, ns, objSide ) + '.r', armIkBaseCtrlRotVals[0], armIkBaseCtrlRotVals[1], armIkBaseCtrlRotVals[2] )

            if not isHindleg:
                mc.setAttr( getName (armIkRotCtrl, ns, objSide ) + '.r',  armIkRotCtrlRotVals[0], armIkRotCtrlRotVals[1], armIkRotCtrlRotVals[2] )
        

    if not ikModeChanged:
        
        mc.setAttr( toggleIconName + '.fkIk', 1 - ikModeOn )
    

def dnKillJob( jobNum ):

    mc.scriptJob( kill = jobNum, f = True )

def copyTransform( refObj ):

    '''
    copy transform object
    '''
    
    copyObj = mc.group( n = refObj + '_copy_grp', em = True )
    mc.delete( mc.parentConstraint( refObj, copyObj ) )
    
    return copyObj


def makeReverseLoc ( startJnt, endJnt, poleVector ):

    newLoc = mc.group( em = 1, world = True )
    mc.delete ( mc.parentConstraint ( startJnt, newLoc ) )

    mc.delete ( mc.aimConstraint ( endJnt, newLoc, upVector=(0, 0, 1), worldUpObject = poleVector, worldUpType='object', aimVector=(0, 1, 0) ) )

    return newLoc

def getCtlTransform ( ctrl, ctrlJnt, toggleIconName, ikModeOn, returnPosition = False ):

    if not mc.ls (ctrl):
        return False

    ctrlParent = mc.listRelatives( ctrl, p = True )[0]
    rotateOrder = mc.getAttr( ctrl + '.ro' )

    mc.setAttr( toggleIconName + '.fkIk', 1 - ikModeOn )
    
    jntTransf = mc.group( em = 1, world = True )
    mc.setAttr( jntTransf + '.ro', rotateOrder )
    mc.delete ( mc.parentConstraint ( ctrlJnt, jntTransf  ))
    

    
    ctlTransf = mc.group( em = 1, world = True )
    mc.setAttr( ctlTransf + '.ro', rotateOrder )
    mc.delete ( mc.parentConstraint ( ctrl, ctlTransf ))
    mc.parent ( ctlTransf, jntTransf )
    
    mc.setAttr( toggleIconName + '.fkIk', ikModeOn )
    
    mc.parent ( ctlTransf, ctrlJnt, relative = True )
    mc.parent ( ctlTransf, ctrlParent )

    ikLocTranslation = mc.getAttr( ctlTransf + '.t' )[0]
    ikLocRotation = mc.getAttr( ctlTransf + '.r' )[0]
    
    if mc.nodeType (ctrl) == 'joint':

        mc.select ( cl = True )
        ctlJntTransf = mc.joint( )
        mc.select ( cl = True )
        mc.parent ( ctlJntTransf, ctrlParent ) 
        rotateOrder = mc.getAttr( ctrl + '.ro' )
        mc.setAttr( ctlJntTransf + '.ro', rotateOrder )
        
        for axis in ['X', 'Y', 'Z']:
            val = mc.getAttr ( ctrl + '.jointOrient' + axis )
            mc.setAttr ( ctlJntTransf + '.jointOrient' + axis, val )

        mc.delete ( mc.parentConstraint ( ctlTransf, ctlJntTransf ))

        ikLocTranslation = mc.getAttr( ctlJntTransf + '.t' )[0]
        ikLocRotation = mc.getAttr( ctlJntTransf + '.r' )[0]
        
        mc.delete ( ctlJntTransf )
    
    #if ctrl == 'Lt_LegAnkleIk_CON':
    #    asdf # what's this?? :)
    
    mc.delete ( jntTransf, ctlTransf )

    if returnPosition:
        
        return ikLocTranslation
    
    else:
        
        return ikLocRotation
    

def getPoleVectorPosition( startObj, midObj, endObj, offsetMulti = 1.0, convex = True, keepWorldY = True ):
    
    '''
    Get position for placement of pole vector constraint target, given 3 positions of limb joints
    Concave option should be True to find position where the limb joint "aims"
    Also, points have to form a triangle to get correct result
    
    @param startPos: start position of limb
    @type startPos: list(3 floats)
    @param midPos: mid position of limb
    @type midPos: list(3 floats)
    @param endPos: end position of limb
    @type endPos: list(3 floats)
    @param offsetMulti: offsetMulti multiply value for pole vector null
    @type offsetMulti: float
    @param convex: if true, return position where limb`s triangle is "aiming", otherwise go opposite direction
    @type convex: bool
    @param keepWorldY: if true maintain world Y position of middle position (midPos), useful for positioning arm/leg pole vector
    @type keepWorldY: bool
    @return: list(3 floats), position for pole vector null
    @author: jakub
    '''
    
    startPos = mc.xform( startObj, q = True, t = True, ws = True )
    midPos = mc.xform( midObj, q = True, t = True, ws = True )
    endPos = mc.xform( endObj, q = True, t = True, ws = True )
    
    startvector = om.MVector( startPos[0], startPos[1], startPos[2] )
    midvector = om.MVector( midPos[0], midPos[1], midPos[2] )
    endvector = om.MVector( endPos[0], endPos[1], endPos[2] )
    
    # get vectors between start-mid and end-mid points
    startvector = midvector - startvector
    endvector = midvector - endvector
    
    upperArmLength = startvector.length()
    
    # get pole vector from start-mid and end-mid vectors cross product
    polevector = ( startvector + endvector ) / 2.0
    
    if not convex:
        polevector = polevector * ( -1 )
    
    polevector.normalize()
    
    # count offset value
    pvOffset = upperArmLength * offsetMulti
        
    polevectoroffset = midvector + ( polevector * pvOffset )
    polevectorPos = [ polevectoroffset.x, polevectoroffset.y, polevectoroffset.z ]
        
    return polevectorPos

