""" 
file: ikFkMatch.py
 
Description:
    This tool switches the IK/FK rigs on the arms and legs of any Rigit rig 
      while matching the locations of the appropriate limb for a seamless rig switch.
    Select any icon on the arms or the lets and click 'IKFK Match', the tool will 
    switch/match the rig of the limb the selected icon is part of.
 
How to Run: 

from rigitAnimTools import ikFkMatch
reload(ikFkMatch)
ikFkMatch.ikFkMatchUI()
"""
from __future__ import division
from __future__ import print_function

import os
import pkgutil
from builtins import range
from builtins import str

import maya.OpenMaya as om
import maya.cmds as mc
import pymel.core as pm
from past.utils import old_div

from . import definitions

pkgpath = os.path.dirname(definitions.__file__)
defList = [name for _, name, _ in pkgutil.iter_modules([pkgpath])]
for defName in defList:
    if not ( '_' in defName):
        exec ( ('from definitions import %s as nameDef' ) % (defName) )
        print ( ('\n\nDefinition %s loaded succesfully.' ) % (defName) )
        break
    else:
        print('No definition file found.')

def checkDrivers ( ns, objSide ):
        
    hipDriverJnt = nameDef.jointDriverNames['hipJnt'] 
    
    if mc.objExists( hipDriverJnt ):
        
        nameDef.jointNames['hipJnt'] = nameDef.jointDriverNames['hipJnt'] 
        nameDef.jointNames['kneeJnt'] = nameDef.jointDriverNames['kneeJnt'] 
        nameDef.jointNames['shoulderJnt'] = nameDef.jointDriverNames['shoulderJnt'] 
        nameDef.jointNames['elbowJnt'] = nameDef.jointDriverNames['elbowJnt'] 

def getCtlType ( name ):

    
    for val in list(nameDef.jointNames.values()):
        if val in name:
            return 3

    for val in list(nameDef.legNames.values()):
        if val in name:
            return 2

    for val in list(nameDef.armNames.values()):
        if val in name:
            return 1

    return 0

def getNameType ( name ):

    if name in list(nameDef.jointNames.keys()):
        return 3

    if name in list(nameDef.legNames.keys()):
        return 2

    if name in list(nameDef.armNames.keys()):
        return 1

    return 0

def getName ( name, ns, side ):
    
    nameType = getNameType ( name )
    sideStr = nameDef.sideCtl[side]

    if nameType == 3:
        sideStr = nameDef.sideJnt[side]

    prefix = ns + sideStr

    if nameType == 3:
        return prefix + nameDef.jointNames[name]

    if nameType == 2:
        return prefix + nameDef.legNames[name]

    if nameType == 1:
        return prefix + nameDef.armNames[name]



def ikFkMatchUI( *args ):
    
    global dnWindowText, dnSelObjJob

    if pm.window( 'rigSwitchWindow', exists = True ):
        pm.deleteUI( 'rigSwitchWindow' )

    rigWind = pm.window( 'rigSwitchWindow', w = 270, h = 100, s = False, title = 'RIG-IT IK-FK Match Tool' )
    
    main_column = pm.rowColumnLayout( nc = 3 )
    pm.separator( st = 'none', hr = False, w = 10, h = 30 )
    dnWindowText = pm.text( l = '[Select an arm or leg control]' )
    pm.separator( st = 'none', hr = False, w = 10 )
    pm.separator( st = 'none', hr = False, w = 10 )
    pm.button( w = 260, h = 25, c = dnSwitchRig, label = 'IKFK Match' )
    pm.separator( st = 'none', hr = False, w = 10, h = 30 )
    pm.setParent( main_column )
    
    rigWind.show()
    pm.window( 'rigSwitchWindow', e = True, w = 285, h = 80 )
    dnChangeText()

    dnSelObjJob = pm.scriptJob( e = ["SelectionChanged", pm.Callback( dnChangeText )], protected = True )
    pm.scriptJob( uid = ['rigSwitchWindow', pm.Callback( dnKillJob, dnSelObjJob )] )

def dnChangeText( *args ):

    # get the selected object, it's TP and the character name
    selObjs = mc.ls( sl = True )

    if selObjs:
        obj = selObjs[0]
        if selObjs:

            topParent = obj
            for a in range( 50 ):

                parent = mc.listRelatives ( topParent, p = True )
                if parent:
                    topParent = parent[0]
                else:
                    break  

        if ( mc.ls ( topParent + '.rigitName' )):
            charName = mc.getAttr ( topParent + '.rigitName' )

            # change the text in the UI
            objStr = 'character: ' + str( charName ) + '            ' + str( obj )
            pm.text( dnWindowText, e = True, l = objStr )
            return
        
    pm.text( dnWindowText, e = True, l = '[Select an arm or leg icon]' )


def dnSwitchRig( *args ):

    
    selObj = pm.ls( sl = True )

    if selObj:

        # namespace
        nameDef.prefix = ''
        
        nameSplit = str( selObj[0] ).split( ':' )
        ns = ''
        if len( nameSplit ) > 1:
            ns = nameSplit[0] + ':'
        objStr = nameSplit[-1]

        # get the selected object, figure out which side and limb it is

        legTypeOn = 0
        objSide = ''

        if nameDef.sideCtl['l'] in objStr:
            objSide = 'l'

        elif nameDef.sideCtl['r']  in objStr:
            objSide = 'r'

        legTypeOn = getCtlType ( objStr )

        
        # get the name of the icon and attr to make the switch and get current setting
        if legTypeOn and objSide:

            if legTypeOn == 1:
                toggleIconName = getName ( 'armToggleCtrl', ns, objSide )
                
            elif legTypeOn == 2:
                toggleIconName = getName ( 'legToggleCtrl', ns, objSide )

            isHindleg = bool ( mc.ls ( getName ('legAnkleIkCtrl', ns, objSide ) ) )

            dnSwitchRigs( ns, objSide, legTypeOn-1, toggleIconName, isHindleg )


def dnSwitchRigs( ns, objSide, legTypeOn, toggleIconName, isHindleg ):
       
    ikModeOn = mc.getAttr( toggleIconName + '.fkIk' )
    ikModeChanged = False

    checkDrivers ( ns, objSide )
    

    if ikModeOn:

        # IK mode
        if legTypeOn:

            ikDrivenJnt = 'toesJnt'
            if isHindleg:
                ikDrivenJnt = 'toes2Jnt'

      
            legWtransVals = pm.xform( getName ('hipJnt', ns, objSide ), q = True, ws = True, t = True )
            pm.xform( getName ('hipFkCtrl', ns, objSide ), ws = True, t = legWtransVals )
            hipRotVals = getCtlTransform( getName ('hipFkCtrl', ns, objSide ), getName ('hipJnt', ns, objSide ), toggleIconName, ikModeOn )
            pm.setAttr( getName ('hipFkCtrl', ns, objSide ) + '.rotate', hipRotVals )
            
            kneeRotVals = getCtlTransform( getName ('kneeFkCtrl', ns, objSide ), getName ('kneeJnt', ns, objSide ), toggleIconName, ikModeOn )
            pm.setAttr( getName ('kneeFkCtrl', ns, objSide ) + '.rotate', kneeRotVals )
            
            footRotVals = getCtlTransform( getName ('fkFootCtrl', ns, objSide ), getName ('footJnt', ns, objSide ), toggleIconName, ikModeOn )
            pm.setAttr( getName ('fkFootCtrl', ns, objSide ) + '.rotate', footRotVals )
            
            toeRotVals = getCtlTransform( getName ('toeFkCtrl', ns, objSide ), getName (ikDrivenJnt, ns, objSide ), toggleIconName, ikModeOn )
            pm.setAttr( getName ('toeFkCtrl', ns, objSide ) + '.rotate', toeRotVals )

        else:

            fkShoulderCtrlRoVal = getCtlTransform( getName ('fkShoulderCtrl', ns, objSide ), getName ('shoulderJnt', ns, objSide ), toggleIconName, ikModeOn )
            pm.setAttr( getName ('fkShoulderCtrl', ns, objSide ) + '.rotate', fkShoulderCtrlRoVal )
            
            fkElbowCtrlRoVal = getCtlTransform( getName ('fkElbowCtrl', ns, objSide ), getName ('elbowJnt', ns, objSide ), toggleIconName, ikModeOn )
            for id, axis in enumerate ( ['x', 'y', 'z'] ):

                attr = getName ('fkElbowCtrl', ns, objSide ) + '.r' + axis
                if not mc.getAttr ( attr, l = True ):
                    pm.setAttr( attr, fkElbowCtrlRoVal[id] )
            
            fkHandCtrlRoVal = getCtlTransform( getName ('fkHandCtrl', ns, objSide ), getName ('handJnt', ns, objSide ), toggleIconName, ikModeOn )
            pm.setAttr( getName ('fkHandCtrl', ns, objSide ) + '.rotate', fkHandCtrlRoVal )
            
    else:
        # FK mode
        if legTypeOn:
            
            ikEndJnt = 'footJnt'
            if isHindleg:
                ikEndJnt = 'toes2Jnt'

            legWtransVal = pm.xform( getName ('hipJnt', ns, objSide ), q = True, ws = True, t = True )
            footWtransVal = pm.xform( getName (ikEndJnt, ns, objSide ), q = True, ws = True, t = True )
            footRots = getCtlTransform( getName ('legIk1Ctrl', ns, objSide ), getName (ikEndJnt, ns, objSide ), toggleIconName, ikModeOn )
            pvIconPlace = getPoleVectorPosition( getName ('hipJnt', ns, objSide ), getName ('kneeJnt', ns, objSide ), getName ('footJnt', ns, objSide ) )

            if isHindleg:
                reverseLoc = makeReverseLoc ( getName (ikEndJnt, ns, objSide ), getName ('footJnt', ns, objSide ), getName ('legIkPoleVtCtrl', ns, objSide ) )

            
            pm.setAttr( toggleIconName + '.fkIk', 1 - ikModeOn)


            if mc.ls ( getName ('legBallIkCtrl', ns, objSide ) ):
                pm.setAttr( getName ('legBallIkCtrl', ns, objSide ) + '.r', ( 0, 0, 0 ) )
            pm.setAttr( getName ('legIk2Ctrl', ns, objSide ) + '.r', ( 0, 0, 0 ) )
            pm.xform( getName ('legIkBaseCtrl', ns, objSide ), ws = True, t = legWtransVal )
            pm.xform( getName ('legIk1Ctrl', ns, objSide ), ws = True, t = footWtransVal )
            pm.setAttr( getName ('legIk1Ctrl', ns, objSide ) + '.r', footRots )
            pm.xform( getName ('legIkPoleVtCtrl', ns, objSide ), ws = True, t = pvIconPlace )

            if isHindleg:
                
                mc.delete ( mc.orientConstraint ( reverseLoc, getName ('legAnkleIkCtrl', ns, objSide ) ) )
                mc.delete ( reverseLoc )
                
            else:

                toeRotVal = getCtlTransform( getName ('legToeIkCtrl', ns, objSide ), getName ('toesJnt', ns, objSide ), toggleIconName, ikModeOn )
                pm.setAttr( getName ('legToeIkCtrl', ns, objSide ) + '.r', toeRotVal )
            
        else:

            # arm libm
            print('asdfdafs', getName ('clavicleJnt', ns, objSide ))
            armIkBaseCtrlRotVals = getCtlTransform( getName ('armIkBaseCtrl', ns, objSide ), getName ('clavicleJnt', ns, objSide ), toggleIconName, ikModeOn )

            armWtransVal = getCtlTransform( getName ('armIk1Ctrl', ns, objSide ), getName ('handJnt', ns, objSide ), toggleIconName, ikModeOn, returnPosition = True  )
            armIk1CtrlRotVals = getCtlTransform( getName ('armIk1Ctrl', ns, objSide ), getName ('handJnt', ns, objSide ), toggleIconName, ikModeOn )
            armIkRotCtrlRotVals = getCtlTransform( getName ('armIkRotCtrl', ns, objSide ), getName ('handJnt', ns, objSide ), toggleIconName, ikModeOn )
  
            pvPos = getPoleVectorPosition ( getName ('shoulderJnt', ns, objSide ), getName ( 'elbowJnt', ns, objSide ), getName ('handJnt', ns, objSide ), offsetMulti = 1.0, convex = True, keepWorldY = True )

            # toggle to IK

            pm.setAttr( toggleIconName + '.fkIk', 1 - ikModeOn )
        
            # apply values to controls 
            pm.setAttr( getName ('armIk2Ctrl', ns, objSide ) + '.t', ( 0, 0, 0 ) )
            pm.setAttr( getName ('armIk2Ctrl', ns, objSide ) + '.r', ( 0, 0, 0 ) )
            mc.setAttr( getName ('armIk1Ctrl', ns, objSide ) + '.t', armWtransVal[0], armWtransVal[1], armWtransVal[2] )

            for id, axis in enumerate ( ['x', 'y', 'z'] ):

                attr = getName ('armIk1Ctrl', ns, objSide ) + '.r' + axis
                if not mc.getAttr ( attr, l = True ):
                    pm.setAttr( attr, armIk1CtrlRotVals[id] )


            
            pm.xform( getName ('armIkPvCtrl', ns, objSide ), ws = True, t = pvPos )
            mc.setAttr( getName ('armIkBaseCtrl', ns, objSide ) + '.r', armIkBaseCtrlRotVals[0], armIkBaseCtrlRotVals[1], armIkBaseCtrlRotVals[2] )

            if not isHindleg:
                mc.setAttr( getName ('armIkRotCtrl', ns, objSide ) + '.r',  armIkRotCtrlRotVals[0], armIkRotCtrlRotVals[1], armIkRotCtrlRotVals[2] )
        

    if not ikModeChanged:
        
        pm.setAttr( toggleIconName + '.fkIk', 1 - ikModeOn )
    

def dnKillJob( jobNum ):

    pm.scriptJob( kill = jobNum, f = True )

def copyTransform( refObj ):

    '''
    copy transform object
    '''
    
    copyObj = mc.group( n = refObj + '_copy_grp', em = True )
    mc.delete( mc.parentConstraint( refObj, copyObj ) )
    
    return copyObj


def makeReverseLoc ( startJnt, endJnt, poleVector ):

    newLoc = mc.group( em = 1, world = True )
    mc.delete ( mc.parentConstraint ( startJnt, newLoc ) )

    mc.delete ( mc.aimConstraint ( endJnt, newLoc, upVector=(0, 0, 1), worldUpObject = poleVector, worldUpType='object', aimVector=(0, 1, 0) ) )

    return newLoc

def getCtlTransform ( ctrl, ctrlJnt, toggleIconName, ikModeOn, returnPosition = False ):

    if not mc.ls (ctrl):
        return False

    ctrlParent = mc.listRelatives( ctrl, p = True )[0]
    rotateOrder = mc.getAttr( ctrl + '.ro' )

    pm.setAttr( toggleIconName + '.fkIk', 1 - ikModeOn )
    
    jntTransf = mc.group( em = 1, world = True )
    mc.setAttr( jntTransf + '.ro', rotateOrder )
    mc.delete ( mc.parentConstraint ( ctrlJnt, jntTransf  ))
    

    
    ctlTransf = mc.group( em = 1, world = True )
    mc.setAttr( ctlTransf + '.ro', rotateOrder )
    mc.delete ( mc.parentConstraint ( ctrl, ctlTransf ))
    mc.parent ( ctlTransf, jntTransf )
    
    pm.setAttr( toggleIconName + '.fkIk', ikModeOn )
    
    mc.parent ( ctlTransf, ctrlJnt, relative = True )
    mc.parent ( ctlTransf, ctrlParent )

    ikLocTranslation = mc.getAttr( ctlTransf + '.t' )[0]
    ikLocRotation = mc.getAttr( ctlTransf + '.r' )[0]
    
    if mc.nodeType (ctrl) == 'joint':

        mc.select ( cl = True )
        ctlJntTransf = mc.joint( )
        mc.select ( cl = True )
        mc.parent ( ctlJntTransf, ctrlParent ) 
        rotateOrder = mc.getAttr( ctrl + '.ro' )
        mc.setAttr( ctlJntTransf + '.ro', rotateOrder )
        
        for axis in ['X', 'Y', 'Z']:
            val = mc.getAttr ( ctrl + '.jointOrient' + axis )
            mc.setAttr ( ctlJntTransf + '.jointOrient' + axis, val )

        mc.delete ( mc.parentConstraint ( ctlTransf, ctlJntTransf ))

        ikLocTranslation = mc.getAttr( ctlJntTransf + '.t' )[0]
        ikLocRotation = mc.getAttr( ctlJntTransf + '.r' )[0]
        
        mc.delete ( ctlJntTransf )
    
    if ctrl == 'Lt_LegAnkleIk_CON':
        asdf
    
    mc.delete ( jntTransf, ctlTransf )

    if returnPosition:
        
        return ikLocTranslation
    
    else:
        
        return ikLocRotation
    

def getPoleVectorPosition( startObj, midObj, endObj, offsetMulti = 1.0, convex = True, keepWorldY = True ):
    
    '''
    Get position for placement of pole vector constraint target, given 3 positions of limb joints
    Concave option should be True to find position where the limb joint "aims"
    Also, points have to form a triangle to get correct result
    
    @param startPos: start position of limb
    @type startPos: list(3 floats)
    @param midPos: mid position of limb
    @type midPos: list(3 floats)
    @param endPos: end position of limb
    @type endPos: list(3 floats)
    @param offsetMulti: offsetMulti multiply value for pole vector null
    @type offsetMulti: float
    @param convex: if true, return position where limb`s triangle is "aiming", otherwise go opposite direction
    @type convex: bool
    @param keepWorldY: if true maintain world Y position of middle position (midPos), useful for positioning arm/leg pole vector
    @type keepWorldY: bool
    @return: list(3 floats), position for pole vector null
    @author: jakub
    '''
    
    startPos = mc.xform( startObj, q = True, t = True, ws = True )
    midPos = mc.xform( midObj, q = True, t = True, ws = True )
    endPos = mc.xform( endObj, q = True, t = True, ws = True )
    
    startvector = om.MVector( startPos[0], startPos[1], startPos[2] )
    midvector = om.MVector( midPos[0], midPos[1], midPos[2] )
    endvector = om.MVector( endPos[0], endPos[1], endPos[2] )
    
    # get vectors between start-mid and end-mid points
    startvector = midvector - startvector
    endvector = midvector - endvector
    
    upperArmLength = startvector.length()
    
    # get pole vector from start-mid and end-mid vectors cross product
    polevector = old_div(( startvector + endvector ), 2)
    
    if not convex: polevector = polevector * ( -1 )
    
    polevector.normalize()
    
    # count offset value
    pvOffset = upperArmLength * offsetMulti
        
    polevectoroffset = midvector + ( polevector * pvOffset )
    polevectorPos = [ polevectoroffset.x, polevectoroffset.y, polevectoroffset.z ]
        
    return polevectorPos

