"""
=====================================================================

    thomasBittnerSmoothSkin 5.5
    smoothes skin, and looks at locked influences

    Made by Thomas Bittner

    cmds.thomasBittnerSmoothSkin(iterations=8, ignoreJointLocks=False )
    cmds.thomasBittnerSmoothSkin(iterations=8, ignoreJointLocks=False, forceMaxInfluences=3 )
    cmds.thomasBittnerSmoothSkin(iterations=8, rigid=1 )

=====================================================================
"""


import maya.OpenMayaMPx as OpenMayaMPx
import maya.api.OpenMaya as OpenMaya2
import maya.OpenMaya as OpenMaya
import maya.api.OpenMayaAnim as OpenMayaAnim2
import maya.cmds as cmds

import sys
import maya.mel
import math
import numpy as np

# name our command
kPluginCmdName="thomasBittnerSmoothSkin"
  
kIterationsFlag = "-i"
kIterationsFlagLong = "-iterations"
kIgnoreJointLocksFlag = "-ijl"
kIgnoreJointLocksFlagLong = "-ignoreJointLocks"
kForceMaxInfluencesFlag = "-m"
kForceMaxInfluencesFlagLong = "-forceMaxInfluences"
kRigidFlag = "-r"
kRigidFlagLong = "-rigid"

class thomasBittnerSmoothSkinCommand(OpenMayaMPx.MPxCommand):
 
    def __init__(self):
        OpenMayaMPx.MPxCommand.__init__(self)
        


    def doIt(self, args):

        # find out how many iterations
        #
        print('running thomasBittnerSmoothSkin 5.5')
        syntax = OpenMaya.MSyntax()
        syntax.addFlag( kIterationsFlag, kIterationsFlagLong, OpenMaya2.MSyntax.kLong)
        syntax.addFlag( kIgnoreJointLocksFlag, kIgnoreJointLocksFlagLong, OpenMaya2.MSyntax.kBoolean)
        syntax.addFlag( kForceMaxInfluencesFlag, kForceMaxInfluencesFlagLong, OpenMaya.MSyntax.kLong)
        syntax.addFlag( kRigidFlag, kRigidFlagLong, OpenMaya.MSyntax.kDouble)

        argData = OpenMaya.MArgDatabase( syntax, args )
        
        iterations = 8
        ignoreJointLocks = False
        ignoreSelfPoint = True
        
        if argData.isFlagSet( kIterationsFlag ):
            iterations  = argData.flagArgumentInt( kIterationsFlag, 0 )  
        if argData.isFlagSet( kIgnoreJointLocksFlag ):
            ignoreJointLocks  = argData.flagArgumentBool( kIgnoreJointLocksFlag, False )
        if argData.isFlagSet(kForceMaxInfluencesFlag):
            maxInfluences = argData.flagArgumentInt(kForceMaxInfluencesFlag, False)
        else:
            maxInfluences = -1
        if argData.isFlagSet( kRigidFlag):
            fRigid = argData.flagArgumentDouble( kRigidFlag, 0 )
            fPower = 1.0 + fRigid * 0.25
        else:
            fRigid = 0.0

        # get selection
        #
        self.meshName, verts, softValues = getVertexSelection()
        self.aVertIds = np.array(verts, dtype=int)
        totalVertexCount = cmds.polyEvaluate(self.meshName, vertex=True)


        # find neighbours of all verts
        #
        iMaxNeighborCount = 8
        aNbsArray = np.empty(self.aVertIds.size * iMaxNeighborCount, dtype=int).reshape(-1, iMaxNeighborCount)
        aNbsArray.fill(totalVertexCount)
        nbIdsArray = []
        aNeighborSizesDenoms = np.empty(self.aVertIds.size)
        meshDag = getDagPath(self.meshName)
        itVertex = OpenMaya2.MItMeshVertex(meshDag)
        for i, vert in enumerate(verts):
            itVertex.setIndex(vert)
            nbIds = itVertex.getConnectedVertices() + [vert]
            minCount = min(iMaxNeighborCount,len(nbIds))
            aNbsArray[i,:minCount] = nbIds[:minCount]
            aNeighborSizesDenoms[i] = minCount
        aNeighborSizesDenoms = (1 / aNeighborSizesDenoms)[:,np.newaxis]
        
        # get vertex list that includes neighbours
        #
        aAllVertIds = np.concatenate([aNbsArray.ravel(),self.aVertIds]) if ignoreSelfPoint else aNbsArray.ravel()
        aAllVertIds = np.unique(aAllVertIds)
        aAllVertIds.sort()
        

        # create mappers and localize aNbsArray
        #
        aMapper = np.zeros(totalVertexCount+1, dtype=int)
        aMapper[aAllVertIds] = np.arange(aAllVertIds.size)
        aMapperSel = aMapper[self.aVertIds]
        aNbsArray = aMapper[aNbsArray]


        # get weights (we add another vertex on top that has zero influences on each joint for vertices that have less neighbors than the maxNeighborCount))
        #
        aWeights, influences, self.infCount = getSkinClusterWeights(self.meshName, aAllVertIds)
        if maxInfluences == -1 and cmds.skinCluster(self.meshName, q=True, obeyMaxInfluences=True):
            maxInfluences = cmds.skinCluster(self.meshName, q=True, maximumInfluences=True)


        aWeights2d = np.concatenate([aWeights, np.zeros(self.infCount)]).reshape(-1,self.infCount)
        aWeights2dOld = np.copy(aWeights2d)

        # get locked joints
        #
        if not ignoreJointLocks:
            locks = [cmds.getAttr('%s.liw' % influence) for influence in influences]
            aLocked = np.array(locks, dtype=bool)
            aNonLocked = ~aLocked # was -aLocked
            aNonLockedSumsBefore = np.sum(aWeights2d[:,aNonLocked], axis=1)

        # now iterate through everything and smooth
        #
        if maxInfluences > 0 and maxInfluences < self.infCount:
            print('respecting maximum influences: ', maxInfluences)
        for iteration in range(iterations):

            aMultipls = aWeights2d[aNbsArray] * aNeighborSizesDenoms[..., np.newaxis]
            if fRigid > 0.0:
                aMultipls = np.power(aMultipls, fPower)
            aWeights2d[aMapperSel] = np.sum(aMultipls, axis=-2)
            if fRigid > 0.0:
                aSums = np.sum(aWeights2d, axis=-1)
                aSums[aSums < 0.00001] = 1
                aSumsDenom = 1.0 / aSums
                aWeights2d *= aSumsDenom[..., np.newaxis]

            if maxInfluences > 0 and maxInfluences < self.infCount:
                aRemoveInfluences2d = np.argsort(aWeights2d[aMapperSel], axis=1)[:,np.arange(self.infCount-maxInfluences)]
                aWeights2d[aMapperSel[:,np.newaxis], aRemoveInfluences2d] = 0.0

        # joint locks
        #
        if not ignoreJointLocks:
            aWeights2d[:,aLocked] = aWeights2dOld[:,aLocked]
            aNonLockedSums = np.sum(aWeights2d[:,aNonLocked], axis=1)
            aNonLockedSums[aNonLockedSums == 0] = 1
            division = aNonLockedSumsBefore / aNonLockedSums
            aWeights2d[:,aNonLocked] *= division[:,np.newaxis]

        # isolating arrays to selection
        #
        self.aWeights2dSel = aWeights2d[aMapperSel]
        self.aWeights2dSelOld = aWeights2dOld[aMapperSel]


        # handling softselection
        #
        if softValues != None:
            aSoft2d = np.array(softValues, dtype='float64')[:,np.newaxis]
            self.aWeights2dSel = self.aWeights2dSel * aSoft2d + self.aWeights2dSelOld * (1-aSoft2d)
        self.redoIt()

        
    def redoIt(self):
        setSkinClusterWeights(self.meshName, list(self.aVertIds), list(self.aWeights2dSel.ravel()), self.infCount)

    def undoIt(self):
        setSkinClusterWeights(self.meshName, list(self.aVertIds), list(self.aWeights2dSelOld.ravel()), self.infCount)
 
    def isUndoable(self):
        return True

    def isRedoable(self):
        return True




def cmdCreator():
    return OpenMayaMPx.asMPxPtr( thomasBittnerSmoothSkinCommand() )
 
 
def initializePlugin(mobject):
    mplugin = OpenMayaMPx.MFnPlugin(mobject, "Thomas Bittner", "5.0", "Any")
    try:
        mplugin.registerCommand( kPluginCmdName, cmdCreator)
    except:
        sys.stderr.write( "Failed to register command: %s\n" % kPluginCmdName )
        raise
 
 
def uninitializePlugin(mobject):
    mplugin = OpenMayaMPx.MFnPlugin(mobject)
    try:
        mplugin.deregisterCommand( kPluginCmdName )
    except:
        sys.stderr.write( "Failed to unregister command: %s\n" % kPluginCmdName )
        raise
 



def getVertexSelection(bIgnoreSoftSelection=False):
    softSelection = OpenMaya2.MGlobal.getRichSelection()
    selection = softSelection.getSelection()

    iter = OpenMaya2.MItSelectionList( selection, OpenMaya2.MFn.kMeshVertComponent )
    elements, weights = [], []
    while not iter.isDone():
        dagPath, component = iter.getComponent()
        fnComp = OpenMaya2.MFnSingleIndexedComponent(component)
        dagPath.pop()
        node = dagPath.fullPathName()

        elements = [fnComp.element(i) for i in range(fnComp.elementCount)]
        if fnComp.hasWeights:
            weights = [fnComp.weight(i).influence for i in range(fnComp.elementCount)]
        else:
            weights = None

        iter.next()

    return node, elements, weights

def getDagPath(sName):
    selList = OpenMaya2.MSelectionList()
    selList.add(sName)
    return selList.getDagPath(0)

def getDependNode(sName):
    selList = OpenMaya2.MSelectionList()
    selList.add(sName)
    return selList.getDependNode(0)


def getSkinClusterWeights(meshName, ids):
    skinPath = getDependNode(maya.mel.eval('findRelatedSkinCluster %s' % meshName))
    meshPath = getDagPath(meshName)
    fnSkinCluster = OpenMayaAnim2.MFnSkinCluster(skinPath);
    fnVtxComp = OpenMaya2.MFnSingleIndexedComponent()
    vtxComponents = fnVtxComp.create( OpenMaya2.MFn.kMeshVertComponent )
    fnVtxComp.addElements(ids);
    
    dWeights, infCount = fnSkinCluster.getWeights(meshPath, vtxComponents)
    aWeights = np.array(list(dWeights), dtype='float64')
    influences = [dp.fullPathName() for dp in fnSkinCluster.influenceObjects()]

    return aWeights, influences, infCount


def setSkinClusterWeights(meshName, ids, weights, influenceCount):
    skinPath = getDependNode(maya.mel.eval('findRelatedSkinCluster %s' % meshName))
    meshPath = getDagPath(meshName)
    fnSkinCluster = OpenMayaAnim2.MFnSkinCluster(skinPath);
    fnVtxComp = OpenMaya2.MFnSingleIndexedComponent()
    vtxComponents = fnVtxComp.create( OpenMaya2.MFn.kMeshVertComponent )
    fnVtxComp.addElements(ids);
    
    infArray = OpenMaya2.MIntArray(range(influenceCount))
    dWeights = OpenMaya2.MDoubleArray(weights)
    fnSkinCluster.setWeights(meshPath, vtxComponents, infArray, dWeights)



