'''
otherUtils.py
@author www.RIG-IT.net

Last Update: 2022-12-25

various simpler tools serving animation of RIG-IT rigs
'''
from __future__ import print_function

import maya.cmds as mc
import maya.mel as mm

from .utils import name
from .utils import shape

topGrpAt = 'rigitTopGrp'
zeroPoseAt = 'zeroPose'
bindPoseAt = 'bindPose'


class otherUtilsWindow:
    
    def __init__(self):
        
        self.window = 'otherUtilsWindow'
        self.title = 'RIG-IT_otherUtils'
        self.size = (300,200)
        
        # check if window is already opened
        if mc.window( self.window, exists = True ):
            mc.deleteUI( self.window, window = True )
        
        # window creation
        self.window = mc.window( self.window, t = self.title, widthHeight = self.size )
        
        mainCL = mc.columnLayout( adjustableColumn=True, columnAlign= 'center' )

        mc.text( label='RIG-IT Other Utils' )
        mc.text( label='' )
        mc.text( label='Zero Pose', font= 'boldLabelFont' )
        mc.separator( st = 'none', hr = False, h = 20 )

        rowOff= 30

        mc.rowColumnLayout( numberOfColumns = 3, adjustableColumn=2, columnWidth= [(1,rowOff),(2,rowOff),(3,rowOff)], p=mainCL )
        mc.text( l='' )
        mc.button( label = 'Reset All Rig', command = self.resetRig )
        mc.text( l='' )

        mc.rowColumnLayout( numberOfColumns = 3, adjustableColumn=2, columnWidth= [(1,rowOff),(2,rowOff),(3,rowOff)], p=mainCL )
        mc.text( l='' )
        mc.button( label = 'Reset Selected Controls Only', command = lambda *args: self.resetRig(True) )
        mc.text( l='' )

        mc.separator( st = 'none', hr = False, h = 40 )
        mc.text( label='Scene Cleanup', font= 'boldLabelFont' )
        mc.separator( st = 'none', hr = False, h = 20 )

        mc.rowColumnLayout( numberOfColumns = 3, adjustableColumn=2, columnWidth= [(1,rowOff),(2,rowOff),(3,rowOff)], p=mainCL )
        mc.text( l='' )
        mc.button( label = 'Fix Shape Names', command = self.fixShapeNames )
        mc.text( l='' )
        
        # display window
        mc.showWindow(self.window)
        mc.window( self.window, e = True, w = self.size[0], h = self.size[1] )
    
    def resetRig(self, onlySelection= False):

        sel = mc.ls(sl=1)
        
        if not sel:
            mc.confirmDialog( title = 'Reset Rig', message = 'Please select any control(s) from rig to reset')
            return

        ns = name.getNamespace(sel[0])
        loadControlPose( rigns = ns, posetype = 'zero', onlySelection = onlySelection, selection= sel )
    
    def fixShapeNames(self, args):

        shape.fixNamesAllScene()





def mainUI():

    mainUI = otherUtilsWindow()


# ------------------------------
# system functions
# ------------------------------

def loadControlPose( rigns = ':', posetype = 'zero', skipcontrols = ['all1_ctl', 'all2_ctl', 'global1_ctl', 'global2_ctl'], verbose = True,
                    onlySelection = False, selection= None ):
    
    '''
    set zero pose to rig
    
    @param rigns: str, rig namespace
    @param posetype: str, type of pose, valid names are: 'zero', 'bind'
    @param skipcontrols: list(str), list of default controls to skip posing
    @param verbose: bool, print some info from script
    @param onlySelection: bool, load pose only for selected controls
    @param selection: list(str), selected controls only used when onlySelection is True, both with namespace or without can be used
    @return: None
    '''
    
    # print initial empty line    
    if verbose:
        
        print('')
    
    #===========================================================================
    # fix namespace string
    #===========================================================================
    
    if not rigns.endswith( ':' ):
        
        rigns += ':'
        
    if rigns == ':':
        
        rigns = ''
    
    #===========================================================================
    # find top group
    #===========================================================================
    
    topgroups = mc.ls( rigns + '*.' + topGrpAt, o = 1 )
    if not topgroups:
        
        raise Exception( '# top group not found under namespace "%s"' % rigns )
    
    topgroup = topgroups[0]
    
    #===========================================================================
    # check which pose was chosen and the pose attribute
    #===========================================================================
    
    poseat = ''
    
    if posetype == 'zero':
        
        poseat = zeroPoseAt
        
    if posetype == 'bind':
        
        poseat = bindPoseAt
    
    if poseat == '':
        
        raise Exception( '# %s is not valid pose name, idle...' % rigns )
    
    if not mc.objExists( topgroup + '.' + poseat ):
        
        topgroup = 'char_grp'
        # raise Exception( '# %s.%s not found, cannot set %s pose' % ( topgroup, poseat, posetype ) )
    
    # ---------------------------------------------
    # remove any namespaces from controls Selection
    # ---------------------------------------------

    selectionNoNs = []
    if onlySelection and selection:

        for s in selection:

            noNs = name.removeNamespace(s)
            selectionNoNs.append( noNs )

    #===========================================================================
    # get values and set them to controls
    #===========================================================================
    
    posevalues = mc.getAttr( topgroup + '.' + poseat )
    
    if posevalues == '':
        
        if verbose:
            
            print('# no pose values stored in %s were found' % ( topgroup + '.' + poseat ))
            
        return
    
    ctrlat_value_pairs = posevalues.split()
    
    # turn off autokey
    autoKeyframeVal = mc.autoKeyframe( q = 1, state = True )
    mc.autoKeyframe( state = False )
    
    for i in range( 0, len( ctrlat_value_pairs ), 2 ):
        
        ctrlat = rigns + ctrlat_value_pairs[i]
        value = float( ctrlat_value_pairs[i + 1] )
        _loadControlPose_setvalues( ctrlat, value, skipcontrols, verbose, onlySelection, selectionNoNs )
    
    # turn on autokey
    mc.autoKeyframe( state = autoKeyframeVal )
    
    if verbose: print('# %s pose set for rig in namespace: %s' % ( posetype, rigns ))
    
    # print an empty line to divide from other comments in log    
    if verbose: print('')
    

def _loadControlPose_setvalues( objattribute, value, skipcontrols, verbose = True, onlySelection= False, selectionNoNs= None ):
    
    '''
    sub-function of loadControlPose to set value to control 
    '''
    
    #===========================================================================
    # perform checks and process namespace, control and attribute names
    #===========================================================================
    
    if not mc.objExists( objattribute ):
        
        if verbose: print('# %s doesn`t exist' % objattribute)
        return
    
    objattributeclean = objattribute    
    
    # remove namespace
    
    if objattribute.count( ':' ):
        
        objattributeclean = objattributeclean.split( ':' )[-1]
    
    # check if we have object.attribute string
    
    if not objattributeclean.count( '.' ):
        
        if verbose: print('# %s is not in object.attribute format, skipping ...' % objattributeclean)
        return
    
    # get object and attribute
    
    objattributelist = objattributeclean.split( '.' )
    ctrlobj = objattributelist[0]
    at = objattributelist[1]
    
    # check if we want to skip this control
    
    if ctrlobj in skipcontrols:
        
        if verbose: '# %s was in Skip List, skipping ...' % ctrlobj
        return
    
    # skip other then Selection controls
    if onlySelection and selectionNoNs:

        if not ctrlobj in selectionNoNs:

            if verbose: '# %s was not in Selection, skipping ...' % ctrlobj
            return

    
    #===========================================================================
    # set value to control
    #===========================================================================
    
    # perform steps for settings control values on humanIk controls
    humanIkControlTypes = ['hikIKEffector', 'hikFKJoint']
    objType = mc.nodeType( objattribute )
    
    if objType in humanIkControlTypes:
        
        objTransform = mc.ls( objattribute, o = True )[0]
        mc.select( objTransform )
        
        hikManipCmd = 'hikManipStart 1 0'  # translate attribute is assumed by default
        if objattribute.count( '.rotate' ):
            
            hikManipCmd = 'hikManipStart 0 1'
        
        mm.eval( hikManipCmd )
        
    # set values on control attribute
    try:
        
        mc.setAttr( objattribute, value )
        
    except:
        
        if verbose: print('# %s could not be set value of %f' % ( objattribute, value ))
        return
    
    # end humanIk manipulator mode
    if objType in humanIkControlTypes:
        
        mm.eval( 'hikManipStop' )


