"""
referenceSwitch.py

Description:

Tool for making Maya reference proxy switch setup. Example use can be rig variations with the same controls
but different geometry or slightly different positions of these controls.
Maya can then switch between those files while keeping animation keyframes attached to controls.
This is using Reference Edits system in Maya. 

made by: www.rig-it.net, (c)2019
"""

import maya.cmds as mc
import maya.mel as mm


def makeProxy( namespaceName, filepathProxies, proxyTags = ['a', 'b', 'c'] ):

    '''
    make proxy from files
    NOTE:
    Very useful for switching between more rig LODs with the same controls and different geometry, deform setups etc.
    Number of items in filepathProxies and proxyTags has to be the same
    
    @param namespaceName: str, namespace name
    @param filepathProxies: list( str),  list of filepaths to be under one reference as proxies, any number of items can be used
    @param proxyTags: list(str), tags to be added to proxies, number of items needs to be the same like filepathProxies
    @return: str, reference node made by Maya
    '''
    
    #===========================================================================
    # check input values
    #===========================================================================
    
    filesLen = len( filepathProxies )
    proxyTagsLen = len( proxyTags )
    
    if not filesLen == proxyTagsLen:
        
        raise Exception( 'number of files and proxy names must be the same' ) 
    
    # =====================================
    # create proxy with shared edits
    # =====================================

    # make sure Shared Edits is ON
    mc.optionVar( intValue = ( 'proxyOptionsSharedEdits', 1 ) )
    
    # make reference (first proxy)
    refFile = mc.file( filepathProxies[0], reference = 1, namespace = namespaceName )
    refNode_a = mc.referenceQuery( refFile, referenceNode = 1 )
    
    # add the rest proxies
    for i in range( filesLen - 1 ):
        
        idx = i + 1
        proxyAddMelCmd = 'proxyAdd "%s" "%s" "%s";' % ( refNode_a, filepathProxies[idx], proxyTags[idx] )
        mm.eval( proxyAddMelCmd )
    
    # fix first proxy names
    # needs to be done after adding proxies or they would inherit first name too
    mc.setAttr( refNode_a + '.proxyTag', proxyTags[0], typ = 'string' )
    mc.lockNode( refNode_a, l = 0 )
    
    # rename proxy nodes
    refNodes = getAllProxiesFromRefNode( refNode_a )
    for i in range( len( refNodes ) ):
        
        mc.lockNode( refNodes[i], l = False )
        refNodes[i] = mc.rename( refNodes[i], namespaceName + proxyTags[i] + '_rfn' )
        
    # make sure we are in default namespace
    mc.namespace( set = ':' )

    return refNodes



def switchProxy( fileRefNode ):

    '''
    switch to proxy of referenced file specified by file reference node
    '''

    mm.eval( 'proxySwitch %s' % fileRefNode )


def getAllProxiesFromRefNode( fileRefNode ):

    '''
    Get all proxy file reference nodes from one reference node
    NOTE: Using connection to proxy manager
    '''

    proxyManagerNode = mc.listConnections( fileRefNode + '.proxyMsg', t = 'proxyManager' )[0]
    allProxies = mc.listConnections( proxyManagerNode + '.proxyList', s = 1, t = 'reference' )

    return allProxies

def getCurrentProxy( proxyManagerNode ):

    '''
    get current active proxy from provided proxyManage node
    '''

    currentProxyNode = mc.listConnections( proxyManagerNode + '.sharedEditsOwner', s = 1, t = 'reference' )[0]

    return currentProxyNode


