'''
Save Shaders
@author www.RIG-IT.net

v1.0.0 created 2021-03

tool and functions for saving and loading Maya scene shaders
'''
from __future__ import print_function

import json
import os
from functools import partial

import maya.cmds as mc
import maya.mel as mm

from . import utils

shadersInfoKey = 'shadersInfo'
shadersInfoFilepathKey = 'shadersInfoFilepath'
shadersSceneFilepathKey = 'shadersSceneFilepath'
shaderNameKey = 'shaderName'
shaderTypeKey = 'shaderType'
shaderGeosKey = 'shaderGeos'
mayaAsciiExt = '.ma'
lockedShaders = ['lambert1']
defaultShaders = ['lambert1', 'standardSurface1']
toolName = 'saveShaders'
toolVersion = '1.0.1'

def save( filePath = '', objectList = None ):
    
    """
    Save all shaders in scene with their connection to scene geometry objects
    Extension of the JSON file should be .json
    Shaders will be exported to Maya Ascii scene in the same folder as filepath
    
    @param filePath: str, full file path for saving the json text file, extension is not needed
    @param objectList: list(str), optional, selected objects to save their shaders
    @return: None
    """
    
    # store original selection
    origSel = mc.ls( sl = True )
    
    # get export folder
    filedirpath = os.path.dirname( filePath )
    
    # check dir folder
    if not os.path.isdir( filedirpath ):
        
        raise Exception( 'problem with getting folder name, please check filepath' )
    
    # make filename for shaders Maya export scene
    filebase = os.path.basename( filePath )
    filenameNoExt = os.path.splitext( filePath )[0]
    filepathMayaShadersNoExt = os.path.join( filedirpath, filenameNoExt )
    filepathMayaShaders = filepathMayaShadersNoExt + mayaAsciiExt
    
    # get all shaders
    allShaders = utils.getAllShadersInScene()
    
    # get scene data
    sceneShadersData = {}
    
    sceneShadersData[shadersInfoKey] = []
    sceneShadersData[shadersInfoFilepathKey] = filePath
    sceneShadersData[shadersSceneFilepathKey] = filepathMayaShaders
    
    for s in allShaders:
        
        shaderData = {}
        
        shadersAssignedGeos = utils.getAssignedObjects( s )
        shaderType = mc.nodeType( s )
        
        shaderData[shaderNameKey] = s
        shaderData[shaderTypeKey] = shaderType
        shaderData[shaderGeosKey] = shadersAssignedGeos
        
        sceneShadersData[shadersInfoKey].append( shaderData )
    
    # save shaders info file
    fileobj = open( filePath, mode = 'w' )    
    json.dump( sceneShadersData, fileobj, sort_keys = True, indent = 4, separators = ( ',', ': ' ) )
    fileobj.close()
    
    # export shaders Maya scene
    mc.select( allShaders )
    mc.file( filepathMayaShaders, typ = 'mayaAscii', exportSelected = True, shader = True )
    
    if origSel:
        mc.select( origSel )  # restore original selection
    
    return sceneShadersData
    

def load( filePath = '', objectList = None, clearAllShaders = True, renameMatchingShaders = True, removeUnusedShaders = True ):
    
    """
    Load shaders from file to scene and apply to geometry objects
    File needs to be previously created with saveShaders.save() function
    
    @param filePath: str, full file path for saving the json text file, extension is not needed
    @param objectList: list(str), optional, apply shaders only to selected objects
    @param clearAllShaders: bool, clear all shaders before import, apply lambert1 to all scene geometry
    @param renameMatchingShaders: bool, if clearAllShaders is False, rename matching shaders before import (recommended)
    @param removeUnusedShaders: bool, optional, remove unused shaders after import
    @return: None
    
    Warning:
    - clearAllShaders could delete shaders in the rig
    - removeUnusedShaders can delete shaders not assigned to geometry but needed for later work 
    """
    
    # read from file
    fileobj = open( filePath, mode = 'rb' )
    fileobjStr = fileobj.read()
    sceneShadersData = json.loads( fileobjStr )
    fileobj.close()
    
    # prepare scene for shaders import
    if clearAllShaders:
        
        geoObjects = [ mc.listRelatives( o, p = True, fullPath=True )[0] for o in mc.ls( typ = ['mesh', 'nurbsSurface'], objectsOnly=True )]
        utils.assignDefaultLambert( geoObjects )
        mm.eval( 'MLdeleteUnused()' )
        mc.refresh()
        mc.dgdirty( a = True )
    
    
    # rename or delete matching shader names
    matchShaderRenamePrefix = 'old_'
    if renameMatchingShaders:
        
        for shaderData in sceneShadersData[shadersInfoKey]:
            
            shaderObj = shaderData[shaderNameKey]
            
            if shaderObj in defaultShaders:
                continue
            
            if mc.objExists( shaderObj ):
                
                # rename existing object - this could be existing shader
                mc.lockNode( shaderObj, l = 0 )
                mc.rename( shaderObj, matchShaderRenamePrefix + shaderObj )
    
    # import shaders
    exportedShadersFilepath = sceneShadersData[shadersSceneFilepathKey]
    mc.file( exportedShadersFilepath, i = True )
    
    # assign shaders to geometry
    count = 0
    # stop()
    for shaderData in sceneShadersData[shadersInfoKey]:
            
            # if count == 100: stop()
            shaderObj = shaderData[shaderNameKey]
            shadersAssignedGeos = shaderData[shaderGeosKey]
            
            print('shaderObj', shaderObj)
            print('shadersAssignedGeos', shadersAssignedGeos)
            
            utils.assign( shadersAssignedGeos, shaderObj, doRefresh = False )
            # count += 1
            
    
    if removeUnusedShaders:
        
        mm.eval( 'MLdeleteUnused()' )



#===============================================================================
# UI functions
#===============================================================================

def toolUI():
    
    """"
    build tool UI
    """
    
    window = utils.capitalizeFirst( toolName )
    windowObj = toolName
    if ( mc.window( windowObj, exists = True ) ):
        
        mc.deleteUI( windowObj )
    
    toolNameUI = utils.capitalizeFirst( toolName )
    window = mc.window( windowObj, title = "RIG-IT {0} {1}".format( toolNameUI, toolVersion ), widthHeight = ( 400, 300 ),
                            maximizeButton = True, resizeToFitChildren = True )
    
    mc.columnLayout( columnAlign = 'center' )
    
    mc.rowColumnLayout( numberOfColumns = 2, cw = [1, 150] )
    mc.text( '' )
    mc.text( '   Save Shaders   ', h = 30 )
    mc.setParent( '..' )
    
    mc.separator()
    mc.setParent( '..' )
    
    mc.columnLayout()
    fileBrowseTfb = mc.textFieldButtonGrp( label = 'Shaders File Path', text = 'example: C:/shaders/joe_shades.txt',
                                           buttonLabel = 'Browse', cw3 = ( 100, 200, 100 ) )
    mc.textFieldButtonGrp( fileBrowseTfb, e = True, buttonCommand = partial( uiFileBrowser, fileBrowseTfb ) )
    mc.setParent( '..' )
    
    mc.text( ' ' )
    
    mc.rowColumnLayout( numberOfColumns = 2, columnWidth = [( 1, 80 ), ( 2, 250 )] )    
    mc.text( '--> F' )
    saveShadersBt = mc.button( l = 'Save Shaders', annotation = 'save Maya scene shaders and their connections to geometry' )
    mc.button( saveShadersBt, e = True, c = lambda x: uiSaveShaders( fileBrowseTfb ) )
    mc.text( ' ' )
    mc.setParent( '..' )
    
    mc.rowColumnLayout( numberOfColumns = 2, columnWidth = [( 1, 80 ), ( 2, 250 )] )    
    mc.text( 'F -->' )
    loadShadersBt = mc.button( l = 'Load Shaders', annotation = 'load Maya scene and apply them to geometry' )
    mc.button( loadShadersBt, e = True, c = lambda x: uiLoadShaders( fileBrowseTfb ) )
    mc.setParent( '..' )
    
    mc.separator( h = 10, style = 'none' )
    
    
    mc.showWindow( window )
    mc.window( windowObj, e = True, w = 400, h = 120 )


def uiSaveShaders( fileBrowseTfb ):
    
    """
    save shaders
    """
    
    uiFilepath = mc.textFieldButtonGrp( fileBrowseTfb, q = True, text = True )
    
    # no file check performed since the file might not be existing yet
    save( filePath = uiFilepath )
    mc.confirmDialog( title = 'Save Shaders', message = '    Shaders succesfully saved    ' )


def uiLoadShaders( fileBrowseTfb ):
    
    """
    load shaders
    """
    
    uiFilepath = mc.textFieldButtonGrp( fileBrowseTfb, q = True, text = True )
    
    # perform file exist check
    if not os.path.isfile( uiFilepath ):
        
        mc.confirmDialog( title = 'Save Shaders', message = 'File path appears invalid, cannot open file' )
    
    else:
        
        load( filePath = uiFilepath )
        mc.refresh()
        mc.confirmDialog( title = 'Save Shaders', message = '    Shaders loading finished    ' )


def uiFileBrowser( fileBrowseTfb ):
    
    """
    open file dialog
    """
    
    multipleFilters = "Text Files (*.txt *.json)"
    res = mc.fileDialog2( fileFilter = multipleFilters, dialogStyle = 2 )
    
    if res:
        
        mc.textFieldButtonGrp( fileBrowseTfb, e = True, text = res[0] )

