'''
shader in rigit.utils
@category Rigging @subcategory Utils
@tags assign shader shadingEngine
@author jakub
'''
from __future__ import print_function

import maya.cmds as mc
import maya.mel as mm


def getShader( object ):
    
    '''
    get shader assigned to object (transform or shape can be passed)
    
    @param object: name of object to look for a shader
    @type object: str
    @return: str, name of shader, return None if no shader was found
    '''
    
    objshapes = mc.listRelatives( object, s = 1, f = 1, noIntermediate = 1 )
    if not objshapes: return None
    
    shadingengine = mc.listConnections( objshapes[0], type = 'shadingEngine' )
    if not shadingengine:
        return None
    
    shadernode = mc.listConnections( shadingengine[0] + '.surfaceShader' )
    if not shadernode:
        return None
    
    return shadernode[0]

def getAssignedObjects( shaderObj, transformsOnly = False ):
    
    '''
    get all objects or their components assigned to a shader
    
    @param object: str, name of object to look for a shaders
    @return: dict, return dictionary with object names as keys and their list of components or empty components

    Note: if components are empty it means that shader is assigned to whole object
    '''
    
    # store selection
    sel = mc.ls( sl = True )
    
    mc.hyperShade( objects = shaderObj )
    assignedObjs = mc.ls( sl = True )
    
    # restore selection
    if sel:
        mc.select( sel )
    
    if not assignedObjs:
        
        return
    
    # get transforms
    if transformsOnly:
        
        tList = []
        
        for obj in assignedObjs:
            
            ts = mc.ls( obj, objectsOnly = True )[0]  # returns shape
            t = mc.listRelatives( ts, p = True )[0]  # get transform
            tList.append( t )
        
        # remove duplicates
        assignedObjs = list( set( tList ) )
    
    return assignedObjs

def getAllShadersInScene():
    
    """
    get all shaders in scene
    """
    
    shadingEngines = mc.ls( type = 'shadingEngine' )
    shaders = []
    
    for se in shadingEngines:
        
        shaderNode = mc.listConnections( se + '.surfaceShader' )
        if shaderNode:
            
            shaders.append( shaderNode[0] )
    
    return shaders
    
def assignDefaultLambert( objects ):
    
    """
    assign default lambert1 shader to objects
    
    @param objects: list(str), list of geometry objects
    @return: None
    """
    
    mc.sets( objects, e = True, forceElement = 'initialShadingGroup' )
    
    
def assign( objects, shadernode, shadingEngine = None, doRefresh = False, verbose = True ):
    
    '''
    assign shader to objects
    (this can be list of transforms containing shapes, shapes or just components)
    
    @param objects: list(str), list of objects to get shader
    @param shadernode: str, name of shader node (must exist for now, later function can be extended to make a shader)
    @param shadingEngine: str (default None), name of render set (a.k.a. Shading Engine), will be made here if nothing passed
    @param doRefresh: bool, refresh scene after shader assignement
    @param verbose: bool, print status updates during function runtime
    @return: list( str, str ), [0] name of shader node [1] name of render set
    '''
    
    if not objects:
        
        if verbose:
            
            print('# saveShaders assign(): no objects passed for shader assignement, skipping ...')
        
        return
    
    if shadernode == 'lambert1':
        
        shadingEngine = 'initialShadingGroup'
        
    # filter out non existing objects
    existObjects = []
    missingObjects = []
    
    for o in objects:
        
        if mc.objExists( o ):
            existObjects.append( o )
        
        else:
            missingObjects.append( o )
    
    # print report on missing objects
    if verbose:
        print( '# saveShaders.assign: these missing objects will be skipped: %s' % missingObjects )
    
    # idle function if no objects from the list were found in scene
    if not existObjects:
        
        return
    
    if not shadingEngine or not mc.objExists( shadingEngine ):
        
        shadingEnginename = removeSuffix( shadernode )
        shadingEngine = mc.sets( existObjects, renderable = True, noSurfaceShader = True, empty = True, name = '%s_shg' % shadingEnginename )
        
    mc.sets( existObjects, e = 1, forceElement = shadingEngine )
    
    # this is Maya bug, even with connectAttr force=1 errors on existing connection, otherwise gives warning, which would result in loads of warnings
    if not mc.isConnected( shadernode + '.outColor', shadingEngine + '.surfaceShader' ):
        
        mc.defaultNavigation( connectToExisting = True, source = shadernode, destination = shadingEngine )
        
    # mm.eval( 'connectNodeToNodeOverride("%s", "%s")' % ( shadernode, shadingEngine ) )
    
    return [ shadernode, shadingEngine ]


def duplicateDefaultLambert( prefix ):
    
    '''
    as duplicate command cannot be used on default lambert shader node,
    this function does it (makes lambert node and copies all lambert1 values to it)
    
    @param prefix: prefix for naming new lambert
    @type prefix: str
    @return: str, name of new lambert
    '''
    
    l1 = 'lambert1'  # default lambert material cannot be renamed, so it`s always named lambert1 (handy!)
    lambertats = mc.listAttr( l1, v = 1, se = 1, scalar = 1 )
    values = [ mc.getAttr( l1 + '.' + a ) for a in lambertats ]
    lambertdup = mc.createNode( 'lambert', n = prefix + '_shd' )
    for a, v in zip( lambertats, values ): mc.setAttr( lambertdup + '.' + a, v )
    
    return lambertdup
    
def setTextureHardwareResolution( shaderNode, resolution = 128, textureAt = 'color' ):
    
    '''
    Set custom hardware texture resolution for selected shader, texture attribute can be changed as well
    
    @param shaderNode: name of shader node
    @type shaderNode: str
    @param resolution: value for custom square resolution
    @type resolution: int
    @param textureAt: name of texture attribute, color by default
    @type textureAt: str
    @return: None
    '''
    
    if not mc.objExists( shaderNode + '.' + textureAt ): return
    
    textureNodeRes = mc.listConnections( shaderNode + '.' + textureAt )
    
    if not textureNodeRes: return
    
    addResolutionAt( textureNodeRes[0], resolution = resolution )
    

def addResolutionAt( textureNode, resolution = 256 ):
    
    '''
    Add custom hardware texture resolution attribute, which Maya system can recognize
    
    @param textureNode: name of texture node to get resolution attribute
    @type textureNode: str
    @param resolution: value for custom square resolution
    @type resolution: int
    @return: None
    '''
    
    resolutionAt = 'resolution'
    
    if not mc.objExists( textureNode + '.' + resolutionAt ):
        
        mc.addAttr( textureNode, ln = resolutionAt, at = 'long', dv = resolution )
    
    mc.setAttr( textureNode + '.' + resolutionAt, resolution )
    

def createConnect2Dtexture( textureType = 'file', textureInputPlug = '', prefix = 'mytexture' ):
    
    '''
    Wrapper to Maya mel script to create 2D texture map with place2dTexture node and its connections
    
    @param textureType: type of 2D texture, default is file
    @type textureType: str
    @param textureInputPlug: texture input attribute (OBJECT.ATTRIBUTE)
    @type textureInputPlug: str
    @return: list(str), list with names of new nodes (texture, texturePlacement) 
    '''
    
    # create MEL command from parameters
    
    mainCreateTextureCommand = 'createRenderNodeCB - as2DTexture "" %s' % textureType
    renderNodeCommand = 'defaultNavigation -force true -connectToExisting -source %node -destination {0}'.format( textureInputPlug )
    
    createRenderNodeCBCommand = '%s "%s"' % ( mainCreateTextureCommand, renderNodeCommand )
    
    # run MEL command
    
    textureNode = mm.eval( createRenderNodeCBCommand )
    place2dTextureNode = mc.listConnections( textureNode + '.uvCoord' )[0]
    
    # rename nodes
    textureNode = mc.rename( textureNode, prefix + '_fil' )
    place2dTextureNode = mc.rename( place2dTextureNode, prefix + '_p2d' )
    
    
    return [textureNode, place2dTextureNode]

    
    
def makeSkinTestShader( assignGeoList = [], prefix = 'skinTest', baseColor = [0.5, 0.5, 0.5], colorOffset = 0.03, repeatUV = [20, 20] ):
    
    '''
    make test shader with subtle color checker texture,
    good to test deformations and sliding skin
    
    @param assignGeoList: optional, list of geometry objects (or groups with geometry) to get this shader assigned
    @type assignGeoList: list(str)
    @param prefix: prefix for naming all new nodes
    @type prefix: str
    @param baseColor: set color of the checker, the second color will be offset from this one
    @type baseColor: list(float), 3 float RGB values normalized in 0-1 range
    @param colorOffset: offset base color RGB values to generate second color
    @type colorOffset: float
    @param repeatUV: checker repeat UV coordinates
    @type repeatUV: list(float), list of 2 float or int values
    @return: list(str), list with new created nodes 0- shader, 1- texture, 2- place2Dtexture
    '''
    
    blinnShader = mc.shadingNode( 'blinn', asShader = True, n = prefix + '_shd' )
    textureNode, place2dNode = createConnect2Dtexture( textureType = 'checker', textureInputPlug = blinnShader + '.c' )
    setTextureHardwareResolution( blinnShader, resolution = 256 )
    
    mc.setAttr( blinnShader + '.specularColor', 0.5, 0.5, 0.5 )
    mc.setAttr( blinnShader + '.eccentricity', 0.3 )
    mc.setAttr( textureNode + '.color1', baseColor[0], baseColor[1], baseColor[2] )
    
    offsetColor = [ baseColor[0] - colorOffset, baseColor[1] - colorOffset, baseColor[2] - colorOffset]
    mc.setAttr( textureNode + '.color1', baseColor[0], baseColor[1], baseColor[2] )
    mc.setAttr( textureNode + '.color2', offsetColor[0], offsetColor[1], offsetColor[2] )
    
    mc.setAttr( place2dNode + '.repeatU', repeatUV[0] )
    mc.setAttr( place2dNode + '.repeatV', repeatUV[1] )
    
    # assign shader to objects
    
    if assignGeoList:
        
        assign( assignGeoList, blinnShader )
    
    return [blinnShader, textureNode, place2dNode]
    

def makeShader( prefix = 'newshader', shaderType = 'lambert', assignToObjects = None ):
    
    """
    make shader and optionally assign to objects
    
    @param prefix: str, prefix for naming new objects
    @param shaderType: str, type of shader
    @param assignToObjects: list(str), optional, list of objects for assigning new shader
    @return: str, name of new shader 
    """
    
    shaderNode = mc.shadingNode( shaderType, asShader = True, n = prefix + '_shd' )
    
    if assignToObjects:
        
        assign( assignToObjects, shadernode = shaderNode )
    
    
    return shaderNode


        
def checkerMaterialBlend( materialObjects, controlObject = None, applyShaders = True, shareOneChecker = True, checkerTextureNode = None, prefixMergeList = None ):
    """
    To add layeredShader to all materialObjects with original material and possibility to blend checker in

    @param materialObjects: list, list of geometry objects with shaders to get checker blend material
    @param controlObject: str, Optional - object for adding the blend attributes (for example Vis control), default None
    @param applyShaders: bool, option to keep shaders only made, not applied to geometry
    @param shareOneChecker: bool, share one checker shader for all objects
    @param checkerTextureNode: str, optional - None by default, here existing checker texture node can be passed
    @param prefixMergeList: list, optional list of prefixes of shaders to be merge with not prefixed shaders
    @return: dictionary, list of new shaders, name of checker texture node  
    
    """

    layeredShaders = []
    checkerShaders = []

    # checker shader creation
    if checkerTextureNode:

        checkerShader = checkerTextureNode
        checkerSet = mc.listConnections ( checkerShader + '.outColor', type = 'objectSet' )[0]
        checkerShaders.append ( checkerShader )

    elif shareOneChecker:

        checkerShader, checkerSet = _createCheckerMaterial ( 'Shared' )
        checkerShaders.append ( checkerShader )

    
    # filter of all scene shaders
    allShaders = mc.ls ( type = 'lambert' )
    
    if controlObject:
        addSection( controlObject, atName = 'Shaders' )

    for shader in allShaders:
        
        if shader == 'lambert1':
            continue
        
        skip = True
        objNodes = getAssignedObjects ( shader )
        
        for objNode in objNodes:
            if objNode in materialObjects:
                skip = False
                break

        shaderName = shader
        shaderSet = mc.listConnections ( shader + '.outColor', type = 'objectSet' )[0]

        # skippig of prefixes
        if prefixMergeList:

            for prefixMerge in prefixMergeList:
                if prefixMerge in shader:
                    shaderName = shader.split ( prefixMerge )[1]
                    layeredShaderSet = shaderName + 'LayerShaderSG'

        # layeredShader creation
        if not skip:

            materialAttrExistance = False
            if controlObject:
                materialAttrExistance = bool ( mc.ls ( controlObject + '.' + shaderName ) )

            if not materialAttrExistance:

                if not shareOneChecker:
                    checkerShader, checkerSet = _createCheckerMaterial ( shaderName )
                    checkerShaders.append ( checkerShader )

                if controlObject:
                    mc.addAttr( controlObject, ln = shaderName, at = 'float', min = 0, max = 1, k = True, dv = 1 )
                layeredShader = mc.shadingNode( 'layeredShader', asShader = 1, name = shaderName + 'LayerShader' )
                layeredShaders.append ( layeredShader )
                mc.setAttr( layeredShader + '.compositingFlag', 1 )
                
                layeredShaderSet = mc.sets( renderable = True, empty = 1, noSurfaceShader = True, name = layeredShader + 'SG' )
                mc.connectAttr( layeredShader + '.outColor', layeredShaderSet + '.surfaceShader', f = 1 )
                mc.connectAttr( checkerShader + '.outColor', layeredShader + '.inputs[0].color', f = 1 )

                mc.setAttr( layeredShader + '.inputs[0].transparency', 0, 0, 0, type = 'double3' )
                if controlObject:
                    for col in ['R', 'G', 'B']:
                        mc.connectAttr( controlObject + '.' + shader, layeredShader + '.inputs[0].transparency' + col )

                mc.connectAttr( shader + '.outColor', layeredShader + '.inputs[1].color', f = 1 )
                mc.connectAttr( shader + '.outTransparency', layeredShader + '.inputs[1].transparency', f = 1 )
            

            if applyShaders:

                # assigning objects
                mc.hyperShade ( objects = shaderSet )
                mc.sets( forceElement = layeredShaderSet, e = 1 )
                mc.select ( cl = True )


    return {'layeredShaders': layeredShaders,
            'checkerShaders': checkerShaders
            }


def _createCheckerMaterial ( prefix ):

    shaderPrefix = prefix + 'Checker'

    checkerShader = mc.shadingNode( 'lambert', asShader = 1, name = shaderPrefix + 'Shader' )
    checkerSet = mc.sets( renderable = True, empty = 1, noSurfaceShader = True, name = shaderPrefix + 'SG' )
    mc.connectAttr( checkerShader + '.outColor', ( checkerSet + '.surfaceShader' ), f = 1 )

    checkerTexture = mc.shadingNode( 'checker', asTexture = 1, name = shaderPrefix + 'Node' )
    checkerPlacement = mc.shadingNode( 'place2dTexture', asUtility = 1 , name = shaderPrefix + '2DPlacement' )
    mc.connectAttr( checkerPlacement + '.outUV', checkerTexture + '.uv' )
    mc.connectAttr( checkerPlacement + '.outUvFilterSize', checkerTexture + '.uvFilterSize' )
    mc.connectAttr( checkerTexture + '.outColor', checkerShader + '.color', force = 1 )

    darkerCol = 0.3
    lighterCol = 0.6
    mc.setAttr( checkerTexture + '.color2', darkerCol, darkerCol, darkerCol, type = 'double3' )
    mc.setAttr( checkerTexture + '.color1', lighterCol, lighterCol, lighterCol, type = 'double3' )
    mc.setAttr( checkerPlacement + '.repeatU', 8 )
    mc.setAttr( checkerPlacement + '.repeatV', 8 )

    return [checkerShader, checkerSet]


def removeSuffix( name ):
    
    name = short( name )
    
    if name.count( '_' ) == 0: return name
    
    if name.count( '_' ) == 1:
        if name.startswith( 'l_' ) or name.startswith( 'r_' ):
            return name
            
    edit = name.split( '_' )[-1]
    edit = name[0: len( name ) - len( edit ) - 1 ]
    
    return edit

def short( name ):
  
  return name.split( '|' )[-1]

def addSection( object, atName = 'settings' ):
    
    mc.addAttr( object, ln = atName, at = 'enum', enumName = '__________', k = 0 )
    mc.setAttr( object + '.' + atName, cb = 1, l = 1 )
    
    
def capitalizeFirst( name ):
    
    '''
    capitalize first letter of given string
    
    @param name: string to process
    @type name: str
    @return: str
    '''
    
    firstletter = name[0].upper()
    suffix = ''
    if len( name ) > 1: suffix = name[1:]
    
    resultstring = firstletter + suffix
    
    return resultstring

    
