"""
skeletonBakeTool.py

made by: www.rig-it.net, (c)2017

How to use:
copy this file to your user folder:
/ C:/< User >/maya/scripts/skeletonBakeTool.py

Then restart Maya and in python script editor type:

import skeletonBakeTool
skeletonBakeTool.makeUI()

Then select something in the rig and push the button.

Version: 1.0.3
First Version Date: 31.7.2017
Update Date: 31.7.2017
Update Date: 8.8.2017 - setting current startFrame to nucleus node
Update Date: 9.8.2017 - adding button for nucleus start frame update, update for simulation
Update date: 10.8.2017 - adding condition if there is nucleus for rig in the scene

"""
from __future__ import print_function

# Bake Animation Tool
import maya.cmds as mc


def makeUI():
    
    windowB = 'myWindow'
            
    if mc.window( windowB, exists = True ):
        mc.deleteUI( windowB )
    else:
        mc.window( windowB, t = 'RIG-IT Skeleton Bake Tool', s = False, wh = ( 200, 150 ), rtf = True )
        mc.rowColumnLayout( nc = 3, columnWidth = [( 1, 50 ), ( 2, 300 ), ( 3, 50 )] )
        mc.separator( h = 10, style = 'none' )
        mc.text( l = 'Baking can not be undone.\n Save the scene before starting\n this action.' )
        
        mc.separator( h = 70, style = 'none' )
        mc.separator( w = 35, style = 'none' )
        mc.button( l = 'Update rig for simulation', c = 'skeletonBakeTool.bakeAnim(updateSceneOnly=True)', w = 50 )
        
        mc.separator( h = 10, style = 'none' )
        mc.separator( style = 'none' )
        mc.separator( h = 15, style = 'none' )
        mc.separator( style = 'none' )
        mc.separator( w = 35, style = 'none' )
        mc.button( l = 'Bake!', c = 'skeletonBakeTool.bakeAnim()', w = 30 )
        
        mc.showWindow()
    

def bakeAnim( updateSceneOnly = False ):

    sel = mc.ls( sl = True, l = True )
    
    if not sel:
        mc.confirmDialog( m = 'Nothing selected - you need to select some object\nfrom this rig for baking the skeleton' )
        return
    
    # import all references in scene
    importAllReferencesInScene()
    
    # get top group
    topGrp = sel[0].split( '|' )[1]
    
    # get namespace
    ns = getNamespace( topGrp )
    nsColon = namespaceAddColon( ns )
    
    # find assetModel_grp
    allRigKids = mc.listRelatives( topGrp, c = True, type = 'transform' )
    assetModelG = None
    
    for k in allRigKids:
        if mc.objExists( k + '.rigitModelGrp'):
            assetModelG = k
    
    if not assetModelG:
        mc.confirmDialog( m = 'Asset model group not detected in scene - probably missing .rigitModelGrp attribute' )
    
    if assetModelG:
        mc.parent( assetModelG, w = True ) 
    
    # get all root joints
    allJoints = mc.listRelatives( topGrp, ad = True, type = 'joint' )
    rootJnt = None
    
    for j in allJoints:
        labelName = mc.getAttr( j + '.otherType' )
        if labelName == 'root1':
            rootJnt = j
    
    allRootJoints = mc.listRelatives( rootJnt, ad = True, type = 'joint' )
    allRootJoints.append( rootJnt )
    allRootJoints.reverse()
    
    # Time Range
    min = mc.playbackOptions( q = True, min = True )
    max = mc.playbackOptions( q = True, max = True )
    
    # adjust nucleus node start frame    
    nucleusNode = nsColon + 'nucleus1'
        
    if mc.objExists( nucleusNode ):    
        
        mc.setAttr( nucleusNode + '.startFrame', min )
    
    if updateSceneOnly:
        
        if mc.objExists( nucleusNode ):
            
            mc.confirmDialog( m = 'Rig updated for simulation.' )
        
        else:
            
            mc.confirmDialog( m = 'Nothing had to be updated, there is no simulation node in rig' )
        
        return
    
    # Baking the animation for all objects
    mc.bakeResults( allRootJoints, simulation = True, t = ( min, max ), sampleBy = 1, oversamplingRate = 1, disableImplicitControl = True, preserveOutsideKeys = True, sparseAnimCurveBake = False, removeBakedAttributeFromLayer = False, removeBakedAnimFromLayer = False, bakeOnOverrideLayer = False, minimizeRotation = True, controlPoints = False, shape = True )
    
        
    # Filter curve
    for j in allRootJoints:
        
        mc.filterCurve( j + ".translateX" )
        mc.filterCurve( j + ".translateY" )
        mc.filterCurve( j + ".translateZ" )
        mc.filterCurve( j + ".rotateX" )
        mc.filterCurve( j + ".rotateY" )
        mc.filterCurve( j + ".rotateZ" )
    
    # remove all setup objects and keep model and skeleton
    
    # unlock all rig objects
    allRigObjs = mc.listRelatives( topGrp, ad = True )
    for o in allRigObjs:
        mc.lockNode( o, l = False )
    
    # all constraints
    allCons = mc.listRelatives( topGrp, ad = True, typ = 'constraint' )
    mc.delete( allCons )
    allEffs = mc.listRelatives( topGrp, ad = True, typ = 'ikEffector' )
    mc.delete( allEffs )
    
    # parent of root1_jnt
    rootParent = mc.listRelatives( rootJnt, p = True )
    mc.parent( rootParent, w = True )
    
    # deleting kids of topGrp
    allRigKidsToDelete = mc.listRelatives( topGrp, c = True, type = 'transform' )
    mc.delete( allRigKidsToDelete )
    
    # parenting model and joints to topGrp
    mc.parent( rootParent, topGrp )
    mc.parent( assetModelG, topGrp )
    
    # make assetModel_group selectable and visible
    mc.setAttr( assetModelG + '.overrideEnabled', 0 )    
    mc.setAttr( rootParent[0] + '.overrideEnabled', 0 )
    mc.showHidden( rootParent )
    

def importAllReferencesInScene( verbose = True, removeNamespace = False ):

    '''
    this will find all references in scene and import them,
    basically baking any existing

    @param verbose: print important warning and notes during function runtime
    @param verbose: str
    @param removeNamespace: optional, remove namespace from objects name
    @type removeNamespace: bool
    '''
    
    allRefNodes = mc.ls( type = 'reference' )
    
    for refNode in allRefNodes:
        
        try:

            namespaceName = mc.referenceQuery( refNode, namespace = 1 )
            fileName = mc.referenceQuery( refNode, filename = 1 )

        except:

            continue

        importReferenceToScene( namespaceName = namespaceName, removeNamespace = removeNamespace, verbose = verbose )    
    

def importReferenceToScene( namespaceName = '', removeNamespace = False, verbose = True ):

    '''
    import reference to scene

    @param namespaceName: remove reference based on given namespace
    @type namespaceName: str
    @param removeNamespace: optional, remove namespace from objects name
    @type removeNamespace: bool
    @param verbose: print important warning and notes during function runtime
    @param verbose: str
    @return: None
    '''

    if not namespaceName:

        return

    namespaceNameColon = namespaceAddColon( namespaceName )
    refObjects = mc.ls( namespaceNameColon + '*', o = 1 )

    if not refObjects:

        mc.namespace( rm = namespaceName )


    refFilename = mc.referenceQuery( refObjects[0], f = 1 )
    refNode = mc.referenceQuery( refFilename, referenceNode = 1 )
    mc.file( referenceNode = refNode, importReference = 1 )

    # remove namespace from object names

    if removeNamespace:

        namespaceList = mc.ls( namespaceNameColon + '*', objectsOnly = True, int = True )

        for node in namespaceList:

            mc.lockNode( node, l = 0 )

        mc.namespace( mv = [namespaceName, ':'], force = True )
        mc.namespace( rm = namespaceName )

    if verbose:

        print('# imported reference with namespace %s from file %s' % ( namespaceName, refFilename ))

def namespaceAddColon( namespace ):
    
    if namespace is not '' and not namespace.endswith( ':' ):
        
        namespace = namespace + ':'
    
    return namespace


def getNamespace( object ):
    
    parts = object.split( ':' )
    if len( parts ) < 2: return ''
    
    ns = ':'.join( parts[ 0:-1 ] )
    
    if ns == ':': ns = ''
    
    return ns


