'''
control module at rigit.rigbase
@category Rigging @subcategory Rigbase
@tags control class
@author: jakub
'''

import maya.cmds as mc
import maya.mel as mm

from . import controlShapes
from . import name
from . import connect
from . import shape
from . import joint
from . import anim

mainCtrlAt = 'rigitCtrl'
internalCtrlAt = 'rigitInternalCtrl'
controlSuffix = 'ctl'
controlInternalSuffix = 'icl'


class Control():
    
    def __init__( self,
                  control = '',
                  offset = '',
                  prefix = 'new',
                  shape = 'cube',
                  scale = 1.0,
                  colorIdx = -1,
                  colorName = '',
                  lockHide = [],
                  lockHideDef = ['s', 'v'],
                  rotOrd = 3,
                  moveTo = '',
                  translateTo = '',
                  translatePos = None,
                  rotateTo = '',
                  rotation = None,
                  axisRemap = 'xyz',
                  ctrlParent = '',
                  noOffset = False,
                  shapeMoveTo = '',
                  msgObj = None,
                  animated = True,
                  useJointFreezeRotate = False
                  ):
        
        localargs = locals()
        
        '''
        Class for building rig control
        
        NOTE: useJointFreezeRotate - useful for custom rotations on Body control, where Translate is in world orientation
        
        @param control: str, control transform name to initialize Control using existing object
        @param offset: str, control offset transform name to initialize Control using existing object
        @param prefix: str, name of control, including side, e.g. 'l_upperArmFk'
        @param shape: str, shape of control curve, valid values: ''(empty str), circleX (default), circleY, circleZ, cube, none (no shape)
        @param scale: float, scale of control
        @param colorIdx: int, color index of control
        @param colorName: str, name of color, valid values are: red, blue, yellow
        @param lockHide: list( str ), attributes to lock and hide (added to lockHideDef list), it`s ok to pass 't' for all translate or just 'tx' for single channel
        @param lockHideDef: list( str ), attributes to lock and hide by default, this is set to scale and visibility
        @param rotOrd: int, rotation order of control
        @param moveTo: str, match translation and rotation of this object
        @param translateTo: str, match translation of this object
        @param translatePos: list 3(float), world position to move control to, if not used it should be False
        @param rotateTo: str, match rotation of this object
        @param rotation: list 3(float), rotate control`s offset group (rotate order will be default XYZ)
        @param axisRemap: str, additional change of control offset group rotation to remap its axis. This has no effect on control shape (orientation)
        @param ctrlParent: str, parent control offset group under this object
        @param noOffset: bool, dont build offset group for this control if True
        @param shapeMoveTo: str, generate control shape by matching control shape transform to provided object
        @param msgObj: str, name of rigbase.Module Main group, so controls can be later found from module connection
        @param animated: bool, if this flag is false, control will be made for non-animation purposes (e.g. muscles) with "_icl" suffix (internal control)
        @param useJointFreezeRotate: bool, for custom rotations use joint transform and freeze rotations, resulting with Translate in world space
        @return: control class instance
        '''
        
        # ---------------------
        # define instance data
        # ---------------------
        
        self.C = None  # control transform
        self.Off = None  # control offset tranform
        
        #=======================================================================
        # instanciate using existing control
        #=======================================================================
        
        controlInitRes = self._instanciateUsingExistingObjects( control, offset, msgObj )
        
        if controlInitRes == 1 or controlInitRes == 2:
            
            return
        
        
        #=======================================================================
        # or build new control
        #=======================================================================
        
        # pass on function`s parameters to _build()
        self._build( localargs )
        
        
    def _build( self, args ):
              
        '''
        make new control
        using values from __init__()
        '''
        
        # args keys below have to match all __init__() function parameters of Control class
        
        control = args['control']
        offset = args['offset']
        prefix = args['prefix']
        shape = args['shape']
        scale = args['scale']
        colorIdx = args['colorIdx']
        colorName = args['colorName']
        lockHide = args['lockHide']
        lockHideDef = args['lockHideDef']
        rotOrd = args['rotOrd']
        moveTo = args['moveTo']
        translateTo = args['translateTo']
        translatePos = args['translatePos']
        rotateTo = args['rotateTo']
        rotation = args['rotation']
        axisRemap = args['axisRemap']
        ctrlParent = args['ctrlParent']
        noOffset = args['noOffset']
        shapeMoveTo = args['shapeMoveTo']
        msgObj = args['msgObj']
        animated = args['animated']
        useJointFreezeRotate = args['useJointFreezeRotate']
        
        # get color
        
        side = name.getSide( prefix )
        colorIdx = self._makeColor( side, colorIdx, colorName )
        
        # get names
        
        ctrlName, offsetName, ctrlOriginalName = self._ctrlNames( prefix, animated )
        
        # check if control name is unique
        
        self._checkControlNameUnique( ctrlName )
        
        # make control
        
        control = self._makeControlShape( ctrlName, side, shape, scale, colorIdx )
        self._makeControlAttributes( control, rotOrd, lockHide, lockHideDef, animated )
        
        # update instance data
        self.C = control
        
        # parent and transform control
        
        offset = self._makeControlOffsetGrp( control, offsetName, ctrlParent, noOffset )
        self._parentControl( control, offset, ctrlParent, noOffset )

        # update instance data
        self.Off = offset
        
        # original control group creation ( for connectAttr values)
        self.Co = self._createOriginalControlGrp ( ctrlOriginalName, axisRemap )

        # match offset group
        self._matchControlTransform( translateTo, rotateTo, moveTo, translatePos, rotation, axisRemap, shapeMoveTo, useJointFreezeRotate )
        
        # finish setting for control original group
        self._createOriginalControlSystem ( ctrlOriginalName, axisRemap )

        # connect control with message object
        self.connectControlToMsgObject( self.C, msgObj )
        
        return { 'control':control, 'offset':offset }
    
    def _createOriginalControlGrp ( self, ctrlOriginalName, axisRemap ):

        if axisRemap == 'xyz':
            return None

        ctrlOriginal = mc.group( em = 1, n = ctrlOriginalName, p = self.Off )

        return ctrlOriginal
        

    def _createOriginalControlSystem ( self, ctrlOriginalName, axisRemap ):

        if axisRemap == 'xyz':
            return None

        # offset group creation
        offsetName = name.removeSuffix ( ctrlOriginalName )
        coOffset = mc.group( em = 1, n = offsetName + 'Off_grp', p = self.Co )
        mc.parent ( coOffset, self.Off )
        tempConstr = mc.orientConstraint ( self.Co, coOffset )  # to keep axis when ro is changed

        # original control for connectAttr
        axis, sign = parceAxisString ( axisRemap )
        newAxisOrder = getRemappedRotationOrder ( axis )
        anim.setDrivenKey( self.C + '.ro', coOffset + '.ro', range ( 6 ), newAxisOrder )
        mc.delete ( tempConstr )
        mc.parent ( self.Co, coOffset )
        mc.connectAttr ( coOffset + '.ro', self.Co + '.ro' )

        for axisO, axisD in zip ( axis, ['x', 'y', 'z'] ):
            mc.connectAttr ( self.C + '.s' + axisO, self.Co + '.s' + axisD )

        mc.parentConstraint ( self.C, self.Co, mo = True ) 


    def _instanciateUsingExistingObjects( self, control, offset, msgObj ):
        
        '''
        function to give warning on invalid attempt to initialize Control
        using existing objects
        '''
        
        ctrlExists = False
        offsetExists = False
        result = 0
        
        if mc.objExists( control ) and mc.nodeType( control ) == 'transform':
            
            self.C = control
            ctrlExists = True
            result = 1
                    
        if mc.objExists( offset ) and mc.nodeType( offset ) == 'transform':
            
            self.Off = offset
            offsetExists = True
        
        if control and not ctrlExists:
            
            mc.warning( 'invalid initialization of control named %s' % control )
            result = 2
            
        if offset and not offsetExists:
            
            mc.warning( 'invalid initialization of control offset named %s' % offset )
            result = 2
        
        # connect control via message
        
        if result == 1:
            
            self.connectControlToMsgObject( control, msgObj )
        
        return result
    
    def _checkControlNameUnique( self, ctrlName ):
        
        matches = mc.ls( ctrlName )
        if matches:
            
            raise Exception( '# control name is not unique: %s , control will not be made' % ctrlName )
    
    
    def connectControlToMsgObject( self, controlObj, msgObj ):
        
        if msgObj and mc.objExists( msgObj ):
            
            connect.message( controlObj, msgObj, 'msg', 'controls' )
    
    
    def _makeColor( self, side, colorIdx, colorName ):
    
        defaultColorIdx = 22  # default yellow
        
        # get color from name

        if colorName == 'default':
            colorIdx = -1
        if colorName == 'red':
            colorIdx = 13
        if colorName == 'blue':
            colorIdx = 6
        if colorName == 'lightblue':
            colorIdx = 18
        if colorName == 'yellow':
            colorIdx = 22
        if colorName == 'green':
            colorIdx = 26
        if colorName == 'darkgreen':
            colorIdx = 7
        if colorName == 'orange':
            colorIdx = 31
        if colorName == 'lightorange':
            colorIdx = 21
        if colorName == 'pink':
            colorIdx = 20
        if colorName == 'purple':
            colorIdx = 9
        
        
        if colorName == 'secondary' and side == 'l':
            colorIdx = 18
        if colorName == 'secondary' and side == 'r':
            colorIdx = 20
        if colorName == 'secondary' and side != 'l' and side != 'r':
            colorIdx = 26
            
        # if colorIdx is still not valid, get color from side or default color
        
        if colorIdx < 0 or colorIdx > 32:
            
            if side.lower() == 'l':
              colorIdx = 6
            elif side.lower() == 'r':
              colorIdx = 13
            else:
              colorIdx = defaultColorIdx
          
        return colorIdx
    
    def _ctrlNames( self, prefix, animated ):
        
        '''
        Setting control naming convention
        
        DO NOT CHANGE THIS FUNCTION,
        unless agreed with rigging supervisor, as animated shots depend on this convention
        '''
        
        ctrlName = prefix + '_ctl'
        offsetName = prefix + 'CtlOff_grp'
        ctrlOriginalName = prefix + 'CtlOrig_grp'
        
        if not animated:
            
            ctrlName = prefix + '_icl'
            offsetName = prefix + 'IclOff_grp'
            ctrlOriginalName = prefix + 'IclOrig_grp'
        
        return [ ctrlName, offsetName, ctrlOriginalName ]
    
    def _makeControlShape( self, ctrlName, side, shape, scale, colorIdx, blockSquash = 0.25, pillarStretch = 2.0 ):
    
        control = ''
        
        # ========================================================
        # create control shape
        # ========================================================
        
        if shape == '' or shape == 'circle': control = controlShapes.circle()
        elif shape == 'circleX': control = controlShapes.circle( normal = 'x' )
        elif shape == 'circleY': control = controlShapes.circle( normal = 'y' )
        elif shape == 'circleZ': control = controlShapes.circle( normal = 'z' )
        elif shape == 'circleXlin': control = controlShapes.circle( normal = 'x', degree = 1 )
        elif shape == 'circleYlin': control = controlShapes.circle( normal = 'y', degree = 1 )
        elif shape == 'circleZlin': control = controlShapes.circle( normal = 'z', degree = 1 )
        elif shape == 'square': control = controlShapes.square()
        elif shape == 'squareX': control = controlShapes.square( normal = [1, 0, 0] )
        elif shape == 'squareY': control = controlShapes.square()
        elif shape == 'squareZ': control = controlShapes.square( normal = [0, 0, 1] )
        elif shape == 'cube': control = controlShapes.cube()
        elif shape == 'sphere': control = controlShapes.sphere()
        elif shape == 'locator': control = controlShapes.locator()
        elif shape == 'blockX': control = controlShapes.cube( scale = [ blockSquash, 1, 1 ] )
        elif shape == 'blockY': control = controlShapes.cube( scale = [ 1, blockSquash, 1 ] )
        elif shape == 'blockZ': control = controlShapes.cube( scale = [ 1, 1, blockSquash ] )
        elif shape == 'pillarX': control = controlShapes.cube( scale = [ pillarStretch, 1, 1 ] )
        elif shape == 'pillarY': control = controlShapes.cube( scale = [ 1, pillarStretch, 1 ] )
        elif shape == 'pillarZ': control = controlShapes.cube( scale = [ 1, 1, pillarStretch ] )
        elif shape == 'rtriangle': control = controlShapes.rtriangle()
        elif shape == 'flag': control = controlShapes.triangle( scale = [1, 4, 1] )
        elif shape == 'c': control = controlShapes.c()
        elif shape == 'd': control = controlShapes.d()
        elif shape == 's': control = controlShapes.s()
        elif shape == 'v': control = controlShapes.v()
        elif shape == 't': control = controlShapes.t()
        elif shape == 'orient': control = controlShapes.orient()
        elif shape == 'aimbase': control = controlShapes.aimbase()
        elif shape == 'arrow': control = controlShapes.arrow()
        elif shape == 'key': control = controlShapes.key()
        elif shape == 'keyX': control = controlShapes.keyX()
        elif shape == 'keyY': control = controlShapes.keyY()
        elif shape == 'keyZ': control = controlShapes.key()
        elif shape == 'starSimple': control = controlShapes.starSimple()
        elif shape == 'cross': control = controlShapes.cross()
        elif shape == 'line': control = controlShapes.line()
        elif shape == 'lineX': control = controlShapes.line( normal = [1, 0, 0] )
        elif shape == 'lineY': control = controlShapes.line( normal = [0, 1, 0] )
        elif shape == 'lineZ': control = controlShapes.line( normal = [0, 0, 1] )
        elif shape == 'eye': control = controlShapes.eye()
        elif shape == 'eyeX': control = controlShapes.eye()
        elif shape == 'eyeY': control = controlShapes.eyeY()
        elif shape == 'eyeZ': control = controlShapes.eyeZ()
        elif shape == 'none': control = mc.group( em = 1 )
        
        else: raise( Exception( 'unregistered shape name "%s"' % shape ) )
        
        # ========================================================
        # adjust control`s transform
        # ========================================================
                
        control = mc.rename( control, ctrlName )
        
        # adjust side orientation for some shapes for right side        
        if side == 'r' and shape in ['rtriangle', 'flag', 'orient', 'arrow']:
            
            scale = scale * ( -1 )
        
        mc.setAttr( control + '.s', scale, scale, scale )
        mc.makeIdentity( control, a = 1, s = 1 )
        
        for s in mc.listRelatives( control, s = 1 ):
            
            mc.setAttr( s + '.ove', 1 )
            mc.setAttr( s + '.ovc', colorIdx )
        
        # adjust pillarX shape
        # WIP ... needs to move all CVs by half the length down the joint (+ for left, - for right)
      
        return control
    
    def _lockHideRotateTest( self, lockHide, lockHideDef ):
        
        if 'r' in  lockHide: return 1
        if 'r' in  lockHideDef: return 1
        
        axesLockNum = 0
        axesLockDefNum = 0
        
        for x in ['x', 'y', 'z']:
            
            if ( 'r' + x ) in  lockHide: axesLockNum += 1
            if ( 'r' + x ) in  lockHideDef: axesLockDefNum += 1
        
        if axesLockNum == 2 or axesLockDefNum == 2: return 1
        
        return 0
        
    
    def _makeControlAttributes( self, control, rotOrd, lockHide, lockHideDef, animated ):
      
        # add control tags to find them easily in scene
        
        if animated:
            
            mc.addAttr( control, ln = mainCtrlAt, dt = 'string' )
        
        else:
            
            mc.addAttr( control, ln = internalCtrlAt, dt = 'string' )
        
        
        # expose rotate order if rotations will be keyable
        
        if self._lockHideRotateTest( lockHide, lockHideDef ) == 0:
            
            mc.setAttr( control + '.ro', rotOrd, k = 1 )
            
        lockHideEdit = []

        for a in lockHide + lockHideDef:
            
            if a == 't' or a == 'r' or a == 's':
                for x in ['x', 'y', 'z']:
                    lockHideEdit.append( a + x )
            else:
                lockHideEdit.append( a )
                
        
        for a in lockHideEdit:
            mc.setAttr( control + '.' + a, l = 1, k = 0, cb = 0 )
    
    
    def _makeControlOffsetGrp( self, control, offsetName, ctrlParent, noOffset ):
        
        if noOffset: return None
        
        offset = mc.group( em = 1, n = offsetName )
        mc.addAttr( offset, ln = 'rigControlOffset', at = 'bool' )
        mc.parent( control, offset )
        
        return offset
    
    def _parentControl( self, control, offset, ctrlParent, noOffset ):
        
        '''
        parent control or control offset under specified parent
        '''
        
        if not ctrlParent or not mc.objExists( ctrlParent ): return
        
        if noOffset: mc.parent( control, ctrlParent )
        else: mc.parent( offset, ctrlParent )
    
    def _matchControlTransform( self, translateTo, rotateTo, moveTo, translatePos, rotation, axisRemap, shapeMoveTo, useJointFreezeRotate ):
        
        # move control (only supported with offset group)
        
        if not self.Off or not mc.objExists( self.Off ):
            
            return
        
        if translateTo and mc.objExists( translateTo ): 
                    
            mc.delete( mc.pointConstraint( translateTo, self.Off ) )
        
        if rotateTo and mc.objExists( rotateTo ):
            
            mc.delete( mc.orientConstraint( rotateTo, self.Off ) )
        
        if moveTo and mc.objExists( moveTo ):
            
            mc.delete( mc.parentConstraint( moveTo, self.Off ) )
        
        if translatePos:
            
            mc.move( translatePos[0], translatePos[1], translatePos[2], self.Off, a = 1, ws = 1 )
        
        if rotation:
            
            mc.rotate( rotation[0], rotation[1], rotation[2], self.Off, a = 1, ws = 1 )
        
        if axisRemap and axisRemap != 'xyz':
            
            mc.parent ( self.Co, w = True )
            axis, sign = parceAxisString ( axisRemap )
            axisTransform = [sign[0] * ( axis[0] == 'x' ), sign[1] * ( axis[1] == 'x' ), sign[2] * ( axis[2] == 'x' ), 0,
                             sign[0] * ( axis[0] == 'y' ), sign[1] * ( axis[1] == 'y' ), sign[2] * ( axis[2] == 'y' ), 0,
                             sign[0] * ( axis[0] == 'z' ), sign[1] * ( axis[1] == 'z' ), sign[2] * ( axis[2] == 'z' ), 0,
                              0, 0, 0, 1 ]

            offCtlTemp = mc.duplicate( self.Off, n = 'transformShape_grp', parentOnly = True )[0]
            shape.copyAtoB ( self.C, offCtlTemp )

            offGrpTemp = mc.duplicate( self.Off, n = 'transformTemp_grp', parentOnly = True )[0]
            mc.setAttr ( offGrpTemp + '.rotateOrder', 0 )
            mc.parent ( offGrpTemp, self.Off )
            mc.xform ( offGrpTemp, matrix = axisTransform )

            mc.parent ( offGrpTemp, w = True )
            mc.delete( mc.parentConstraint( offGrpTemp, self.Off ) )
            shape.copyAtoB ( offCtlTemp, self.C )

            mc.delete ( offCtlTemp, offGrpTemp )

            mc.parent ( self.Co, self.Off )
                      
        if shapeMoveTo:
            
            offGrpShapeObj = mc.duplicate( self.Off, n = 'shapeMoveToControlObj', renameChildren = True )[0]
            mc.delete( mc.parentConstraint( shapeMoveTo, offGrpShapeObj ) )
            ctrlShapeObj = mc.listRelatives( offGrpShapeObj, c = True, type = 'transform' )[0]
            shape.copyAtoB( ctrlShapeObj, self.C )
            mc.delete( offGrpShapeObj )
        
        if useJointFreezeRotate:
            
            ctrlJoint = joint.makeSingleJoint( prefix = 'newCtrl_jnt', refObject = self.Off, freezeRotation = True )
            mc.setAttr( ctrlJoint + '.drawStyle', 2 )  # hide joint icon
            mc.setAttr( ctrlJoint + '.radius', l = True, k = False, cb = False )  # lock & hide joint radius attribute
            shape.copyAtoB( self.C, ctrlJoint )
            
            self.matchControlAttributes( self.C, ctrlJoint )
            mc.delete( self.C )  # get rid of control transform, we dont need it anymore
            ctrlJoint = mc.rename( ctrlJoint, self.C )  # match original control name on new joint transform
            
            mc.setAttr( self.Off + '.r', 0, 0, 0 )  # reset rotation on control offset group
            mc.parent( self.C, self.Off, a = True )
            
            self.C = ctrlJoint
        
        

    def matchControlAttributes( self, sourceObj, destObj ):
        
        """
        match control attributes on destination object
        This is used for replacing control transform with joint
        """
        
        # process transform channels
        for at in ['t', 'r', 's']:
            
            for axis in ['x', 'y', 'z']:
                
                atLockStat = mc.getAttr( sourceObj + '.' + at + axis, l = True )
                
                if atLockStat:
                    
                    mc.setAttr( destObj + '.' + at + axis, l = True, k = False, cb = False )
        
        # process visibility
        atLockStat = mc.getAttr( sourceObj + '.v', l = True )
        if atLockStat:
            
            mc.setAttr( destObj + '.v', l = True, k = False, cb = False )

        rotOrd = mc.getAttr( sourceObj + '.ro' )
        keyable = mc.getAttr( sourceObj + '.ro', k = True )
        mc.setAttr( destObj + '.ro', rotOrd, k = keyable )

    def getOrigControl( self ):
        
        """
        To get control transform node which can be directly connected
        with its tranforms despite it has axisRemap or not.
        """
        if self.Co:
            return self.Co
        
        return self.C

  

def getAllControlsUnderGroup( grp ):
    
    """
    get all RIGIT controls under given group
    
    @param grp: str, name of group to searching controls in its hierarchy
    @return: list(str), list of controls found under group, this is string names, not Control class instances
    """
    
    objsUnderGroup = mc.listRelatives( grp, ad = True, type = 'transform' )
    controlsUnderGroup = []
    
    for o in objsUnderGroup:
        
        if mc.objExists( o + '.' + mainCtrlAt ):
            
            controlsUnderGroup.append( o )
    
    return controlsUnderGroup

def parceAxisString ( axisString ):

    """
    For parcing string and its signs.
    Example: For 'yzx' it returns (['y', 'z', 'x'], [1, 1, 1]). For '+yz-x' it return (['y', 'z', 'x'], [1, 1, -1])
    """

    axis = []
    sign = []
    prevChar = '+'
    for char in axisString:
        if char in ['x', 'y', 'z']:
            axis.append ( char )
            sign.append ( 1 - 2 * ( prevChar == '-' ) )
        prevChar = char

    return axis, sign


mayaRotOrder = ['xyz', 'yzx', 'zxy', 'xzy', 'yxz', 'zyx']

def getRemappedRotationOrder ( parcedAxisString ):

    """
    Returns which rotation order index should be mapped on which index of remapped control by parcedAxisString ( )
    """
    newOrder = []
    for order in mayaRotOrder:
        currentRemap = ['r', 'i', 'g']
        for id, axis in enumerate ( ['x', 'y', 'z'] ):
            pos = order.index( axis )
            currentRemap[pos] = parcedAxisString[id]
        
        newOrder.append ( mayaRotOrder.index( ''.join( currentRemap ) ) )
    
    return newOrder
    
