'''
rivet in rigit.tools
@category Rigging @subcategory Tools
@tags rivet button
@author jakub

functions for making rivet setup on meshes and nurbs
'''

import maya.cmds as mc

from . import name
from . import shape
from . import surface
from . import hair
from . import transform
from . import curve


# add function to automaticly find type of shape and closest edges (mesh) or point (nurbs)

def classicrivet( shapecomponents, object, prefix = '' ):
    
    '''
    NOTE:
    - shapecomponents edges example: ['pSphere1.e[237]', 'pSphere1.e[257]']
    - shapecomponents surface point example: ['nurbsSphere1.uv[2.6][0.5]']
    
    @param shapecomponents: list of 2 edge names or 1 surface point name
    @type shapecomponents: list( str )
    @param object: name of riveted object
    @type object: str
    @param prefix: name added to every new object
    @type prefix: str
    @return: list(str), list with various rivet object names
    '''
    
    # call rivet function based on type of surface
    
    shapetype = mc.nodeType( mc.ls( shapecomponents[0], o = 1 )[0] )
    rivetres = None
    
    # get transform object`s name
    
    shapeobj = mc.ls( shapecomponents[0], o = 1 )[0]
    
    objprefix = name.removeSuffix( object )
    if not prefix: prefix = objprefix    
    prefix = prefix + 'Rivet'
    
    
    # call rivet function based on shape
    
    if shapetype == 'mesh': rivetres = classicrivet_mesh( shapecomponents, shapeobj, object, prefix )
    if shapetype == 'nurbsSurface': rivetres = classicrivet_surface( shapecomponents[0], object, prefix )
    
    
    return rivetres


def classicrivet_mesh( edges, shapeobj, object, prefix, closestUvParam = False, usePinIfPossible = True ):
    
    '''
    make classic rivet using surface from 2 edges
    
    NOTE:
    - function returns list with first item being list of shape input plugs
      shape input plugs can be used to be later reconnected to something else to prevent cycle
    
    
    @param edges: list( str, str ), 2 edges, e.g. ['pSphere1.e[237]', 'pSphere1.e[257]']
    @param object: str, name of object to be attached to rivet transform output
    @param prefix: str, name added to every new object
    @param closestUvParam: bool, object attachment will be closer to its original location using closest coords on created nurbs plane
    @param usePinIfPossible: bool, if True, it uses uvPin instead of follicle solution for Maya 2020 build and newer
    @return: list( list(str), str, str, str ), ( (shapeinput1,..shapeinput4), aimConstraint, pointOnSurfaceInfo, pointConstraint )
    '''
    
    #===========================================================================
    # create nurbs loft between edges
    #===========================================================================
    
    edges = mc.ls( edges, fl = 1 )
    namePOSI = ''
    
    # get edge indeces
    
    e1 = edges[0].split( '.e' )[1]
    e1 = e1.replace( '[', '' )
    e1 = e1.replace( ']', '' )
    e1 = float( e1 )
    
    e2 = edges[1].split( '.e' )[1]
    e2 = e2.replace( '[', '' )
    e2 = e2.replace( ']', '' )
    e2 = float( e2 )
    
    # create curves from edge
    
    nameCFME1 = mc.createNode( 'curveFromMeshEdge', n = prefix + 'A_cme' )
    mc.setAttr( nameCFME1 + '.ihi', 1 )
    mc.setAttr( nameCFME1 + '.ei[0]', e1 )
    
    nameCFME2 = mc.createNode( 'curveFromMeshEdge', n = prefix + 'B_cme' )
    mc.setAttr( nameCFME2 + '.ihi', 1 )
    mc.setAttr( nameCFME2 + '.ei[0]', e2 )
    
    # create loft
    
    nameLoft = mc.createNode( 'loft', n = prefix + 'Loft_lft' )    
    mc.setAttr( nameLoft + '.ic', s = 2 )
    mc.setAttr( nameLoft + '.u', 1 )
    mc.setAttr( nameLoft + '.rsn', 1 )
    mc.setAttr( nameLoft + '.autoReverse', False )
    
    # make connections
    
    mc.connectAttr( nameCFME1 + '.oc', nameLoft + '.ic[0]' )
    mc.connectAttr( nameCFME2 + '.oc', nameLoft + '.ic[1]' )
    mc.connectAttr( shapeobj + '.w', nameCFME1 + '.im' )
    mc.connectAttr( shapeobj + '.w', nameCFME2 + '.im' )
    
    # make nurbs shape
    
    nurbsobj = mc.createNode( 'nurbsSurface' )
    nurbsxform = mc.listRelatives( nurbsobj, p = 1 )[0]
    nurbsxform = mc.rename( nurbsxform, prefix + 'Loft_sur' )
    nurbsobj = mc.listRelatives( nurbsxform, s = 1 )[0]
    mc.hide( nurbsxform )
    objectparent = mc.listRelatives( object, p = 1 )
    
    if objectparent:
        
        mc.parent( nurbsxform, objectparent[0] )
    
    mc.connectAttr( nameLoft + '.outputSurface', nurbsobj + '.create' )
    
    # check and fix loft curve direction
    # compare perimeter of loft after reversing one curve
    origPerimLen = _measureLoftPerimeter( nurbsxform )
    
    reverseCurveNode = mc.createNode( 'reverseCurve', n = prefix + 'Loft_rvc' )
    mc.connectAttr( nameCFME1 + '.oc', reverseCurveNode + '.inputCurve' )
    mc.connectAttr( reverseCurveNode + '.outputCurve', nameLoft + '.ic[0]', f = True )
    reversedPerimLen = _measureLoftPerimeter( nurbsxform )
    
    if origPerimLen < reversedPerimLen:
        
        mc.connectAttr( nameCFME1 + '.oc', nameLoft + '.ic[0]', f = True )
        mc.delete( reverseCurveNode )
    
    #===========================================================================
    # attach object transform to surface
    #===========================================================================
    
    params = [0.5, 0.5] 

    if closestUvParam:
        params = surface.getClosestPointOnSurface( object, nurbsobj, normalizeCoords = True )[0]

    if usePinIfPossible and float ( mc.about( v = True ) ) >= 2020:
        attachres = [createUvPin ( object, nurbsobj, prefix, params )]
    else:
        attachres = attachToSurface( object, surfaceObj = nurbsobj, prefix = prefix, params = params )
    
    # add shape inputs list to return
    
    shapeinputs = [ [ nameCFME1 + '.im', nameCFME2 + '.im' ] ]
    
    
    results = shapeinputs + attachres
    
    
    return results
    
    
def classicrivet_surface( surfacePoint, object, prefix ):
    
    '''
    make classic rivet using surface from 2 edges
    
    NOTE:
    - function returns list with first item being list of shape input plugs
      shape input plugs can be used to be later reconnected to something else to prevent cycle
    
    
    @param surfacePoint: name of nurbs surface point, e.g. nurbsSphere1.uv[2.6][0.5]
    @type edges: str
    @param object: name of riveted object
    @type object: str
    @param prefix: name added to every new object
    @type prefix: str
    @return: list( list(str), str, str, str ), ( (shapeinput1,..shapeinput4), aimConstraint, pointOnSurfaceInfo, pointConstraint )
    '''
    
    uvedits = surfacePoint.split( '[' )
    shapeobj = uvedits[0].split( '.' )[0]
    u = float( uvedits[1].replace( ']', '' ) )
    v = float( uvedits[2].replace( ']', '' ) )
    
    
    #===========================================================================
    # attach object transform to surface
    #===========================================================================
    
    attachres = attachToSurface( object, surfaceObj = shapeobj, prefix = prefix, params = [0.5, 0.5] )
    
    # add shape inputs list to return
    
    shapeinputs = [ [ attachres[0] + '.is' ] ]
    results = shapeinputs + attachres
    
    
    return results
    

def attachToSurface( object, surfaceObj, prefix = '', params = [0.5, 0.5] ):
    
    '''
    function used by rivet to attach object to nurbsSurface
    
    @param object: name of transform object to be attached
    @type object: str
    @param surfaceObj: name of nurbs shape (this can be loft node or anything with NURBS data output)
    @type surfaceObj: str
    @param prefix: prefix for naming new objects
    @type prefix: str
    @param params: list with UV parameters to position object on surface
    @type params: list( float, float )
    @param usePercentage: flag for using percentage for parameters on pointOnSurfaceInfo
    @type usePercentage: bool
    @return: list(str), list of objects names from attachment setup (pointOnSurfaceInfo node, aimConstraint node, pointConstraint node)
    '''
    
    # edit prefix
    
    if not prefix: prefix = name.removeSuffix( object ) + 'AttachToNurbs'
    
    # create surface follicle
    
    follicleXform, follicle = hair.makeFollicleOnSurface( 
                                                    surface = surfaceObj,
                                                    prefix = prefix,
                                                    uvCoordinates = params,
                                                    useCustomParent = True,
                                                    customParent = object
                                                    )
    
    return [ follicleXform, follicle ]

    
    
def attachToSurface_pointOnSurface( object, surfaceplug, prefix = '', params = [0.5, 0.5], usePercentage = 1 ):
    
    '''
    function used by rivet to attach object to nurbsSurface
    
    @param object: name of transform object to be attached
    @type object: str
    @param surfaceplug: name of ...
    @type surfaceplug: str
    @param prefix: prefix for naming new objects
    @type prefix: str
    @param params: list with UV parameters to position object on surface
    @type params: list( float, float )
    @param usePercentage: flag for using percentage for parameters on pointOnSurfaceInfo
    @type usePercentage: bool
    @return: list(str), list of objects names from attachment setup (pointOnSurfaceInfo node, aimConstraint node, pointConstraint node)
    '''
    
    # edit prefix
    
    if not prefix: prefix = name.removeSuffix( object ) + 'AttachToNurbs'
    
    # create point on surface info
    
    namePOSI = mc.createNode( 'pointOnSurfaceInfo', n = prefix + '_psi' )    
    mc.setAttr( namePOSI + '.turnOnPercentage', usePercentage )
    mc.setAttr( namePOSI + '.parameterU', params[0] )
    mc.setAttr( namePOSI + '.parameterV', params[1] )
    mc.connectAttr( surfaceplug, namePOSI + '.is' )
    
    
    # create aim constraint
    
    nameAC = mc.createNode( 'aimConstraint', p = object, n = prefix + '_aic' )
    mc.setAttr( nameAC + '.tg[0].tw', 1 )
    mc.setAttr( nameAC + '.a', 0, 1, 0 )
    mc.setAttr( nameAC + '.u', 0, 0, 1 )
    mc.setAttr( nameAC + '.v', k = 0 )
    
    for channel in ['t', 'r', 's']:
        for axis in ['x', 'y', 'z']:
            mc.setAttr( nameAC + '.' + channel + axis, k = 0 )
    
    # create normal and tangent vector products to rotate transform
    
    rivetNormalVectorProduct = mc.createNode( 'vectorProduct', n = prefix + 'Normal_vtp' )
    mc.setAttr( rivetNormalVectorProduct + '.operation', 3 )
    mc.connectAttr( namePOSI + '.n', rivetNormalVectorProduct + '.input1' )
    mc.connectAttr( object + '.parentInverseMatrix', rivetNormalVectorProduct + '.matrix' )
    mc.connectAttr( rivetNormalVectorProduct + '.output', nameAC + '.target[0].targetTranslate' )
        
    rivetTangentVectorProduct = mc.createNode( 'vectorProduct', n = prefix + 'Tangent_vtp' )
    mc.setAttr( rivetTangentVectorProduct + '.operation', 3 )
    mc.connectAttr( namePOSI + '.tv', rivetTangentVectorProduct + '.input1' )
    mc.connectAttr( object + '.parentInverseMatrix', rivetTangentVectorProduct + '.matrix' )
    mc.connectAttr( rivetTangentVectorProduct + '.output', nameAC + '.wu' )
    for axis in ['x', 'y', 'z']: mc.connectAttr( nameAC + '.cr' + axis, object + '.r' + axis )
    
    
    # create point constraint and position vector product to translate transform
    
    namePC = mc.createNode( 'pointConstraint', p = object, n = prefix + '_poc' )
    
    rivetPosVectorProduct = mc.createNode( 'vectorProduct', n = prefix + 'Pos_vtp' )
    mc.setAttr( rivetPosVectorProduct + '.operation', 4 )
    mc.connectAttr( namePOSI + '.position', rivetPosVectorProduct + '.input1' )
    mc.connectAttr( object + '.parentInverseMatrix', rivetPosVectorProduct + '.matrix' )
    mc.connectAttr( rivetPosVectorProduct + '.output', namePC + '.target[0].targetTranslate' )
    mc.connectAttr( namePC + '.constraintTranslate', object + '.t' )
    
    
    return [ namePOSI, nameAC, namePC ]
    
    

    
    
def _measureLoftPerimeter( loftSurface ):
    
    pos1 = mc.xform( loftSurface + '.cv[0][0]', q = 1, t = 1, ws = 1 )
    pos2 = mc.xform( loftSurface + '.cv[0][3]', q = 1, t = 1, ws = 1 )
    pos3 = mc.xform( loftSurface + '.cv[1][3]', q = 1, t = 1, ws = 1 )
    pos4 = mc.xform( loftSurface + '.cv[1][0]', q = 1, t = 1, ws = 1 )
    
    perimLength = transform.measureDistanceBetweenMultiplePositions( [pos1, pos2, pos3, pos4] )
    
    return perimLength
    
    
    
        
def createUvPin ( objectNode, shapeObj, prefix, params, normalizedIsoParms = True ):

    '''
    Function to attach object to nurbsSurface or mesh surface. Works in Maya 2020 or newer.
    
    @param object: str, name of transform object to be attached
    @param shapeObj: str, name of nurbs or poly shape
    @param prefix: str, prefix for naming new objects
    @param params: list( float, float ), list with UV parameters to position object on surface
    @param usePercentage: bool, flag for using percentage for parameters on pointOnSurfaceInfo
    @return: list(str), list of objects names from attachment setup (pointOnSurfaceInfo node, aimConstraint node, pointConstraint node)
    '''

    if float ( mc.about( v = True ) ) < 2020:
        raise Exception ( 'uvPin node is node supported in running Maya version' )

    origObjShape = shape.getOrigShape ( shapeObj )

    if not origObjShape:

        temp = mc.cluster ( shapeObj )
        mc.delete ( temp )
        origObjShape = shape.getOrigShape ( shapeObj )
   
    mc.setAttr( objectNode + '.inheritsTransform', 0 )
    mc.xform ( objectNode, ws = True, m = [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1] )


    pin = mc.createNode( 'uvPin', name = prefix + '_pin' )
    mc.setAttr ( pin + '.coordinate[0].coordinateU', params[0] )
    mc.setAttr ( pin + '.coordinate[0].coordinateV', params[1] )
    mc.setAttr ( pin + '.normalizedIsoParms', normalizedIsoParms )
    mc.connectAttr ( pin + '.outputMatrix[0]', objectNode + '.offsetParentMatrix' )

    shapetype = mc.nodeType( shapeObj )


    if shapetype == 'mesh':

        if not mc.polyEditUV( shapeObj + '.map[0]', query = True ):  # if obj has no UVs
            raise Exception ( 'geo %s misses uv map' % shapeObj )

        mc.connectAttr ( origObjShape + '.outMesh', pin + '.originalGeometry' )
        mc.connectAttr ( shapeObj + '.worldMesh[0]', pin + '.deformedGeometry' )

    if shapetype == 'nurbsSurface':

        mc.connectAttr ( origObjShape + '.local', pin + '.originalGeometry' )
        mc.connectAttr ( shapeObj + '.worldSpace[0]', pin + '.deformedGeometry' )


    return pin
