'''
export.py at softmodSculpt

functions for exporting and importing sculpt setup

Copyright: www.rig-it.net
'''
# from __future__ import print_function

import json
import os

import maya.cmds as mc

from . import setup
from ...utils import deformer
from ...utils import name

softmodfilename = 'softmodSculpt_export'
softmodfolder = 'softmodSculpt_export'
manimext = 'manim'
jsonext = 'json'
deformControlSuffixBase = 'Deform'


def exportSetup( filepath = '', animationpath = '', softmods = [], autoCreateFilePath = True, exportAnimation = True, verbose = True ):
    
    '''
    write all or selected Soft Mods and all its information (edges, geometries, parameters etc.)
    
    @param filepath: full path (not including extension) to the file to write the Soft Mod setup config from current Maya scene
    @type filepath: str
    @param animationpath: full path (not including extension) to the animation file for Soft Mods controls (will also write static values)
    @type animationpath: str
    @param softmods: list of Soft Mod main groups (where each contains 1 softMod system)
    @type softmods: list(str)
    @param autoCreateFilePath: create filepath to export softmods if filepath was not provided (use user job folder)
    @type autoCreateFilePath: bool
    @param exportAnimation: export animation file as well, if file not provided, save it to the same folder as filepath
    @type exportAnimation: bool
    @return: list(str), list containing result information from the tool: 0- full path of the export file, 1- full path of the animation file
    '''

    # note start of export

    if verbose:

        print('')
        print('# softmodSculpt Export - started export process')

    # find all softmods
    
    allSoftModMainGroups1 = mc.ls( '*.%s' % setup.mainGrpAt, o = 1 )
    allSoftModMainGroups2 = mc.ls( '*:*.%s' % setup.mainGrpAt, o = 1 )
    allSoftModMainGroups = allSoftModMainGroups1 + allSoftModMainGroups2
    
    infoData = []  # collect all info for each softmod here
    
    # get info on each (parameters of controls, geometry shapes) - will need its primary shape for edges indeces
    
    infoData = _exportControl_collectData( allSoftModMainGroups, infoData, verbose )
    
    # fix control names
    _fixControlNames( infoData, verbose )
    
    # write softmod file (json file) - add path of animation file to main soft mod file so it can be found automaticly
    # optionaly create folder in user directory if no valid filepath was provided
    
    filepath = _completeFixAndCreateSoftModFilePath( filepath, autoCreateFilePath, verbose, mode = 'export' )
    infoData = _exportSoftModSetups( filepath, infoData, verbose )
    
    # optional - write animation (PFFL anim)
    
    animationpath = _makeAnimationFilepath( filepath, animationpath )
    _exportAnimation( infoData, animationpath, exportAnimation, verbose )
    
    
    return [ filepath, animationpath ]


def importSetup( filepath = '', animationpath = '', softmods = [], autoCreateFilePath = True, importAnimation = True, verbose = True ):
    
    '''
    create softmodSculpt setups in Maya scene from exported file (made by exportSetup function)
    The script assumes that all geometry is existing already and has not softmodSculpts applied yet
    
    TO DO: skip existing softmodSculpts
    
    @param filepath: full path (file extension doesnt have to be provided) to the file to import the Soft Mod setup config to current Maya scene
    @type filepath: str
    @param animationpath: full path (file extension doesnt have to be provided) to the animation file for Soft Mods controls (will also write static values)
    @type animationpath: str
    @param softmods: list of Soft Mod main groups (where each contains 1 softMod system)
    @type softmods: list(str)
    @param autoCreateFilePath: create filepath to export softmods if filepath was not provided (use user job folder)
    @type autoCreateFilePath: bool
    @param exportAnimation: export animation file as well, if file not provided, save it to the same folder as filepath
    @type exportAnimation: bool
    @return: list(str), list containing result information from the tool: 0- full path of the export file, 1- full path of the animation file
    '''
    
    # store selection
    
    origsel = mc.ls( sl = 1 )
    
    # open softmod file (json file)
    
    filepath = _completeFixAndCreateSoftModFilePath( filepath, autoCreateFilePath, verbose, mode = 'import' )
    infoData = _readInfoDataFromFile( filepath )
    
    # create softmodScultp setups
    
    _createSoftmodSetups( infoData )
    
    # optional - open animation file (PFFL anim)
    
    animationpath = _makeAnimationFilepath( filepath, animationpath )
    _importAnimation( infoData, animationpath, verbose )
    
    # load original selection
    
    if origsel: mc.select( origsel )
    else: mc.select( d = 1 )


def _createSoftmodSetups( infoData ):
    
    setupResList = []
    
    for softmodData in infoData:
        
        # skip data from corrupted softmod setups
        
        if not softmodData or softmodData['missingControl']:
            
            continue
        
        # collect softmod setup data
        
        falloffMode = softmodData['falloffMode']
        falloffRadius = softmodData['falloffRadius']
        deformCtrl = softmodData['deformCtrl']
        centreControl = softmodData['centreControl']
        prefix = softmodData['prefix']
        edgeIdA = softmodData['edgeIdA']
        edgeIdB = softmodData['edgeIdB']
        deformedShapes = softmodData['deformedShapes']
        edgeBuildShape = softmodData['edgeBuildShape']
        
        edges = [ '%s.e[%d]' % ( edgeBuildShape, edgeIdA ), '%s.e[%d]' % ( edgeBuildShape, edgeIdB ) ]
        
        
        # make softmod setup
        
        if not mc.objExists( edges[0] ) or not mc.objExists( edges[1] ):
            
            print('# skipping this setup, one of these objects do not exist: %s' % edges)
            continue
        
        setupRes = setup.make( edges, prefix = prefix, useAutoIncrement = False )
        
        # add more geos
        
        for dshape in deformedShapes:
            
            if dshape == edgeBuildShape:
                
                continue
            
            setup.addRemGeoFromSculpt( [ dshape ], deformCtrl, mode = 'add' )        
        
        # update setup result list
        
        setupResList.append( setupRes )
    
    
    return setupResList


def _readInfoDataFromFile( filepath ):
    
    # open json file
    
    fileobj = open( filepath, mode = 'r' )    
    softmodDt = json.load( fileobj )
    fileobj.close()
    
    infoData = softmodDt['softModDataList']
    
    return infoData


def _importAnimation( infoData, animationpath, verbose = True ):
    
    if verbose: print('# softmodSculpt - importing animation to %s' % animationpath)
    
    raise Exception( '# import animation not supported' )
    
    # animfileobj.close()


def _exportAnimation( infoData, animationpath, exportAnimation, verbose = True ):
    
    if verbose: print('# softmodSculpt - exporting animation to %s' % animationpath)
    
    raise Exception( '# export animation not supported' )
    
    origsel = mc.ls( sl = 1 )
    
    animatedAts = []
    keyableAts = []
    
    for sminfo in infoData:
        
        if sminfo['missingControl']:
            
            continue
        
        deformctrl = sminfo['deformCtrl']
        centreControl = sminfo['centreControl']
        
        mc.select( deformctrl, centreControl )
        
        keyableAts.extend( keyable_attributes )
        animatedAts.extend( animated_attributes )        
    
    # save animation to manim file
    
    
    # select original selection
    
    if origsel: mc.select( origsel )
    else: mc.select( d = 1 )
    
    return infoData
    
    
def _makeAnimationFilepath( filepath, animationpath ):
    
    if os.path.exists( animationpath ):
        
        animationpath, filext = os.path.splitext( animationpath )
        animationpath = '%s.%s' % ( animationpath, manimext )
        
        return animationpath
    
    # animation path doesnt exist, so it will be determined from softmod file path
    
    filepathdir = os.path.dirname( filepath )
    animationpath = '%s/%s.%s' % ( filepathdir, softmodfilename, manimext )
    
    return animationpath
    


def _exportSoftModSetups( filepath, infoData, verbose = True ):
    
    if verbose: print('# softmodSculpt - exporting softmodSculpt setups to %s' % filepath)
    
    # make dictionary file for json format
    softmodsDt = { 'softModDataList': infoData }
    
    # save json file
    fileobj = open( filepath, mode = 'w' )    
    json.dump( softmodsDt, fileobj, sort_keys = True, indent = 4, separators = ( ',', ': ' ) )
    fileobj.close()
    
    return infoData


def _completeFixAndCreateSoftModFilePath( filepath, autoCreateFilePath, verbose = True, mode = 'import' ):
    
    '''
    function for making correct file path for saving/loading softmodSculpt export file
    as the tool can automatically find the best file and even make subfolder, the same automatic option
    is for loading the softmodSculpt setup
    
    notes on parameters:
    
    mode: valid values are 'import' - for loading setup, 'export' - for exporting setup
    '''
    
    # get destination folder from filepath
    
    folder_sms = os.path.dirname( filepath )
    
    if not filepath:
        
        # auto-generate filepath
        
        if autoCreateFilePath:
            
            job = os.getenv( 'JOB', 'no_job' )
            user = os.getenv( 'USER' )
            jobuserfolder_sms = '/jobs/%s/users/%s/%s' % ( job, user, softmodfolder )
            homeuserfolder_sms = '/home/%s/%s' % ( user, softmodfolder )
            
            folder_sms = jobuserfolder_sms
            if job == 'no_job': folder_sms = homeuserfolder_sms
            
            # create softmodScult folder if it does not exist
            
            if mode == 'export':
                
                if not os.path.exists( folder_sms ):
                    
                    if verbose: print('# softmodSculpt - creating folder %s ...' % folder_sms)
                    
                    os.makedirs( folder_sms )
            
            filepath = folder_sms + '/' + softmodfilename
            
            if mode == 'import':
                
                filepath = '%s.%s' % ( filepath, jsonext )
        
        else:
            
            raise Exception( 'no filepath was provided, use autoCreateFilePath=True option to generate default folder' )
            
    
    # check valid softmodSculpt folder
    
    if not os.path.exists( folder_sms ) and mode == 'export':  # this check is for "export mode" because folder will be optionaly created
        
        raise Exception( 'folder doesnt exist: %s' % folder_sms )
     
    if not os.path.exists( filepath ) and mode == 'import':  # this check is for "import" mode as the file needs to exist
        
        raise Exception( 'filepath doesnt exist: %s' % filepath )
    
    # remove extension and add json extension
    
    if mode == 'export':
        
        filepath, filext = os.path.splitext( filepath )
        filepath = '%s.%s' % ( filepath, jsonext )
    
    
    return filepath


def _exportControl_collectData( allSoftModMainGroups, infoData, verbose = True ):
    
    '''
    collect data for each softmod setup using softmod group as starting point
    '''
    
    for softmodMainGrp in allSoftModMainGroups:
        
        softmodData = {}
        softmodData = _exportControl_collectData_group( softmodMainGrp, softmodData, verbose )
        
        infoData.append( softmodData )
    
    return infoData
        
        
def _exportControl_collectData_group( softmodMainGrp, softmodData, verbose = True ):
    
    '''
    collect data for softmod setup using softmod group as starting point
    '''
    
    # get control settings
    
    if not mc.objExists( softmodMainGrp + '.' + setup.deformControlMsgAt ) or not mc.objExists( softmodMainGrp + '.' + setup.centreControlMsgAt ):
        
        if verbose:
            
            print('softmodSculpt Export, softmodSculpt group %s seems corrupted, skipping export of this one' % softmodMainGrp)
        
        softmodData['missingControl'] = True
        
        return softmodData
    
    deformControl = mc.listConnections( softmodMainGrp + '.' + setup.deformControlMsgAt )
    if deformControl: deformControl = deformControl[0]
    
    centreControl = mc.listConnections( softmodMainGrp + '.' + setup.centreControlMsgAt )
    if centreControl: centreControl = centreControl[0]
    
    softmodData['missingControl'] = True
    
    if deformControl and centreControl and mc.objExists( deformControl ) and mc.objExists( centreControl ):
        
        softmodData['missingControl'] = False
        softmodData['falloffMode'] = mc.getAttr( deformControl + '.falloffMode' )
        softmodData['falloffRadius'] = mc.getAttr( deformControl + '.falloffRadius' )
        softmodData['deformCtrl'] = deformControl
        softmodData['centreControl'] = centreControl
        
        prefixname = name.removeSuffix( softmodMainGrp )
        prefixendindex = len( prefixname ) - len( setup.moduleSuffix )
        softmodData['prefix'] = prefixname[:prefixendindex]
    
    else:
        
        if verbose:
            
            print('softmodSculpt Export, deform control or centre control from %s not found, skipping export of this one' % softmodMainGrp)
        
        return softmodData
    
    
    # get edges
    
    cmeA = mc.listConnections( '%s.%s' % ( deformControl, setup.curveFromMeshEdgeAMsgAt ) )[0]
    cmeB = mc.listConnections( '%s.%s' % ( deformControl, setup.curveFromMeshEdgeBMsgAt ) )[0]
    
    edgeA = mc.getAttr( '%s.edgeIndex' % cmeA )[0][0]
    edgeB = mc.getAttr( '%s.edgeIndex' % cmeB )[0][0]
    
    softmodData['edgeIdA'] = edgeA
    softmodData['edgeIdB'] = edgeB
    
    
    # get shapes
    
    deformernode = mc.listConnections( deformControl + '.' + setup.deformerMsgAt )
    if deformernode: deformernode = deformernode[0]
    
    deformedshapes = deformer.getConnectedShapes( deformernode, returnShapeTypes = [ 'mesh'] )
    
    # convert shapes to transforms which is more robust if Maya adds new shapes
    # in animation scenes
    deformedshapes = [ mc.listRelatives( s, p = 1 )[0] for s in deformedshapes ]
    deformedshapes = list( set( deformedshapes ) )
    
    if verbose:
        
        print('# softmodSculpt Export, found deformed shapes on %s : %s' % ( deformControl, deformedshapes ))
    
    softmodData['deformedShapes'] = deformedshapes

    # find shape used to attach rivet
    curveFromMeshEdgeHistory = mc.listHistory( cmeA )
    geo = None

    for h in curveFromMeshEdgeHistory:

        if 'shape' in mc.nodeType( h, inherited = 1 ):

            geoShape = h
            geo = mc.listRelatives( geoShape, p = 1 )[0]
            break

    softmodData['edgeBuildShape'] = geo  # assuming that edges were built on the first returned deformed shape
    
    return softmodData
    

def _fixControlNames( infoData, verbose = True ):
    
    for d in infoData:
        
        if not d or d['missingControl']:
            
            continue
        
        prefix = d['prefix']
        deformControl = d['deformCtrl']
        centreControl = d['centreControl']
        
        try:
            
            mc.rename( deformControl, prefix + setup.deformCtrlPrefix + '_ctl' ) 
            mc.rename( centreControl, prefix + setup.centreCtrlPrefix + '_ctl' )
        
        except:
            
            if verbose:
                
                print('# softmodSculpt Export, could not rename deform/centre controls from %s%s_grp' % ( prefix, setup.moduleSuffix ))
            
            pass
        

def deleteSetupInScene( verbose = True ):

    """
    delete main group with softmods in current scene
    """

    # delete all main groups in case they are not under top group

    if verbose:

        print('# softmodSculpt - deleting all softmodSculpt setups in scene')

    softModGroups = mc.ls( '*.%s' % setup.mainGrpAt, o = 1 )

    for grp in softModGroups:

        mc.delete( grp )


    # delete top group

    softModTopGroups = mc.ls( '*.%s' % setup.topGrpAt, o = 1 )

    for grp in softModTopGroups:

        mc.delete( grp )







