'''
setup.py at softmodSculpt

functions for making the sculpt deformation and controls

Copyright: www.rig-it.net (c)2021

CAUTION:
Use this tool at your own risk. Skuld s.r.o. and its services RIG-IT take no responsibility
for any damage caused by using this tool
'''
# from __future__ import print_function

import maya.cmds as mc

from ...utils import connect
from ...utils import control
from ...utils import curve
from ...utils import name
from ...utils import rivet
from ...utils import transform

# PIPELINE TOOL ATTRIBUTES - DO NOT CHANGE

# attributes and names to find softmod objects in scene

topGrpAt = 'softmodTopGrp'
mainGrpAt = 'softmodMainGrp'
deformControlAt = 'softmodDeformCtrl'
centreControlAt = 'softmodCentreCtrl'
topGrpName = 'softmodSculptTop_grp'
deformerMsgAt = 'deformerMsg'
controlMsgAt = 'controlMsg'
deformCtrlPrefix = 'Deform'
moduleSuffix = 'Module'
centreCtrlPrefix = 'Centre'

# other internal attributes

mainGrpMsgAt = 'mainGrpMsg'
deformControlMsgAt = 'deformCtrlMsg'
centreControlMsgAt = 'centreCtrlMsg'
curveFromMeshEdgeAMsgAt = 'curveFromMeshEdgeNodeAMsg'
curveFromMeshEdgeBMsgAt = 'curveFromMeshEdgeNodeBMsg'


def make( edges, prefix = 'softmodSculpt', useAutoIncrement = True ):
    
    '''
    @param edges: 2 edges for making rivet
    @type edges: list( str ), 2 string values
    @param mesh: name of mesh transform to be deformed
    @type mesh: str
    @param prefix: base prefix for all new nodes
    @type prefix: str
    @return: list( str, Control, Control ), (module group, main control, centre control)
    
    create relative soft modification on mesh with a control
    '''
    
    # get mesh objects
    
    mesh = edges[0].split( '.' )[0]
    meshxform = mesh
    
    if 'shape' in mc.nodeType( mesh, inherited = 1 ): 
        
        meshxform = mc.listRelatives( mesh, p = 1 )[0]
    
    # -------------------------
    # make top group
    # -------------------------
    
    topGrp = topGrpName
    
    if not mc.objExists( topGrpName ):
        
        topGrp = mc.group( n = topGrpName, em = 1 )
        mc.addAttr( topGrp, ln = topGrpAt, dt = 'string' )
    
    
    # -------------------------
    # get last increment number
    # -------------------------
    
    if useAutoIncrement:
        
        prefix = prefix + name.findNextIncrement( prefix, suffix = moduleSuffix + '_grp' )
    
    # -------------------------
    # make module group
    # -------------------------
    
    moduleGrp = mc.group( n = prefix + 'Module_grp', p = topGrp, em = 1 )
    
    mc.addAttr( moduleGrp, ln = mainGrpAt, dt = 'string' )
    
    
    #===========================================================================
    # make rivet group
    #===========================================================================
    
    if len( edges ) != 2:
        
        raise Exception( '# number of edges passed is not 2, idle...' )
    
    rivetGrp = mc.group( n = prefix + '_grp', em = 1, p = moduleGrp )
    mc.setAttr( rivetGrp + '.v', 0 )
    
    # setup rivet
    
    rivetobjs = rivet.classicrivet( edges, rivetGrp, prefix = prefix )
    
    snapshotrivetgrp = mc.group( n = prefix + 'RivetSnapshot_grp', p = rivetGrp, em = 1 )
    mc.parent( snapshotrivetgrp, moduleGrp )
    
    # now disconnect rivet system to prevent cycle
    
    for obj in rivetobjs[1:]: mc.setAttr( obj + '.nodeState', 2 )  # using blocked, because hasNoEffect doesn`t work on constraints
    
    shapeinputs = rivetobjs[0]
    for si in shapeinputs:
    	
    	src = mc.listConnections( si, s = 1, d = 0, p = 1 )[0]
    	mc.disconnectAttr( src, si )
    
    
    #===========================================================================
    # make softmodSculpt control
    #===========================================================================
    
    # measure edge width
    
    edgebox = mc.exactWorldBoundingBox( edges )
    edgeboxdist = transform.measureDistanceBetween2Points( [ edgebox[0], edgebox[1], edgebox[2] ], [ edgebox[3], edgebox[4], edgebox[5]  ] )
    ctrlscale = edgeboxdist * 0.3
    falloffdist = edgeboxdist * 2.0
    
    # make controls
    
    mainctrl = control.Control( prefix = prefix + deformCtrlPrefix, lockHideDef = ['v'], shape = 'sphere', moveTo = rivetGrp, ctrlParent = moduleGrp, scale = ctrlscale )    
    
    # connect main control to main group
    
    connect.message( mainctrl.C, moduleGrp, mainGrpMsgAt, deformControlMsgAt )
    
    # add cluster attributes to control
    
    mc.addAttr( mainctrl.C, ln = deformControlAt, dt = 'string' )
    
    mc.addAttr( mainctrl.C, ln = 'softModSettings', at = 'enum', enumName = '__________' )
    mc.setAttr( mainctrl.C + '.softModSettings', cb = 1 )
    
    mc.addAttr( mainctrl.C, ln = 'envelope', at = 'float', dv = 1.0, min = 0, max = 1, k = 1 )
    mc.addAttr( mainctrl.C, ln = 'falloffMode', at = 'enum', enumName = 'volume:surface', k = 1 )
    mc.addAttr( mainctrl.C, ln = 'falloffRadius', at = 'float', dv = falloffdist, min = 0, k = 1 )
    
    
    # attach control    
    
    mc.parentConstraint( rivetGrp, mainctrl.Off )
    
    
    # make softmodSculpt centre control
    
    centrectrl = control.Control( prefix = prefix + centreCtrlPrefix, shape = 'locator', moveTo = rivetGrp, ctrlParent = moduleGrp, scale = ctrlscale )
    mc.parentConstraint( rivetGrp, centrectrl.Off )
    
    # connect centre control to main group and add id attribute
    
    connect.message( centrectrl.C, moduleGrp, mainGrpMsgAt, centreControlMsgAt )
    mc.addAttr( centrectrl.C, ln = centreControlAt, dt = 'string' )    
    
    
    # connect centre control with main control using line
    
    connectline = curve.makeConnectionLine( mainctrl.C, centrectrl.C, prefix = prefix )
    mc.parent( connectline, moduleGrp )
    
    # connect setup nodes to main control
    
    cfeMessageAts = [ curveFromMeshEdgeAMsgAt, curveFromMeshEdgeBMsgAt ]
    
    for cfromedgenodePlug, msgAt in zip( rivetobjs[0], cfeMessageAts ):
        
        cfromedgenode = mc.ls( cfromedgenodePlug, o = 1 )[0]
        connect.message( mainctrl.C, cfromedgenode, msgAt, deformControlMsgAt )
    
    
    #===========================================================================
    # make object space softmod handle setup
    #===========================================================================
    
    # this will cancel any transformations of deformed object in world space
    # these transformations would lead to added deformations
    
    # make control world position null
    
    centreCompensOff = mc.group( n = prefix + centreCtrlPrefix + 'TrOff_grp', em = 1, p = moduleGrp )
    centreCompensGrp = mc.group( n = prefix + centreCtrlPrefix + 'Tr_grp', em = 1, p = centreCompensOff )
    
    mc.parentConstraint( meshxform, centreCompensOff, mo = 1 )
    mc.pointConstraint( centrectrl.C, centreCompensGrp )
    
    # make softmod handle groups
    
    worldctrloff = mc.group( n = prefix + 'WorldCtrlOff_grp', em = 1, p = moduleGrp )
    worldctrloffrot = mc.group( n = prefix + 'WorldCtrlOffRot_grp', em = 1, p = worldctrloff )
    worldctrlgrp = mc.group( n = prefix + 'WorldCtrl_grp', em = 1, p = worldctrloffrot )
    localctrloff = mc.group( n = prefix + 'LocalCtrlOff_grp', em = 1, p = centreCompensOff )
    
    mc.parentConstraint( mainctrl.Off, localctrloff )
    mc.delete( mc.parentConstraint( meshxform, worldctrloff ) )
    mc.delete( mc.parentConstraint( mainctrl.C, worldctrloffrot ) )
    mc.setAttr( worldctrloff + '.t', 0, 0, 0 )
    mc.setAttr( worldctrloff + '.r', 0, 0, 0 )
    
    
    # connect transforms of rivet and controls to world space setup
    
    mc.connectAttr( localctrloff + '.t', worldctrloffrot + '.t' )
    mc.connectAttr( localctrloff + '.r', worldctrloffrot + '.r' )
    mc.connectAttr( localctrloff + '.ro', worldctrloffrot + '.ro' )
    
    mc.connectAttr( mainctrl.C + '.t', worldctrlgrp + '.t' )
    mc.connectAttr( mainctrl.C + '.r', worldctrlgrp + '.r' )
    mc.connectAttr( mainctrl.C + '.s', worldctrlgrp + '.s' )
    mc.connectAttr( mainctrl.C + '.ro', worldctrlgrp + '.ro' )
    
    
    #===========================================================================
    # make softmod deformer
    #===========================================================================
    
    softmodDef = mc.softMod( mesh, n = prefix + '_sfm', wn = ( worldctrlgrp, worldctrlgrp ), bs = 1 )[0]
    mc.connectAttr( worldctrlgrp + '.parentInverseMatrix', softmodDef + '.bindPreMatrix' )
    mc.connectAttr( centreCompensGrp + '.t', softmodDef + '.falloffCenter' )
    
    # link deformer to main control for easy search
    
    connect.message( mainctrl.C, softmodDef, deformerMsgAt, controlMsgAt )
    
    #===========================================================================
    # reconnect shape data into rivet
    #===========================================================================
    
    beforesmoftmod = mc.listConnections( softmodDef + '.input[0].inputGeometry', s = 1, d = 0, p = 1 )[0]
    for si in shapeinputs:
    	
    	mc.connectAttr( beforesmoftmod, si )
    	
    for obj in rivetobjs[1:]: mc.setAttr( obj + '.nodeState', 0 )
    
    
    
    #===========================================================================
    # connect cluster attributes from control
    #===========================================================================

    mc.connectAttr( mainctrl.C + '.envelope', softmodDef + '.envelope' )
    mc.connectAttr( mainctrl.C + '.falloffMode', softmodDef + '.falloffMode' )
    mc.connectAttr( mainctrl.C + '.falloffRadius', softmodDef + '.falloffRadius' )
    
    
    #==========================================================================
    # cleanup
    #==========================================================================
    
    mc.delete( snapshotrivetgrp )


    #==========================================================================
    # select control for convenience
    #==========================================================================

    mc.select( mainctrl.C )
    
    return moduleGrp, mainctrl, centrectrl


def deleteControls( verbose = True ):
    
    """
    delete controls based on user selection
    """
    
    selection = mc.ls( sl = True, o = True )
    
    if not selection:
        
        pass
    
    # collect main groups to delete
    
    mainGroups = []
    
    for sel in selection:
        
        try:
            
            mainGroup = mc.listConnections( sel + '.' + mainGrpMsgAt )[0]
            mainGroups.append( mainGroup )
                    
        except:
            
            if verbose:
                
                print('# could not find connection to main group, idle...')
    
    print('mainGroups', mainGroups)
    mainGroups = list( set( mainGroups ) )  # remove duplicate names
    
    if verbose:
        
        print('# deleting SoftMod control main groups %s' % mainGroups)
    
    mc.delete( mainGroups )
    
    
            


def addRemGeoFromSculpt( geoms, ctrl, mode = 'add', verbose = True ):
    
    '''
    @param geoms: geometry to be added/removed from deformer set
    @param ctrl: [str] a transform of the Sculpt Soft Mode control
    @param mode: [str] mode of action, use: 'add', 'remove'
    '''
    
    deformernode = mc.listConnections( ctrl + '.deformerMsg' )[0]
    deformerset = mc.listConnections( deformernode, type = 'objectSet' )[0]
    
    # check if objects exist
    
    existingGeoms = []
    
    for g in geoms:
        
        if not mc.objExists( g ):
            
            if verbose:
                
                print('# object not found: %s - skipping %s operation on softMod deformer' % ( g, mode ))
            
            continue
        
        existingGeoms.append( g )
    
    
    if mode == 'add':
        
        mc.sets( existingGeoms, add = deformerset )
    
    if mode == 'remove':
        
        mc.sets( existingGeoms, rm = deformerset )

   

    
    
