'''
speedMeasureLoc.py

About this tool:

SpeedMeasureLoc creates annotation locator attached to selected object which measures
speed of that object. Various speed units are available.
The tool also measures acceleration but this is hidden at this version.

How to Run:

from rigitAnimTools import speedMeasureLoc
speedMeasureLoc.toolUI()


All copyrights reserved to RIG-IT.net
'''
from __future__ import division
from __future__ import print_function

import math

import maya.OpenMaya as om
import maya.cmds as mc
import maya.mel as mm

toolName = 'speedMeasureLoc'
currentTimeAt = 'currentTime'
speedGroundTopGrp = 'speedGroundTop_grp'
speedGroundSetupIdAt = 'speedGroundSetupId'
positionAt = 'position'
oldPositionAt = 'oldPosition'
startFrameAt = 'startFrame'
speedAt = 'speed'
oldSpeedAt = 'oldSpeed'
accelerationAt = 'acceleration'
groundContactAt = 'groundContact'
accelerationMinAt = 'accelerationMin'
accelerationMaxAt = 'accelerationMax'
stopImpulseAt = 'stopImpulse'
stopImpulseFrameLengthAt = 'stopImpulseFrameLength'
stopImpulseStartFrameAt = 'stopImpulseStartFrame'
accelerationNormAt = 'accelerationNorm'
groundDistanceAt = 'groundDistance'
groundDistToleranceAt = 'groundDistTolerance'
groundLevelYAt = 'groundLevelY'
systemEnableAt = 'systemEnable'
stopImpulseShaderValInAt = 'impactVaIn'
accelShaderValInAt = 'accelerationVaIn'
speedGroundExpressionName = 'speedGroundSetup_exp'
centimetersPerUnitAt = 'centimetersPerUnit'
speedCustomAt = 'speedCustom'
speedCustomStrAt = 'speedCustomStr'
speedMeasureTypeAt = 'speedMeasureType'
annotLocResMsgAt = 'annotLocResMsg'
rigitSpeedMeasureSysAt = 'rigitSpeedMeasureSys'

#===============================================================================
# UI functions
#===============================================================================

uiuiObjectsTextField = None


def toolUI():
    
    """
    build tool UI
    """
    
    window = name_capitalizeFirst( toolName )
    windowObj = toolName
    if ( mc.window( windowObj, exists = True ) ):
        
        mc.deleteUI( windowObj )
    
    
    window = mc.window( windowObj, title = "RIG-IT {0} v1.0.0".format( name_capitalizeFirst( toolName ) ), widthHeight = ( 300, 300 ),
                            maximizeButton = True, resizeToFitChildren = True )
    mc.columnLayout()
    mc.text( '   Setup   ', h = 30 )
    mc.separator()
    mc.setParent( '..' )
    
    mc.rowColumnLayout( numberOfColumns = 2, columnWidth = [( 1, 50 ), ( 2, 250 )] )    
    mc.text( '==>' )
    # mc.columnLayout(columnWidth = 200)
    mc.button( l = 'Setup measure locator on selected object', c = setupOnSelected )
    mc.text( ' ' )
    mc.setParent( '..' )
    
    mc.rowColumnLayout( numberOfColumns = 2, columnWidth = [( 1, 50 ), ( 2, 250 )] )    
    mc.text( '\_/' )
    mc.button( l = 'Delete Speed Measures From Scene', c = deleteSystemsFromScene )
    mc.setParent( '..' )
    
    mc.separator( h = 10, style = 'none' )
    
    
    mc.showWindow( window )
    mc.window( windowObj, e = True, w = 350, h = 120 )

def setupOnSelected( *args ):
    
    """
    get selection of objects in scene
    """
    
    selection = mc.ls( sl = 1 )
    
    if not selection:
        
        return
    
    attachObj = selection[0]
    attachObjPrefix = name_removeSuffix( attachObj ) + '_'
    
    setupRes = speedGroundSetup( attachObj, prefix = attachObjPrefix, relativeObject = None, groundGeometry = None, startFrame = 1, groundDistTolerance = 1.0, shapeScale = 1.0 )
    indicatorMesh = setupRes[2]
    annotObj = setupRes[3]
    
    mc.select( annotObj )
    
    
    

def deleteSystemsFromScene( *args ):
    
    """
    delete speed measure objects from scene
    """
    
    userResult = mc.confirmDialog( title = 'Delete Speed Measure System From Scene', message = 'Do you want to delete Speed Measure system from this scene?',
                          button = ['Yes', 'No'], defaultButton = 'Yes', cancelButton = 'No', dismissString = 'No',
                          messageAlign = 'center' )
        
    if userResult == 'No':
        
        return
    
    speedMeasureSysObjs = mc.ls( '*.' + rigitSpeedMeasureSysAt, o = True, recursive = True )
    
    if speedMeasureSysObjs:
        
        mc.delete( speedMeasureSysObjs )
        print('# RIGIT Speed Measure system objects deleted. List of deleted objects: {0}'.format( speedMeasureSysObjs ))
    


#===============================================================================
# tool functions
#===============================================================================

def speedGroundSetup( movingObject, prefix = '', relativeObject = None, groundGeometry = None, startFrame = 1, groundDistTolerance = 1.0, shapeScale = 1.0 ):
    
    '''
    setup simulation based ground impact system for driving flex blendShapes and other creature effects
    Useful to be attached to foot transform to detect when foot hits the ground and stands
    and generate usually normalized scalar values to drive blendShapes etc.
    
    @param movingObject: name of moving object, for example foot joint
    @type movingObject: str
    @param prefix: prefix to name new objects, if nothing provided, will autogenerate from movingObject
    @type prefix: str
    @param relativeObject: optional, name of relative space object, default is world
    @type relativeObject: str
    @param groundGeometry: optional, name of ground geometry for custom terrain, default is Y world zero value + offset 
    @type groundGeometry: str
    @param startFrame: start frame for simulation
    @type startFrame: int
    @param groundDistTolerance: distance threshold tolerance for assuming zero ground distance
    @type groundDistTolerance: float
    @return: list(str), 0- name of speed ground main group, 1- name of speed group top group 
    '''
    
    prefix = prefix + 'speedGround'
    
    # make groups
    
    topGrp = speedGroundTopGrp
    
    if not mc.objExists( speedGroundTopGrp ):
        
        topGrp = mc.group( n = speedGroundTopGrp, em = 1 )
        mc.addAttr( topGrp, ln = rigitSpeedMeasureSysAt, at = 'bool' )
        mc.setAttr( topGrp + '.' + rigitSpeedMeasureSysAt, l = True )
    
    mainGrp = mc.group( n = prefix + 'main_grp', em = 1, p = topGrp )
    relativeSpaceGrp = mc.group( n = prefix + 'relativeSpace_grp', em = 1, p = mainGrp )
    speedGroundOffGrp = mc.group( n = prefix + 'objOffset_grp', em = 1, p = relativeSpaceGrp )
    speedGroundObj = mc.group( n = prefix + 'obj_grp', em = 1, p = speedGroundOffGrp )
    
    # make annotation locator
    annotLoc = mc.spaceLocator( n = prefix + 'Annot_loc' )[0]
    annotLocShape = mc.listRelatives( annotLoc, s = True )[0]
    mc.hide( annotLocShape )
    
    annotShape = mc.createNode( 'annotationShape', n = prefix + '_antShape' )
    annotObj = mc.listRelatives( annotShape, p = 1 )[0]
    annotObj = mc.rename( annotObj, prefix + '_ant' )
    annotObjGrp = mc.group( annotObj, n = prefix + 'Annot_grp' )
    mc.parent( annotLoc, annotObjGrp )
    
    mc.connectAttr( annotLocShape + '.worldMatrix[0]', annotShape + '.dagObjectMatrix[0]' )
    
    mc.setAttr( annotObj + '.text', '0.0', type = 'string' )
    mc.delete( mc.pointConstraint( movingObject, annotObjGrp ) )
    mc.move( shapeScale * 2, shapeScale * 2, shapeScale * 2, annotObj, r = True )
    mc.parent( annotObjGrp, mainGrp )
    mc.parentConstraint( movingObject, annotObjGrp, mo = True )
    
    mc.addAttr( mainGrp, ln = annotLocResMsgAt, at = 'message' )
    mc.connectAttr( annotObj + '.msg', mainGrp + '.' + annotLocResMsgAt )
    
    # adjust annotation object attributes
    for channel in ['r', 's', 'v']:
        
        if not channel == 'v':
            
            for axis in ['x', 'y', 'z']:
                
                mc.setAttr( annotObj + '.' + channel + axis, l = True, k = False )
        
        else:
            
            mc.setAttr( annotObj + '.' + channel, l = True, k = False )
            
    mc.addAttr( annotObj, ln = speedMeasureTypeAt, at = 'enum', enumName = 'meterSec:kilomHour:milesHour', k = 1, dv = 0 )
    mc.addAttr( annotObj, ln = centimetersPerUnitAt, at = 'float', k = 1, dv = 1 )
    
    #===========================================================================
    # setup impact system
    #===========================================================================
    
    # add attributes
    
    sectionLine = '____'
    
    attribute_addSection( mainGrp, 'userSettings' + sectionLine )
    
    mc.addAttr( mainGrp, ln = startFrameAt, at = 'short', k = 1, dv = startFrame )
    mc.addAttr( mainGrp, ln = accelerationMinAt, at = 'float', k = 1, dv = 10 )
    mc.addAttr( mainGrp, ln = accelerationMaxAt, at = 'float', k = 1, dv = 50 )
    mc.addAttr( mainGrp, ln = stopImpulseFrameLengthAt, at = 'short', k = 1, dv = 4, min = 0 )
    mc.addAttr( mainGrp, ln = groundLevelYAt, at = 'float', k = 1, dv = 0.0 )
    mc.addAttr( mainGrp, ln = groundDistToleranceAt, at = 'float', k = 1, dv = groundDistTolerance )
    mc.addAttr( mainGrp, ln = centimetersPerUnitAt, at = 'float', k = 1, dv = 1 )
    mc.addAttr( mainGrp, ln = systemEnableAt, at = 'short', k = 1, min = 0, max = 1, dv = 1 )
    mc.addAttr( mainGrp, ln = speedMeasureTypeAt, at = 'enum', enumName = 'meterSec:kilomHour:milesHour', k = 1, dv = 0 )
    
    attribute_addSection( mainGrp, 'systemOutput' + sectionLine )
    
    mc.addAttr( mainGrp, ln = groundContactAt, at = 'short', k = 1, min = 0, max = 1 )
    mc.addAttr( mainGrp, ln = stopImpulseAt, at = 'float', k = 1 )
    mc.addAttr( mainGrp, ln = accelerationNormAt, at = 'float', k = 1 )
    
    
    attribute_addSection( mainGrp, 'simulationResults' + sectionLine )
    
    mc.addAttr( mainGrp, ln = currentTimeAt, at = 'float', k = 1 )
    mc.addAttr( mainGrp, ln = groundDistanceAt, at = 'float', k = 1 )
    mc.addAttr( mainGrp, ln = speedAt, at = 'float', k = 1 )
    mc.addAttr( mainGrp, ln = oldSpeedAt, at = 'float', k = 1 )
    mc.addAttr( mainGrp, ln = accelerationAt, at = 'float', k = 1 )
    mc.addAttr( mainGrp, ln = speedCustomAt, at = 'float', k = 1 )
    mc.addAttr( mainGrp, ln = speedCustomStrAt, dt = 'string' )
    
    mainGrpId = _speedGroundGetUniqueId()
    mc.addAttr( mainGrp, ln = speedGroundSetupIdAt, at = 'short', dv = mainGrpId, min = 0, k = 1 )
    mc.setAttr( mainGrp + '.' + speedGroundSetupIdAt, l = 1 )
    
    attribute_addCompoundAndChildren( mainGrp, positionAt, [positionAt + 'X', positionAt + 'Y', positionAt + 'Z'], keyable = 1, attrType = 'float' )
    attribute_addCompoundAndChildren( mainGrp, oldPositionAt, [oldPositionAt + 'X', oldPositionAt + 'Y', oldPositionAt + 'Z'], keyable = 1, attrType = 'float' )
    
    mc.addAttr( mainGrp, ln = stopImpulseStartFrameAt, at = 'short', k = 1 )
    
    # connect annotation locator attributes to main group
    for at in [speedMeasureTypeAt, centimetersPerUnitAt]:
        
        mc.connectAttr( annotObj + '.' + at, mainGrp + '.' + at )
    
    
    # attach speed ground object to moving object
    
    if relativeObject:
        
        mc.parentConstraint( relativeObject, relativeSpaceGrp )
    
    mc.delete( mc.pointConstraint( movingObject, speedGroundOffGrp ) )
    mc.pointConstraint( movingObject, speedGroundObj )
    
    # connect ground level
    
    _setupGroundLevelConnection( speedGroundObj, groundGeometry, prefix, mainGrp, relativeSpaceGrp )
    
    
    # setup system connections
    
    timeNode = mc.ls( type = 'time' )[0]
    
    mc.connectAttr( timeNode + '.outTime', mainGrp + '.' + currentTimeAt )
    mc.connectAttr( speedGroundObj + '.t', mainGrp + '.' + positionAt )
    
    accelConditionNode = mc.createNode( 'condition', n = prefix + 'Accel_con' )
    mc.setAttr( accelConditionNode + '.operation', 2 )
    mc.connectAttr( accelConditionNode + '.outColorR', mainGrp + '.' + accelerationNormAt )
    
    accelRemapNode = remapValue_setupNode( prefix = prefix + 'Acceleration',
                                          inputPlug = mainGrp + '.' + accelerationAt,
                                          outMinVal = 0,
                                          outMaxVal = 1
                                          )
    
    accelRemapNegNode = remapValue_setupNode( prefix = prefix + 'AccelerationNeg',
                                          inputPlug = mainGrp + '.' + accelerationAt,
                                          outMinVal = 0,
                                          outMaxVal = 1
                                          )
    
    connect_negate( mainGrp + '.' + accelerationMinAt, destplug = accelRemapNegNode + '.inputMin' )
    connect_negate( mainGrp + '.' + accelerationMaxAt, destplug = accelRemapNegNode + '.inputMax' )
    mc.connectAttr( mainGrp + '.' + accelerationMinAt, accelRemapNode + '.inputMin' )
    mc.connectAttr( mainGrp + '.' + accelerationMaxAt, accelRemapNode + '.inputMax' )
    
    mc.connectAttr( accelRemapNode + '.outValue', accelConditionNode + '.firstTerm' )
    mc.connectAttr( accelRemapNegNode + '.outValue', accelConditionNode + '.secondTerm' )
    mc.connectAttr( accelRemapNode + '.outValue', accelConditionNode + '.colorIfTrueR' )
    mc.connectAttr( accelRemapNegNode + '.outValue', accelConditionNode + '.colorIfFalseR' )
    
    _setupStopImpulse( prefix, mainGrp )
    
    
    # setup shader test object
    
    shadedCheckMesh, stopImpulseShader, accelerationShader = _speedGroundMakeShadedCheckMesh( prefix, scale = shapeScale )
    mc.parent( shadedCheckMesh, mainGrp )
    mc.pointConstraint( movingObject, shadedCheckMesh )
    
    mc.connectAttr( mainGrp + '.' + stopImpulseAt, shadedCheckMesh + '.' + stopImpulseShaderValInAt )
    mc.connectAttr( mainGrp + '.' + accelerationNormAt, shadedCheckMesh + '.' + accelShaderValInAt )
    
    
    # make expression
    
    speedGroundExpressions = [ e for e in mc.ls( type = 'expression' ) if e == speedGroundExpressionName ]
    
    if not len( speedGroundExpressions ):
        
        expString1 = 'python( "from rigitAnimTools import speedMeasureLoc" );\n'
        expString2 = 'python( "import maya.cmds as mc" );\n'
        expString3 = 'python( "mc.undoInfo( stateWithoutFlush = False )" );\n'
        expString4 = 'python( "speedMeasureLoc.speedGroundExpressionCall()" );\n'
        expString5 = 'python( "mc.undoInfo( stateWithoutFlush = True )" )'
        expString = expString1 + expString2 + expString3 + expString4 + expString5
        expNode = mc.expression( string = expString, n = speedGroundExpressionName, object = mainGrp )
        
        mc.addAttr( expNode, ln = rigitSpeedMeasureSysAt, at = 'bool' )
        mc.setAttr( topGrp + '.' + rigitSpeedMeasureSysAt, l = True )
    
    
    return [mainGrp, topGrp, shadedCheckMesh, annotObj]



def _speedGroundMakeShadedCheckMesh( prefix, scale = 1.0 ):
    
    '''
    make shaded mesh for checking results of speedGround setup
    
    @param prefix: prefix for new objects
    @type prefix: str
    @param scale: scale to be used for setting size of the test shaded object
    @type scale: float
    @return: list(str), names of new objects, 0- new shaded mesh, 1- stopImpulse shader, 2- acceleration shader
    '''
    
    testMesh = mc.polySphere( n = prefix + 'CheckResult_msh', ch = 0, radius = scale )[0]
    
    testMeshStopImpulseFaces = [testMesh + '.f[180:359]', testMesh + '.f[380:399]' ]
    testMeshAccelFaces = [testMesh + '.f[0:179]', testMesh + '.f[360:379]' ]
    
    # add attributes
    
    mc.addAttr( testMesh, ln = stopImpulseShaderValInAt, at = 'float', min = 0, k = 1 )
    mc.addAttr( testMesh, ln = accelShaderValInAt, at = 'float', min = 0, k = 1 )
    
    # make shaders
    
    stopImpulseShader = mc.shadingNode( 'surfaceShader', n = prefix + 'StopImpulse_shd', asShader = True )
    accelerationShader = mc.shadingNode( 'surfaceShader', n = prefix + 'Acceleration_shd', asShader = True )
    
    stopImpulseColor = [1, 1, 1]
    accelColor = [1, 0, 0]
    
    for shaderNode, colorVals, driverAt in zip( [stopImpulseShader, accelerationShader], [stopImpulseColor, accelColor], [stopImpulseShaderValInAt, accelShaderValInAt] ):
        
        for cval, channel in zip( colorVals, ['R', 'G', 'B'] ):
            
            anim_setDrivenKey( testMesh + '.' + driverAt, shaderNode + '.outColor' + channel, [0, 1], [0, cval] )
    
    # assign shaders
    
    shader_assign( testMeshStopImpulseFaces, stopImpulseShader )
    shader_assign( testMeshAccelFaces, accelerationShader )
    
    
    return [testMesh, stopImpulseShader, accelerationShader]
    


def _setupGroundLevelConnection( speedGroundObj, groundGeometry, prefix, mainGrp, relativeSpaceGrp ):
    
    if not groundGeometry:
        
        _addGroundLevelOffsetY( speedGroundObj + '.ty', prefix, mainGrp )
        return
    
    surfaceGrp = mc.group( n = prefix + 'Surface_grp', em = 1, p = relativeSpaceGrp )
    surfaceDistGrp = mc.group( n = prefix + 'SurfaceDistance_grp', em = 1, p = surfaceGrp )
    
    mc.geometryConstraint( groundGeometry, surfaceGrp )
    mc.aimConstraint( speedGroundObj, surfaceGrp, aim = [0, 1, 0] )
    mc.pointConstraint( speedGroundObj, surfaceGrp )
    mc.pointConstraint( speedGroundObj, surfaceDistGrp )
    
    _addGroundLevelOffsetY( surfaceDistGrp + '.ty', prefix, mainGrp )
    

def _addGroundLevelOffsetY( input1plug, prefix, mainGrp ):
    
    groundLevelYaddNode = mc.createNode( 'addDoubleLinear', n = prefix + 'GroundLevelYAdd_adl' )
    mc.connectAttr( input1plug, groundLevelYaddNode + '.i1' )
    mc.connectAttr( mainGrp + '.' + groundLevelYAt, groundLevelYaddNode + '.i2' )
    
    negMultiNode = mc.createNode( 'multDoubleLinear', n = prefix + 'NegativeVal_mdl' )
    mc.setAttr( negMultiNode + '.i1', -1 )
    mc.connectAttr( groundLevelYaddNode + '.o', negMultiNode + '.i2' )
    
    absNode = mc.createNode( 'condition', n = prefix + 'AbsoluteValue_con' )
    mc.setAttr( absNode + '.operation', 4 )  # less then
    mc.setAttr( absNode + '.secondTerm', 0 )  # less then
    mc.connectAttr( groundLevelYaddNode + '.o', absNode + '.firstTerm' )
    mc.connectAttr( negMultiNode + '.o', absNode + '.colorIfTrueR' )
    mc.connectAttr( groundLevelYaddNode + '.o', absNode + '.colorIfFalseR' )
    
    mc.connectAttr( absNode + '.outColorR', mainGrp + '.' + groundDistanceAt )
    
    
def _setupStopImpulse( prefix, mainGrp ):
    
    '''
    setup stop impulse value output system
    it only needs expression to set when object is not moving
    and frame at which it stopped moving 
    '''
    
    frameDifferenceNode = mc.createNode( 'plusMinusAverage', n = prefix + 'StopImpulseTimeDelta_pma' )
    mc.setAttr( frameDifferenceNode + '.operation', 2 )
    mc.connectAttr( mainGrp + '.' + currentTimeAt, frameDifferenceNode + '.input1D[0]' )
    mc.connectAttr( mainGrp + '.' + stopImpulseStartFrameAt, frameDifferenceNode + '.input1D[1]' )
    
    impulseDurationNormalizeNode = mc.createNode( 'multiplyDivide', n = prefix + 'StopImpulseTimeNorm_mdv' )
    mc.setAttr( impulseDurationNormalizeNode + '.operation', 2 )
    mc.connectAttr( frameDifferenceNode + '.output1D', impulseDurationNormalizeNode + '.input1X' )
    mc.connectAttr( mainGrp + '.' + stopImpulseFrameLengthAt, impulseDurationNormalizeNode + '.input2X' )
    
    impulseCurveNode = remapValue_setupNode( prefix = prefix + 'StopImpluse',
                          inputPlug = impulseDurationNormalizeNode + '.outputX',
                          inMinVal = 0,
                          inMaxVal = 1,
                          outMinVal = 0,
                          outMaxVal = 1,
                          interpolation = 'smooth' )
    
    mc.setAttr( impulseCurveNode + '.value[0].value_FloatValue', 1 )
    mc.setAttr( impulseCurveNode + '.value[1].value_FloatValue', 0 )
    mc.connectAttr( impulseCurveNode + '.outValue', mainGrp + '.' + stopImpulseAt )


def _speedGroundGetUniqueId():
    
    groups = mc.ls( '*.' + speedGroundSetupIdAt, o = 1, recursive = True )
    
    
    if not groups: return 0
    
    idList = [ mc.getAttr( g + '.' + speedGroundSetupIdAt ) for g in groups ]
    idList.sort()
    for i in range( idList[-1] + 2 ):
        
        if i not in idList:
            
            return i

    
def speedGroundExpressionCall():
    
    '''
    this function is called from ground impact/acceleration expression
    '''
    
    # run simulation system
    
    mainGrps = mc.ls( '*.' + speedGroundSetupIdAt, o = 1, recursive = True )
    
    for mainGrp in mainGrps:
        
        _speedGroundExpressionCallPerGroup( mainGrp )


def _speedGroundExpressionCallPerGroup( mainGrp ):
    
    '''
    this function is called from the main expression function and updates groung impact group
    '''
    
    # idle function if system is disabled
    
    systemEnableState = mc.getAttr( mainGrp + '.' + systemEnableAt )
    
    if not systemEnableState:
        
        return
    
    currentFrame = mc.currentTime( q = 1 )
    startFrame = mc.getAttr( mainGrp + '.' + startFrameAt )
    centimetersPerUnit = mc.getAttr( mainGrp + '.' + centimetersPerUnitAt )
    framesRateInSecond = anim_getFpsSetting()
    speedMeasureType = mc.getAttr( mainGrp + '.' + speedMeasureTypeAt )  # metersPerSecond / kilometersPerHour / milesPerHour
    
    # get annotation locator
    annotLoc = mc.listConnections( mainGrp + '.' + annotLocResMsgAt )[0]
    
    # reset values if simulation is at start frame
    
    if currentFrame == startFrame:
        
        currentPositionVals = mc.getAttr( mainGrp + '.' + positionAt )[0]
        mc.setAttr( mainGrp + '.' + oldPositionAt, currentPositionVals[0], currentPositionVals[1], currentPositionVals[2] )
        mc.setAttr( mainGrp + '.' + speedAt, 0.0 )
        mc.setAttr( mainGrp + '.' + oldSpeedAt, 0.0 )
        mc.setAttr( mainGrp + '.' + accelerationAt, 0.0 )
        mc.setAttr( mainGrp + '.' + speedCustomAt, 0.0 )
        mc.setAttr( mainGrp + '.' + speedCustomStrAt, '0.0', type = 'string' )
        mc.setAttr( annotLoc + '.text', '0.0', type = 'string' )
        mc.setAttr( mainGrp + '.' + groundContactAt, 0 )
        mc.setAttr( mainGrp + '.' + stopImpulseStartFrameAt, startFrame )
        
        return
    
    
    # get new values
    
    oldPosition = mc.getAttr( mainGrp + '.' + oldPositionAt )[0]
    position = mc.getAttr( mainGrp + '.' + positionAt )[0]
    distanceOffset = transform_measureDistanceBetween2Points( oldPosition, position )
    
    secondsTime = 1.0 / framesRateInSecond
    speed = float(distanceOffset / secondsTime)
    oldSpeed = mc.getAttr( mainGrp + '.' + oldSpeedAt )
    acceleration = float(( speed - oldSpeed ) / secondsTime)
    groundContact = mc.getAttr( mainGrp + '.' + groundContactAt )
    groundDistance = mc.getAttr( mainGrp + '.' + groundDistanceAt )
    groundDistTolerance = mc.getAttr( mainGrp + '.' + groundDistToleranceAt )
    
    speedTypeConvertFactor = { 0:1, 1:3.6, 2:2.237 }
    speedTypeConvertNames = { 0:'m/s', 1:'km/h', 2:'miles/h' }
    centimeterToMeter = centimetersPerUnit / 100.0
    speedCustom = speed * centimeterToMeter * speedTypeConvertFactor[speedMeasureType]
    
    
    # update attributes
    
    if math.fabs( groundDistance ) <= groundDistTolerance:
        
        mc.setAttr( mainGrp + '.' + groundContactAt, 1 )
        
        if not groundContact:
            
            mc.setAttr( mainGrp + '.' + stopImpulseStartFrameAt, currentFrame )
    
    else:
        
        mc.setAttr( mainGrp + '.' + groundContactAt, 0 )
    
    
    
    # store current values as old values
    
    currentPositionVals = mc.getAttr( mainGrp + '.' + positionAt )[0]
    mc.setAttr( mainGrp + '.' + oldPositionAt, currentPositionVals[0], currentPositionVals[1], currentPositionVals[2] )
    mc.setAttr( mainGrp + '.' + oldSpeedAt, mc.getAttr( mainGrp + '.' + speedAt ) )
    
    # store new values as current values
    
    mc.setAttr( mainGrp + '.' + speedAt, speed )
    mc.setAttr( mainGrp + '.' + accelerationAt, acceleration )
    mc.setAttr( mainGrp + '.' + speedCustomAt, speedCustom )
    
    
    speedCustomRound = round( speedCustom, 2 )
    # speedCustomRoundStr = str( speedCustomRound ) + ' ' + speedTypeConvertNames[speedMeasureType]
    speedCustomRoundStr = '{}'.format(speedCustomRound)
    mc.setAttr( mainGrp + '.' + speedCustomStrAt, speedCustomRoundStr, type = 'string' )
    mc.setAttr( annotLoc + '.text', speedCustomRoundStr, type = 'string' )
    
    


def attribute_addSection( object, atName = 'settings' ):
    
    mc.addAttr( object, ln = atName, at = 'enum', enumName = '__________', k = 0 )
    mc.setAttr( object + '.' + atName, cb = 1, l = 1 )
    
def attribute_addCompoundAndChildren( object, compoundName, childNames, keyable = True, attrType = 'float' ):
    
    '''
    add compound attribute and children attributes
    useful for making scalar 3-dimensional attributes for custom Translate, Rotation etc.
    
    @param object: str, name of object
    @param compoundName: str, name of parameter
    @param childNames: list(str), names of compound attributes, for example with Scale it is scaleX, scaleY, scaleZ
    @param keyable: bool, make each compound keyable
    @param attrType: str, type of attribute, for example 'float', 'short' etc.
    @return: None
    '''
    
    mc.addAttr( object, ln = compoundName, nc = len( childNames ), at = 'compound' )
    
    for atName in childNames:
        
        mc.addAttr( object, ln = atName, keyable = keyable, at = attrType, p = compoundName )
        
def transform_measureDistanceBetween2Points( pointA, pointB ):
    
    '''
    @param objectA: first point
    @type objectA: list(3 floats)
    @param objectB: second point
    @type objectB: list(3 floats)
    @return: float, distance between given points in same space
    '''
    
    vectA = om.MVector( pointA[0], pointA[1], pointA[2] )
    vectB = om.MVector( pointB[0], pointB[1], pointB[2] )
    pointDif = vectB - vectA
    
    return pointDif.length()

def remapValue_setupNode( prefix = '', inputPlug = None, outputPlug = None, inMinVal = 0, inMaxVal = 1, outMinVal = 0, outMaxVal = 1, interpolation = 'spline' ):
    
    '''
    setup remapValue node with options to connect input, output and different settings
    
    @param prefix: prefix for naming new node
    @type prefix: str
    @param inputPlug: input value plug (OBJECT.ATTRIBUTE) or just a scalar value to be set as input value
    @type inputPlug: float OR str
    @param outputPlug: output value plug (OBJECT.ATTRIBUTE)
    @type outputPlug: str
    @param inMinVal: input minimum value OR object plug (OBJECT.ATTRIBUTE)
    @type inMinVal: float OR str
    @param inMaxVal: input maximum value OR object plug (OBJECT.ATTRIBUTE)
    @type inMaxVal: float OR str
    @param outMinVal: output minimum value OR object plug (OBJECT.ATTRIBUTE)
    @type outMinVal: float OR str
    @param outMaxVal: output maximum value OR object plug (OBJECT.ATTRIBUTE)
    @type outMaxVal: float OR str
    @param interpolation: type of interpolation, valid values: 'none','linear','smooth','spline'
    @type interpolation: str
    @return: str, name of new remapValue node
    '''
    
    if not prefix:
        
        prefix = 'new'
    
    remapValueNode = mc.createNode( 'remapValue', n = prefix + '_rmv' )
    _smartConnect( inMinVal, remapValueNode + '.inputMin' )
    _smartConnect( inMaxVal, remapValueNode + '.inputMax' )
    _smartConnect( outMinVal, remapValueNode + '.outputMin' )
    _smartConnect( outMaxVal, remapValueNode + '.outputMax' )
    mc.setAttr( remapValueNode + '.value[%d].value_Interp' % 0, 3 )
    
    
    if inputPlug:
        
        _smartConnect( inputPlug, remapValueNode + '.inputValue' )
    
    if outputPlug:
        
        mc.connectAttr( remapValueNode + '.outValue', outputPlug )
    
    
    return remapValueNode
    
def _smartConnect( connectionValue, connectionPlug ):
    
    '''
    this detects if connection value is existing Maya plug or it assumes it is scalar value
    based on result it either connects it or sets it
    '''
    
    if mc.objExists( connectionValue ):
        
        mc.connectAttr( connectionValue, connectionPlug )
    
    else:
        
        mc.setAttr( connectionPlug, connectionValue )

def connect_negate( sourceplug, destplug = None, prefix = 'utilnode' ):
    
    '''
    negate single input value: X * (-1)
    
    @param sourceplug: name of source plug (object.attribute)
    @type sourceplug: str
    @param destplug: name of destination plug (object.attribute)
    @type destplug: str
    @param prefix: prefix name for new objects
    @type prefix: str
    @return: str, name of new inverse node
    '''
    
    multnode = mc.createNode( 'multDoubleLinear', n = prefix + '_mdl' )
    mc.setAttr( multnode + '.i2', -1 )
    mc.connectAttr( sourceplug, multnode + '.i1' )
    
    if not prefix:
        
        nodeName = sourceplug.split( '.' )[-1]
        prefix = name_removeSuffix( nodeName )
    
    if destplug:
        
        mc.connectAttr( multnode + '.o', destplug, f = 1 )
    
    return multnode

def shader_assign( objects, shadernode, shadingEngine = None ):
    
    '''
    assign shader to objects
    (this can be list of transforms containing shapes, shapes or just components)
    
    @param objects: list of objects to get shader
    @type objects: list(str)
    @param shadernode: name of shader node (must exist for now, later function can be extended to make a shader)
    @type shadernode: str
    @param shadingEngine: name of render set (a.k.a. Shading Engine), will be made here if nothing passed
    @type shadingEngine: str (default None) 
    return: list( str, str ), [0] name of shader node [1] name of render set
    '''
    
    if shadernode == 'lambert1':
        
        shadingEngine = 'initialShadingGroup'
        mc.sets( objects, e = 1, forceElement = 'initialShadingGroup' )
        return [ shadernode, shadingEngine ]
    
    if not shadingEngine or not mc.objExists( shadingEngine ):
        
        shadingEnginename = name_removeSuffix( shadernode )
        shadingEngine = mc.sets( renderable = True, noSurfaceShader = True, empty = True, name = '%s_she' % shadingEnginename )
    
    
    mc.sets( objects, e = 1, forceElement = shadingEngine )
    
    # this is Maya bug, even with connectAttr force=1 errors on existing connection, otherwise gives warning, which would result in loads of warnings
    if not mc.isConnected( shadernode + '.outColor', shadingEngine + '.surfaceShader' ):
        
        mc.connectAttr( shadernode + '.outColor', shadingEngine + '.surfaceShader', f = 1 )
        
    mm.eval( 'connectNodeToNodeOverride("%s", "%s")' % ( shadernode, shadingEngine ) )
    
    return [ shadernode, shadingEngine ]

def anim_getFpsSetting( getString = False ):
    
    """
    Get Frames Per Second playback Maya setting
    
    @param getString: bool, return string value, otherwise return float number of FPS
    @return: float / str (if getString is True)
    """
    
    framesRateInSecondStr = mc.currentUnit( q = True, time = True )
    framesRateInSecondOptions = { 'game':15, 'film':24, 'pal':25, 'ntsc':30, 'show':48, 'palf':50, 'ntscf':60 }
    if 'fps' in framesRateInSecondStr:
        
        fpsStr = framesRateInSecondStr.replace( 'fps', '' )
        framesRateInSecond = float( fpsStr )
        
    else:
        
        framesRateInSecond = framesRateInSecondOptions[framesRateInSecondStr]
    
    
    if getString:
        
        return framesRateInSecondStr
    
    else:
        
        return framesRateInSecond
    
def anim_setDrivenKey( driver, driven, driverVals, drivenVals, cleanDrivenPlug = True, preinf = 0, postinf = 0, intangtype = 'linear', outtangtype = 'linear', verbose = True ):
    
    '''
    easy to use setDrivenKey function, can setup more driven attributes
    it can re-create existing keyframes as well
    
    infinity modes key:
    
    0 - constant
    1 - linear
    3 - cycle - there is strange jump in Maya from 1 to 3
    4 - cycle with offset
    5 - oscillate
    
    
    tangent types key:
    
    - fixed
    - linear
    - flat
    - step
    - slow
    - fast
    - spline
    - clamped
    - plateau
    - stepNext
    
    
    @param driver: driver plug (object.attribute)
    @type driver: str
    @param driven: driven plug (object.attribute)
    @type driven: str
    @param driverVals: driver values
    @type driverVals: float list
    @param drivenVals: driven values
    @type drivenVals: float list
    @param cleanDrivenPlug: clean driven plug in case there was already another SDK connection
    @type cleanDrivenPlug: bool
    @param preinf: preInfinity mode of the driven curve, Constant by default
    @type preinf: int
    @param postinf: postInfinity mode of the driven curve, Constant by default
    @type postinf: int
    @param tangtype: tangentType of the driven curve, Linear by default
    @type tangtype: int
    @param verbose: print comments during function runtime
    @type verbose: bool
    @return: str, name of new function curve
    '''
    
    # prepare driven plug for new SDK
    
    if cleanDrivenPlug:
        
        anim_setDrivenKey_cleanDrivenPlug( driven )
    
    # attempt to unlock the plug, in case this is referenced object
    # it needs to be able to continue
    
    try: mc.setAttr( driven, l = 0 )
    except:
        
        if verbose:
            
            print('# unable to unlock %s to be driven by %s, it has probably incoming connection, skipping setDrivenKey creation ...' % ( driven, driver ))
    
    
    # make SDK
    
    for i in range( len( driverVals ) ):
     
        driverVal = driverVals[i]
        drivenVal = drivenVals[i]
        
        mc.setDrivenKeyframe( driven , cd = driver, dv = driverVal, v = drivenVal )
    
    # adjust SDK
    
    drivencrv = mc.listConnections( driven, s = 1, d = 0, skipConversionNodes = 1 )[0]
    drivenCurves = [ drivencrv ]
    
    # make blend weighted node check
    
    if mc.nodeType( drivencrv ) in ['blendWeighted', 'pairBlend']:
        
        drivenCurves = mc.listConnections( drivencrv, skipConversionNodes = 1, s = 1, d = 0 )
    
    for drivenCrv in drivenCurves:
        
        mc.setAttr( drivenCrv + '.pre', preinf )
        mc.setAttr( drivenCrv + '.pst', postinf )
        mc.keyTangent( drivenCrv, e = 1, itt = intangtype, ott = outtangtype )
    
    
    # legacy return value condition
    
    if len( drivenCurves ) == 1:
        
        return drivenCurves[0]
    
    else:
        
        return drivenCurves


def anim_setDrivenKey_cleanDrivenPlug( driven ):
    
    con = mc.listConnections( driven, s = 1, scn = 1 )
    
    if not con: return
    
    if 'animCurve' in mc.nodeType( con[0], inherited = 1 ):
        
        # if driven was connected to animCurve node
        # and this node wasn`t connected to anything else,
        # delete it
        
        if len( mc.listConnections( con[0], d = 1, scn = 1 ) ) == 1:
            
            mc.delete( con[0] )
            
        # otherwise just disconnect it
        
        else: connect.disconnect( driven )
        
def name_removeSuffix( name ):
    
    name = name_short( name )
    
    if name.count( '_' ) == 0: return name
    
    if name.count( '_' ) == 1:
        if name.startswith( 'l_' ) or name.startswith( 'r_' ):
            return name
            
    edit = name.split( '_' )[-1]
    edit = name[0: len( name ) - len( edit ) - 1 ]
    
    return edit        

def name_short( name ):
    
    return name.split( '|' )[-1]

def name_capitalizeFirst( name ):
    
    '''
    capitalize first letter of given string
    
    @param name: string to process
    @type name: str
    @return: str
    '''
    
    firstletter = name[0].upper()
    suffix = ''
    if len( name ) > 1: suffix = name[1:]
    
    resultstring = firstletter + suffix
    
    return resultstring
