'''
anim and there is also motion path function(s)
@category Rigging @subcategory Utils
@tags animation function curve setdrivenkey motion path

animation utilities
'''
from __future__ import print_function

import maya.cmds as mc
import maya.mel as mm


# from .utils import connect


def functionCurve( ftimes = [0, 1], values = [0, 1], name = 'sdkFunction', startTangent = 'linear', midTangent = 'linear', endTangent = 'linear', infinityType = 4 ):
    
    '''
    @param [float list] function time values (x)
    @values [float list] resulting values of function (y)
    @name [string] name of the function curve
    @param infinityType [int] 0=Constant, 1=Linear, 3=Cycle, 4=Cycle with Offset, 5=Oscillate
    
    infinity modes key:
    
    0 - constant
    1 - linear
    2 - NO MODE HERE
    3 - cycle
    4 - cycle with offset
    5 - oscillate
    '''
    
    if len( ftimes ) is not len( values ):
        
        mm.eval( 'warning -sl 1 ("number of ftimes and values must be equal, function idle...\n")' )
        return
    
    function = mc.createNode( 'animCurveUU', n = name + '_anc' )
    
    mc.setAttr( function + '.pre', infinityType )
    mc.setAttr ( function + '.pst', infinityType )
    
    
    for i in range( len( ftimes ) ):
        
        tang = ''
        
        if i == 0:
            
            tang = startTangent
            
        if i == len( ftimes ) - 1:
            
            tang = endTangent
        
        else:
            
            tang = midTangent
        
        mc.setKeyframe( function, f = ftimes[i], v = values[i], itt = tang, ott = tang )
    
  
    return function


def createMotionPath( pathCurve, pathObject, prefix = 'pathObject', follow = True, fractionMode = True, frontAxis = 'z', upAxis = 'y', worldUpType = 'scene', worldUpVector = [1, 0, 0], worldUpObject = '', uValue = 0.0, inverseFront = False ):

    """
    make motionPath setup using motionPath Maya node
    
    NOTE: this Maya command can be used as alternative for createMotionPath:
          mc.pathAnimation()
    
    @param pathCurve: str, name of nurbs curve for the motionPath node
    @param prefix: str, name prefix for new objects
    @param follow: bool, object will aim along the curve path
    @param fractionMode: bool, position of path object is based on fraction of length of curve, if False it is based on curve parameter
    @param frontAxis: str, set front aim axis of object, to reverse use inverseFront argument, valid values: 'x','y','z'
    @param upAxis: str, set up axis of object, valid values: 'x','y','z'
    @param worldUpType: str, choose worldUp type for object, valid values are: 'scene','object','objectrotation', 'vector', 'normal'
    @param worldUpVector: list(float, float, float), vector defining up vector in reference space, used by 'objectrotation' and 'vector'
    @param worldUpObject: str, name of reference object used by 'objectrotation' worldUpType mode
    @param uValue: float, value setting position of object on path
    @param inverseFront: bool, rotate object 180 degrees so it is facing the path in opposite direction
    @return: str, name of new motionPath node
    """
    
    mPathNode = mc.createNode( 'motionPath', n = prefix + '_mpt' )
    
    mc.setAttr( mPathNode + '.f', follow )
    mc.setAttr( mPathNode + '.fm', fractionMode )
    mc.setAttr( mPathNode + '.fa', {'x':0, 'y':1, 'z':2}[frontAxis] )
    mc.setAttr( mPathNode + '.ua', {'x':0, 'y':1, 'z':2}[upAxis] )
    mc.setAttr( mPathNode + '.wut', {'scene':0, 'object':1, 'objectrotation':2, 'vector':3, 'normal':4}[worldUpType] )
    mc.setAttr( mPathNode + '.wu', worldUpVector[0], worldUpVector[1], worldUpVector[2] )
    mc.setAttr( mPathNode + '.uValue', uValue )
    mc.setAttr( mPathNode + '.inverseFront', inverseFront )
    
    if worldUpType == 'object' or worldUpType == 'objectrotation' and mc.objExists( worldUpObject ):
        
        mc.connectAttr( worldUpObject + '.worldMatrix', mPathNode + '.wum' )
    
    mc.connectAttr( pathCurve + '.worldSpace', mPathNode + '.geometryPath' )
    mc.connectAttr( mPathNode + '.msg', pathObject + '.specifiedManipLocation' )
    
    mc.connectAttr( mPathNode + '.' + 'rotate', pathObject + '.r' )
    mc.connectAttr( mPathNode + '.' + 'allCoordinates', pathObject + '.t' )

    hideMotionPathPositionMarkers(pathCurve)
    
    return mPathNode



def hideMotionPathPositionMarkers( curve ):
    
    curveShape = mc.listRelatives( curve, s = 1 )[0]
    posMarkerShapes = mc.listRelatives( curveShape, ad = 1, f = 1, typ = 'positionMarker' )
    
    if posMarkerShapes:
        
        for pm in posMarkerShapes: mc.setAttr( pm + '.v', 0 )
    
    
def hasMotionPath( object ):
    
    '''
    test if object has motionPath node attached to it
    
    @param object: name of object to check
    @type object: str
    @return: bool, True if motionPath was found
    '''
    
    pathObjHist = mc.listConnections( object )
    
    if pathObjHist:
        
        motionNodes = mc.ls( pathObjHist, typ = 'motionPath' )
        if motionNodes: return 1
    
    return 0


def setDrivenKey( driver, driven, driverVals, drivenVals, cleanDrivenPlug = True, preinf = 0, postinf = 0, intangtype = 'linear', outtangtype = 'linear', verbose = True ):
    
    '''
    easy to use setDrivenKey function, can setup more driven attributes
    it can re-create existing keyframes as well
    
    infinity modes key:
    
    0 - constant
    1 - linear
    3 - cycle - there is strange jump in Maya from 1 to 3
    4 - cycle with offset
    5 - oscillate
    
    
    tangent types key:
    
    - fixed
    - linear
    - flat
    - step
    - slow
    - fast
    - spline
    - clamped
    - plateau
    - stepNext
    
    
    @param driver: driver plug (object.attribute)
    @type driver: str
    @param driven: driven plug (object.attribute)
    @type driven: str
    @param driverVals: driver values
    @type driverVals: float list
    @param drivenVals: driven values
    @type drivenVals: float list
    @param cleanDrivenPlug: clean driven plug in case there was already another SDK connection
    @type cleanDrivenPlug: bool
    @param preinf: preInfinity mode of the driven curve, Constant by default
    @type preinf: int
    @param postinf: postInfinity mode of the driven curve, Constant by default
    @type postinf: int
    @param tangtype: tangentType of the driven curve, Linear by default
    @type tangtype: int
    @param verbose: print comments during function runtime
    @type verbose: bool
    @return: str, name of new function curve
    '''
    
    # prepare driven plug for new SDK
    
    if cleanDrivenPlug:
        
        pass
        # setDrivenKey_cleanDrivenPlug( driven )
    
    # attempt to unlock the plug, in case this is referenced object
    # it needs to be able to continue
    
    try: mc.setAttr( driven, l = 0 )
    except:
        
        if verbose:
            
            print('# unable to unlock %s to be driven by %s, it has probably incoming connection, skipping setDrivenKey creation ...' % ( driven, driver ))
    
    
    # make SDK
    
    for i in range( len( driverVals ) ):
     
        driverVal = driverVals[i]
        drivenVal = drivenVals[i]
        
        mc.setDrivenKeyframe( driven , cd = driver, dv = driverVal, v = drivenVal )
    
    # adjust SDK
    
    drivencrv = mc.listConnections( driven, s = 1, d = 0, skipConversionNodes = 1 )[0]
    drivenCurves = [ drivencrv ]
    
    # make blend weighted node check
    
    if mc.nodeType( drivencrv ) in ['blendWeighted', 'pairBlend']:
        
        drivenCurves = mc.listConnections( drivencrv, skipConversionNodes = 1, s = 1, d = 0 )
    
    for drivenCrv in drivenCurves:
        
        mc.setAttr( drivenCrv + '.pre', preinf )
        mc.setAttr( drivenCrv + '.pst', postinf )
        mc.keyTangent( drivenCrv, e = 1, itt = intangtype, ott = outtangtype )
    
    
    # legacy return value condition
    
    if len( drivenCurves ) == 1:
        
        return drivenCurves[0]
    
    else:
        
        return drivenCurves


def setDrivenKey_cleanDrivenPlug( driven ):
    
    con = mc.listConnections( driven, s = 1, scn = 1 )
    
    if not con: return
    
    if 'animCurve' in mc.nodeType( con[0], inherited = 1 ):
        
        # if driven was connected to animCurve node
        # and this node wasn`t connected to anything else,
        # delete it
        
        if len( mc.listConnections( con[0], d = 1, scn = 1 ) ) == 1:
            
            mc.delete( con[0] )
            
        # otherwise just disconnect it
        
        else: connect.disconnect( driven )

def getFpsSetting( getString = False ):
    
    """
    Get Frames Per Second playback Maya setting
    
    @param getString: bool, return string value, otherwise return float number of FPS
    @return: float / str (if getString is True)
    """
    
    framesRateInSecondStr = mc.currentUnit( q = True, time = True )
    framesRateInSecondOptions = { 'game':15, 'film':24, 'pal':25, 'ntsc':30, 'show':48, 'palf':50, 'ntscf':60 }
    if 'fps' in framesRateInSecondStr:
        
        fpsStr = framesRateInSecondStr.replace( 'fps', '' )
        framesRateInSecond = float( fpsStr )
        
    else:
        
        framesRateInSecond = framesRateInSecondOptions[framesRateInSecondStr]
    
    
    if getString:
        
        return framesRateInSecondStr
    
    else:
        
        return framesRateInSecond


def scaleAnimCurveHeight ( animCurve, animScale ):

    '''
    stretching output values of animCurve
    @param animCurve: str, name of anim curve
    @param animScale: float, scale value

    '''

    mc.selectKey( animCurve, add = 1, k = 1 )

    mc.scaleKey( fp = 0, fs = animScale, iub = False, ts = 1, tp = 0, vp = 0, vs = animScale, animation = 'keys' )
    mc.selectKey( clear = 1 )

def scaleAnimCurveLength ( animCurve, animScale ):

    '''
    stretching time of animCurve
    @param animCurve: str, name of anim curve
    @param animScale: float, scale value

    '''

    mc.selectKey( animCurve, add = 1, k = 1 )

    mc.scaleKey( fp = 0, fs = animScale, iub = False, ts = 2.119999, tp = 1, vp = 0, vs = 1, animation = 'keys' )
    mc.selectKey( clear = 1 )


def scaleAnimCurvesHeight ( animCurves, animScale ):

    '''
    stretching output values of animCurves
    @param animCurves: list(str), list of anim curves
    @param animScale: float, scale value

    '''

    for animCurve in animCurves:

        scaleAnimCurveHeight ( animCurve, animScale )
        
def scaleAllHeightTLByValue ( value ):

    '''
    stretching output values of all translate anim curves in scene
    @param value: float, scale value

    '''

    allTranslateACs = mc.ls ( type = 'animCurveTL' )

    scaleAnimCurvesHeight ( allTranslateACs, value )

def scaleAllLengthByValue ( value ):

    '''
    stretching time of all anim curves in scene
    @param value: float, scale value

    '''

    animCurves = mc.ls ( type = 'animCurveTL' )
    animCurves.extend ( mc.ls ( type = 'animCurveTA' ) )
    animCurves.extend ( mc.ls ( type = 'animCurveTT' ) )
    animCurves.extend ( mc.ls ( type = 'animCurveTU' ) )

    for animCurve in animCurves:

        scaleAnimCurveLength ( animCurve, value )

