'''
blendShape at rigitAnimTools.utils
@category Rigging @subcategory Utils
@tags shapes blendshape morph
@author jakub
'''

import maya.cmds as mc


def getBlendShapeNodes( deformedObj ):
    
    """
    get all blendShape nodes from mesh
    
    @param deformedObj: deformed geometry to look for blendShapes
    @type deformedObj: str
    @return: list(str), list of blendShape nodes found connected to provided object
    """
    
    hist = mc.listHistory( deformedObj )
    blendShapes = mc.ls( hist, type = 'blendShape' )
    
    return blendShapes

def getGeoFromBlendShape( bsNode ):
    
    """
    get geometry transform object deformed by given blendShape node
    
    @param bsNode: str, name of blendShape node
    @return: str, name of geometry transform which has shape deformed by this blendShape, if nothing found returns None
    """
    
    his = mc.listHistory( bsNode, f = 1 )
    shapeTypes = ['mesh', 'nurbsSurface', 'nurbsCurve', 'lattice']
    geoObject = None
    
    for h in his:
        
        if mc.nodeType( h ) in shapeTypes:
            
            geoObject = h
            break
        
    return geoObject

def getTargetNames( bsnode ):
    
    '''
    simple querying aliases of target weight attributes
    
    @param bsnode: str, source blendShape
    @return: list(str), list of target names
    '''
    
    weightIndeces = getTargetIndeces( bsnode )
    
    atnames = [ mc.aliasAttr( '%s.weight[%d]' % ( bsnode, i ), q = 1 ) for i in weightIndeces ]
    
    return atnames

def getTargetIndeces( bsnode ):
    
    """
    get list of target indeces on blendShape node
    NOTE: returns all valid items even from sparse arrays
    
    @param bsnode: str, source blendShape
    @return: list(str), list of target indeces
    """
    
    targetIndeces = mc.getAttr( bsnode + '.weight', mi = 1 )
    
    return targetIndeces

def simpleExtract( bsnode, parentgroup = 'extractedShapes_grp', hideParentGrp = True, zeroAllBefore= True ):
    
    '''
    Extract geometry from given blendShape without option to use inbetweens
    Parents new shapes under new group

    NOTE: for reconnecting New shapes, use returned lists: snapshotObjectList, shapeDriverPlugs
          - because this is usually done after new blendshape is made
          - note that some plugs can be None, so perform condition before connecting
    
    @param bsnode: str, blendShape node
    @param parentgroup: str, if parent group is provided, new shapes will be parented under it, otherwise new group is made
    @param hideParentGrp: bool, hide extracted shapes parent group to keep memory load lower
    @param zeroAllBefore: bool, zero all targets before extracting, by default it does not change values which is handy for baking in changes to targets
    @return: list(str), index 0 - name of new shapes group, indeces 1-n - names of generated shape objects
    '''
    
    shapeDuplicateObj = getGeoFromBlendShape( bsnode )
    bsNames = getTargetNames( bsnode )
    
    if not mc.objExists( parentgroup ):
        parentgroup = mc.group( n = parentgroup, em = 1 )
    
    # hide parent group - this keeps memory smaller
    if hideParentGrp:
        mc.hide( parentgroup )
    
    shapeDriverPlugs = []
    for bsn in bsNames:
        
        # check input connection
        plugs = mc.listConnections( bsnode + '.' + bsn, d=False, s=True, p=True )
        if plugs:
            shapeDriverPlugs.append( plugs[0] )
            mc.disconnectAttr( plugs[0], bsnode + '.' + bsn )
        else:
            shapeDriverPlugs.append( None )
    
    snapshotObjectList = []
    
    for tgIdx, bsn in enumerate( bsNames ):
        
        # skip empty target
        if not bsn:
            continue
        
        origval = mc.getAttr( bsnode + '.' + bsn )
        
        # extract target shapes
        mc.setAttr( '%s.%s' % ( bsnode, bsNames[tgIdx] ), 1 )
        
        shapeObj = mc.duplicate( shapeDuplicateObj, n = bsNames[tgIdx] )[0]
        mc.parent( shapeObj, parentgroup )
        snapshotObjectList.append( shapeObj )
        
        if zeroAllBefore:
            
            mc.setAttr( bsnode + '.' + bsn, 0 )
        
        else:
            
            mc.setAttr( bsnode + '.' + bsn, origval )
    
    # reconnect original shapes
    for plug, bsn in zip( shapeDriverPlugs, bsNames ):

        if plug:
            mc.connectAttr( plug, bsnode + '.' + bsn )
    
    # assign default lambert shader to extracted shapes
    mc.sets( snapshotObjectList, e = True, forceElement = 'initialShadingGroup' )

    # unlock all attributes
    for o in snapshotObjectList:

        mc.lockNode( o, l = 0 )
        
        for channel in ['t', 'r', 's']:
            for axis in ['x', 'y', 'z']:
                mc.setAttr( o + '.' + channel + axis, l = 0, k = 1 )
        
        mc.setAttr( o + '.v', l = 0, k = 1 )
    
    return [parentgroup, snapshotObjectList, shapeDriverPlugs]