'''
functions for working with curves (nurbsCurves and bezierCurves) at rigit.utils
@category Rigging @subcategory Utils
@tags curve nurbs bezier
@author jakub
'''

import maya.cmds as mc

from . import name


def from2objects( objectA, objectB, prefix = '', spans = 1, degree = 3 ):
    
    '''
    make curve between 2 objects
    
    @param objectA: first object to attach curve to
    @type objectA: str
    @param objectB: second object to attach curve to
    @type objectB: str
    @param prefix: prefix for new created nodes
    @type prefix: str
    @param spans: number of spans for new curve
    @type spans: int
    @param degree: degree of new curve
    @type degree: int
    @return: str, name of new curve
    '''
    
    nullA = mc.group( n = 'nullA_from2objects_grp', em = 1 )
    nullB = mc.group( n = 'nullB_from2objects_grp', em = 1 )
    
    mc.delete( mc.pointConstraint( objectA, nullA ) )
    mc.delete( mc.pointConstraint( objectB, nullB ) )
    
    posA = mc.xform( nullA, q = 1, t = 1, ws = 1 )
    posB = mc.xform( nullB, q = 1, t = 1, ws = 1 )
    
    mc.delete( nullA, nullB )
    
    line = from2positions( posA, posB, prefix = prefix, spans = spans, degree = degree )
    
    return line

def from2positions( posA, posB, prefix = 'new', spans = 1, degree = 3 ):
    
    '''
    make curve between 2 points in space
    
    @param posA: first position to build curve from
    @type posA: list( float, float, float )
    @param posB: second position to build curve from
    @type posB: list( float, float, float )
    @param prefix: prefix for new created nodes
    @type prefix: str
    @param spans: number of spans for new curve
    @type spans: int
    @param degree: degree of new curve
    @type degree: int
    @return: str, name of new curve
    '''
    
    crv = mc.curve( n = prefix + '_crv', p = ( posA, posB ), d = 1 )
    mc.rebuildCurve( crv, keepRange = False, spans = spans, degree = degree )
    
    return crv


def makeConnectionLine( objectA, objectB, prefix = '', overrideMode = 1, curveParent = '' ):
    
    '''
    @param objectA: first object to attach curve to
    @type objectA: str
    @param objectB: second object to attach curve to
    @type objectB: str
    @param prefix: prefix for new created nodes
    @type prefix: str
    @param overrideMode: turn on overrideDisplay for new curve and use this number for its enum type (normal, template, reference)
    @type overrideMode: int (valid values: 0,1,2)
    @param curveParent: optional, name of parent for created connection curve
    @type curveParent: str
    @return: str, name of the curve
    '''
    
    objABase = name.removeSuffix( objectA )
    objBBase = name.removeSuffix( objectB )
    
    if prefix == '': prefix = objABase + '_' + objBBase
    
    line = from2objects( objectA, objectB, prefix = prefix, spans = 1, degree = 1 )
    
    mc.cluster( line + '.cv[0]', n = prefix + 'A_cluster', wn = ( objectA, objectA ), bs = 1 )
    mc.cluster( line + '.cv[1]', n = prefix + 'B_cluster', wn = ( objectB, objectB ), bs = 1 )
    
    lineShape = mc.listRelatives( line, s = 1 )[0]
    mc.setAttr( line + '.it', 0, l = 1 )
    mc.setAttr( line + '.ove', 1 )
    mc.setAttr( line + '.ovdt', overrideMode )
    
    # parent curve
    
    if mc.objExists( curveParent ):
        
        mc.parent( line, curveParent )
    
    
    return line

