'''
filereference at rigit.utils
@category Rigging @subcategory Utils
@tags file reference proxy
@author: jakub
'''
from __future__ import print_function

import maya.cmds as mc
import maya.mel as mm

from . import name

def isReference(obj):

    """
    test if object is reference
    """

    isReference = True

    try:
        namespaceName = mc.referenceQuery( obj, namespace = 1 )

    except:
        isReference = False
    
    return isReference


def importAllReferencesInScene( verbose = True, removeNamespace = False ):

    '''
    this will find all references in scene and import them,
    basically baking any existing

    @param verbose: print important warning and notes during function runtime
    @param verbose: str
    @param removeNamespace: optional, remove namespace from objects name
    @type removeNamespace: bool
    '''

    allRefNodes = mc.ls( type = 'reference' )

    for refNode in allRefNodes:

        try:

            namespaceName = mc.referenceQuery( refNode, namespace = 1 )
            fileName = mc.referenceQuery( refNode, filename = 1 )

        except:

            continue

        importReferenceToScene( namespaceName = namespaceName, removeNamespace = removeNamespace, verbose = verbose )


def removeReferenceFromScene( namespaceName = '' ):

    '''
    remove reference from scene including removing its namespace

    NOTE - function does not removed nodes parented under reference or connected to reference,
            if this is not done before, operation will result in creation of Foster Parent group

    @param namespaceName: remove reference based on given namespace
    @type namespaceName: str
    @return: None
    '''

    if not namespaceName:

        return

    namespaceNameColon = name.namespaceAddColon( namespaceName )
    refObjects = mc.ls( namespaceNameColon + '*', o = 1 )

    if not refObjects:

        mc.namespace( rm = namespaceName )


    refFilename = mc.referenceQuery( refObjects[0], f = 1 )
    referenceNode = mc.referenceQuery( refFilename, referenceNode = 1 )
    mc.file( removeReference = 1, referenceNode = referenceNode )



def importReferenceToScene( namespaceName = '', removeNamespace = False, verbose = True ):

    '''
    import reference to scene

    @param namespaceName: str, remove reference based on given namespace
    @param removeNamespace: bool, optional, remove namespace from objects name
    @param verbose: str, print important warning and notes during function runtime
    @return: None
    '''

    if not namespaceName:

        return

    namespaceNameColon = name.namespaceAddColon( namespaceName )
    refObjects = mc.ls( namespaceNameColon + '*', o = 1 )

    if not refObjects:

        mc.namespace( rm = namespaceName )

    # check if this is from referenced file
    if not mc.referenceQuery(refObjects[0], isNodeReferenced= True):
        return

    refFilename = mc.referenceQuery( refObjects[0], f = 1 )
    refNode = mc.referenceQuery( refFilename, referenceNode = 1 )
    if verbose:
        print('refFilename: {} , refNode: {}'.format( refFilename, refNode ))
    mc.file( referenceNode = refNode, importReference = 1 )

    # remove namespace from object names

    if removeNamespace:
        namespaceList = mc.ls( namespaceNameColon + '*', objectsOnly = True, recursive = True )

        for node in namespaceList:
            mc.lockNode( node, l = 0 )

        mc.namespace( mv = [namespaceName, ':'], force = True )
        mc.namespace( rm = namespaceName )

    if verbose:

        print('# imported reference with namespace %s from file %s' % ( namespaceName, refFilename ))

def removeAllNamespacesInScene( onlyWithoutRef = False, verbose = True ):
    
    """
    remove all namespaces in scene
    
    @param onlyWithoutRef: bool, remove only local namespaces without reference file connected
    @param verbose: bool, print important information during the function runtime
    
    NOTE: this will first import all references in scene
    """
    
    if not onlyWithoutRef:
        
        importAllReferencesInScene( removeNamespace = True )
    
    allNs = mc.namespaceInfo( listOnlyNamespaces = True, recurse = True )
    allNs.reverse()
    
    for ns in allNs:
        
        if not ns in ['UI', 'shared']:
            
            try:
                mc.namespace( rm = ns, mergeNamespaceWithRoot = True )
                
            except:
                if verbose:
                    print('# problem removing this namespace: %s', ns)
                
                pass
        
    

def makeProxy( namespaceName, filepathProxies, proxyTags = ['anim', 'render'] ):

    '''
    make proxy from files
    NOTE:
    Very useful for switching between more rig LODs with the same controls and different geometry, deform setups etc.
    Number of items in filepathProxies and proxyTags has to be the same
    
    @param namespaceName: str, namespace name
    @param filepathProxies: list( str, str, str),  list of filepaths to be under one reference as proxies
    @param proxyTags: list(str), tags to be added to proxies
    @return: str, reference node made by Maya
    '''
    
    # =====================================
    # create proxy with shared edits
    # =====================================

    # make sure Shared Edits is ON
    mc.optionVar( intValue = ( 'proxyOptionsSharedEdits', 1 ) )

    # make reference (first proxy)
    refFile = mc.file( filepathProxies[0], reference = 1, namespace = namespaceName )
    refNode_a = mc.referenceQuery( refFile, referenceNode = 1 )

    # add second proxy
    proxyAddMelCmd = 'proxyAdd "%s" "%s" "%s";' % ( refNode_a, filepathProxies[1], proxyTags[1] )
    mm.eval( proxyAddMelCmd )

    # add third proxy
    # proxyAddMelCmd = 'proxyAdd "%s" "%s" "%s";' % ( refNode_a, filepathProxies[2], proxyTags[2] )
    # mm.eval( proxyAddMelCmd )
    
    # fix first proxy names
    # needs to be done after adding proxies or they would inherit first name too
    mc.setAttr( refNode_a + '.proxyTag', proxyTags[0], typ = 'string' )
    mc.lockNode( refNode_a, l = 0 )
    
    # rename proxy nodes
    refNodes = getAllProxiesFromRefNode( refNode_a )
    for i in range( len( refNodes ) ):
        
        print('refNodes[i]', refNodes[i])
        mc.lockNode( refNodes[i], l = False )
        refNodes[i] = mc.rename( refNodes[i], namespaceName + proxyTags[i] + '_rfn' )
        
    # make sure we are in default namespace
    mc.namespace( set = ':' )

    return refNodes


def switchProxy( fileRefNode ):

    '''
    switch to proxy of referenced file specified by file reference node
    '''

    mm.eval( 'proxySwitch %s' % fileRefNode )


def getAllProxiesFromRefNode( fileRefNode ):

    '''
    Get all proxy file reference nodes from one reference node
    NOTE: Using connection to proxy manager
    '''

    proxyManagerNode = mc.listConnections( fileRefNode + '.proxyMsg', t = 'proxyManager' )[0]
    allProxies = mc.listConnections( proxyManagerNode + '.proxyList', s = 1, t = 'reference' )

    return allProxies

def getCurrentProxy( proxyManagerNode ):

    '''
    get current active proxy from provided proxyManage node
    '''

    currentProxyNode = mc.listConnections( proxyManagerNode + '.sharedEditsOwner', s = 1, t = 'reference' )[0]

    return currentProxyNode



