'''
name in rigit.utils
@category Rigging @subcategory Utils
@tags name rename prefix suffix side
@author jakub
'''
from __future__ import print_function

import re

import maya.cmds as mc


def makeBaseName( side, name ):
    
    baseName = name
    if side != '': baseName = side + '_' + name
    
    return baseName

    
def short( name ):
  
  return name.split( '|' )[-1]
  
def removeSuffix( name ):
    
    name = short( name )
    
    if name.count( '_' ) == 0: return name
    
    if name.count( '_' ) == 1:
        if name.startswith( 'l_' ) or name.startswith( 'r_' ):
            return name
            
    edit = name.split( '_' )[-1]
    edit = name[0: len( name ) - len( edit ) - 1 ]
    
    return edit

def getSuffix( name ):
    
    name = short( name )
    
    if name.count( '_' ) == 0:
        
        return ''
    
    edit = name.split( '_' )[-1]
    
    
    return edit

def getPrefix( name, min3NameParts = False ):
    
    name = short( name )
    
    if name.count( '_' ) == 0:
        
        return name
    
    res = name.split( '_' )
    
    if min3NameParts and len( res ) > 2:
        
        return res[0]
    
    if not min3NameParts:
        
        return res[0]
    
    return ''


def getBase( name, keepSide = True ):
    
    edit = name.split( '|' )[-1]
    edit = removeEndNumbers( edit )
    
    if name.count( '_' ):
        
        edit = removeSuffix( edit )
        editLowerCase = edit.lower()
        edit = removeEndNumbers( edit )
        
        if editLowerCase.startswith( 'l_' ) or editLowerCase.startswith( 'r_' ):
            
            if not keepSide:
                
                return edit[2:]
    
    
    return edit



def getSide( name ):
    
    '''
    return side of object, from left 'l', from right 'r', from other empty string ''
    
    @param name: object name
    @type name: str
    @return: str, returns one of these: 'l','r',''
    '''
    
    edit = name.split( '|' )[-1]
    edit = removeNamespace( edit )
    
    # make sure side is lower case for naming conventions
    edit = edit.lower()
    
    if edit.startswith( 'l_' ): return 'l'
    if edit.startswith( 'r_' ): return 'r'
    
    return ''

def flipSide( name ):
    
    namespacePart = getNamespace( name )
    if namespacePart == ':': namespacePart = ''
    cleanName = removeNamespace( name )
    
    side = getSide( name )
    
    if not side == 'l' and not side == 'r':
        
        return name
    
    if side == 'l':
        
        return namespacePart + 'r_' + cleanName[2:]
    
    if side == 'r':
        
        return namespacePart + 'l_' + cleanName[2:]
    

def getSideAsSign( name ):
    
    '''
    @param name: object`s name
    @type name: str
    @return: int, valid values: -1,0,1
    
    return 0 for center, 1 for left and -1 for right
    '''
    
    editStr = name.lower()
    
    if len( name ) > 1:
        
        editStr = getSide( name )
    
    if editStr == 'l': return 1
    if editStr == 'r': return -1
    
    return 0
    

def removeSide( name ):
    
    lside = 'l_'
    rside = 'r_'
    midside = 'c_'
    
    edit = name
    
    if name[:2] in [ lside, rside, midside ]:
        
        edit = name[2:]
    
    return edit

def namespaceAddColon( namespace ):
    
    if namespace != '' and not namespace.endswith( ':' ):
        
        namespace = namespace + ':'
    
    return namespace

def namespaceRemoveColon( namespace ):
    
    if namespace != '' and namespace.endswith( ':' ):
        
        return namespace[:-1]
    
    return namespace


   
def getNamespace( object ):
    
    parts = object.split( ':' )
    if len( parts ) < 2: return ''
    
    ns = ':'.join( parts[ 0:-1 ] )
    
    if ns == ':': ns = ''
    
    return ns

def removeNamespace( name ):
    
    if not ':' in name: return name
    
    edit = name.split( ':' )[-1]
    
    return edit


def removePrefix( name, prefix = '', customPrefix = False ):
    
    name = short( name )
    
    if not customPrefix:

        if name.count( '_' ) == 0: return name

        edit = name.split( '_' )[0]
        edit = name[ len( edit ) + 1 : ]

    else:

        prefixLen = len( prefix )
        edit = name[ prefixLen: ]

    return edit
    
def removeEndNumbers( name ):
    
    '''
    @param name: string to strip off numbers off the end
    @type name: str
    @return: str
    
    remove all numbers at the end of the name, e.g. 'l_leg1' --> 'l_leg'
    '''
    
    pattern = re.compile( '[0-9]*$' )
    res = re.search( pattern, name )
    endlen = len( res.group( 0 ) )
    
    if not endlen: return name
    else: return name[ 0:len( name ) - endlen ]


def removeFromString( string, remove = 'digit' ):
    
    '''
    remove digit, alphabetic or space characters
    
    @param string: 
    @type string: str
    @param remove: choose 'digit', 'alpha' or 'space'
    @type remove: str 
    @param verbose: print messages
    @type verbose: boolean
    @return str: string with removed characters
    
    @author: juan
    '''
    
    if remove == 'digit':
        string = re.sub( '[0-9^]', '', string )
    elif remove == 'alpha':
        string = re.sub( '[^0-9]', '', string )
    elif remove == 'space':
        string = re.sub( '[ ]', '', string )
    else:
        mc.error( 'remove param must be `digit`, `alpha` or `space`' )
            
    return string

    
def returnEndNumbers( name ):
    
    '''
    @param name: string to strip off numbers off the end
    @type name: str
    @return: str
    
    return end numbers in the name, for example c_cube_geo10 will return 10 
    '''
    
    pattern = re.compile( '[0-9]*$' )
    res = re.search( pattern, name )
    return res.group( 0 )
    

def findNextIncrement( prefix, suffix = '_ctl' ):
    
    '''
    @param prefix: prefix before increment
    @type prefix: str
    @param suffix: suffix to isolate increment
    @type suffix: str
    
    find next increment for a numbered module
    
    EXAMPLE:
    
    making cube1_ctrl, cube2_ctrl etc.
    run this: findNextIncrement( 'cube', '_ctrl' )
    this will find that next number needs to be 3
    
    NOTE: this function should not get full name of object from scene
    '''
    
    nameUnique = False
    nextNumberStr = '1'
    objshort = prefix + nextNumberStr + suffix
    endNumberInt = 1
    
    while not nameUnique:
        
        # perform unique name test
        objsFound = mc.ls( objshort )
        objsFoundLen = len( objsFound )
        
        if objsFoundLen == 0:
            
            nameUnique = True
        
        else:
            
            endNumberInt += 1
            nextNumberStr = '{}'.format(endNumberInt)
            objshort = prefix + nextNumberStr + suffix
        
        
    return nextNumberStr

def newNameWithNextIncrement( objname, autofindSuffix = True, suffix = '' ):
    
    """
    find new name for object based on next increment number
    NOTE: this is way of renaming to prevent clashing names
    
    @param objname: str, name of object which can be both long or short path
    @param autofindSuffix: bool, find if object has suffix automaticly unless it is specified in suffix parameter
    @param suffix: str, optional, user defined suffix that should be part of objname string end, otherwise it will be ignored
    @return: str, new name for object 
    """
    
    objnameShort = short( objname )
    
    if autofindSuffix:
        
        suffix = getSuffix( objnameShort )
    
    elif not objnameShort.endswith( suffix ):
        
        suffix = ''
        
    suffixUndersc = suffix
    if not suffixUndersc.startswith( '_' ):
        
        suffixUndersc = '_' + suffixUndersc
    
    nosuffix = removeSuffix( objnameShort )
    nosuffixEndNumbers = removeEndNumbers( nosuffix )
    nextIncrem = findNextIncrement( prefix = nosuffixEndNumbers, suffix = suffixUndersc )
    
    # make new name
    newname = nosuffixEndNumbers + nextIncrem + suffixUndersc
    
    return newname
    

def getAvailableNamespace( namespaceName ):
    
    '''
    this function will check given namespace and return the same if there is no such namespace,
    but in case this namespace already exists in scene, it will increment +1 until this is unique
    
    NOTE: this works locally in any current namespace, which can be under other namespace then root ":",
          it is user or function to set the local namespace before this function is used
    '''
    
    allNsList = mc.namespaceInfo( ls = 1 )
    
    availableNamespace = namespaceName
    
    # remove numbers from end of namespace for incrementing
    
    namespaceNameNoNums = removeEndNumbers( namespaceName )
    counter = 1
    whatIfCount = 1000
    
    while availableNamespace in allNsList:
        
        edit = namespaceNameNoNums + '_%03d' % counter
        availableNamespace = edit
        
        if counter > whatIfCount:
            
            break
    
    return availableNamespace
    
    

def renameSimple( objlist, search, replace, verbose = True ):
    
    '''
    super simple rename, doesn`t bother to find object by long name or return list of renamed objects
    
    @param objlist: list of objects to rename
    @type objlist: list(str)
    @param search: search string
    @type search: str
    @param replace: replace string
    @type replace: str
    @return: list(str), list of renamed names
    '''
    
    renamedlist = []
    
    for o in objlist:
        
        if mc.objExists( o ) == 1:
            
            short = o.split( '|' )[-1]
            newName = short.replace( search, replace )
            
            longName = mc.ls( o, l = True )[0]
            
            # unlock node            
            lockedState = mc.lockNode( longName, q = 1, l = 1 )[0]
            
            if lockedState:
                
                mc.lockNode( longName, l = 0 )
            
            try:
                
                mc.rename( longName, newName )
            
            except:
                
                if verbose:
                    
                    print('# %s could not be renamed, skipping ...' % o)
                    
                continue
            
            # lock node if it was locked
            if lockedState:
                
                mc.lockNode( newName, l = 1 )
            
            renamedlist.append( newName )
    
    return renamedlist

def renameAllLowerCase( objlist ):
    
    """
    rename list of objects to lower case version of their name
    
    @param objlist: list(str), list of object names to be renamed
    @return: list(str), list of new names
    """
    
    newNames = []
    
    for o in objlist:
        
        newname = o.lower()
        mc.rename( o, newname )
        newNames.append( o )
        
    return newNames
    


    
def capitalizeFirst( name ):
    
    '''
    capitalize first letter of given string
    
    @param name: string to process
    @type name: str
    @return: str
    '''
    
    firstletter = name[0].upper()
    suffix = ''
    if len( name ) > 1: suffix = name[1:]
    
    resultstring = firstletter + suffix
    
    return resultstring

def lowercaseFirst( name ):
    
    '''
    change first letter of given string to lower case
    
    @param name: string to process
    @type name: str
    @return: str
    '''
    
    firstletter = name[0].lower()
    suffix = ''
    if len( name ) > 1: suffix = name[1:]
    
    resultstring = firstletter + suffix
    
    return resultstring


def sortAlphanumericList( list ):
    
    '''
    given a list of type l = ['alpha2','alpha34','alpha86','alpha1'] sorted by traling number
    
    @param list: list to re-oder
    @type list: list(str)
    @return: list(str), re-odered list
    
    @notes: beware, if list contains non matching alpha characters on items, it will be ordered alphabetically \
    for example list of type = ['alphaA_2', 'alphaC_1','alphaE_5','alphaB_3','alphaD_4']
    '''
    
    checkText = lambda text: int( text ) if text.isdigit() else text
    
    alphanumeric = lambda key: [ checkText( t ) for t in re.split( '([0-9]+)', key ) ]
    
    return sorted( list, key = alphanumeric )
    
def getClashingNamesInScene():
    
    """
    return list of clashing names in scene,
    if no names are clashing, empty list will be returned
    
    @return: list( str ), list of clashing names
    """
    
    clashing = [ o for o in mc.ls() if o.count( '|' ) ]
    
    return clashing
    

def simpleFixClashingNamesInScene( skipObjects = None, keepSuffix = True ):
    
    """
    simple mechanism to fix clashing names in scene    
    This script just adds a number at the end of the object to make it unique
    
    NOTE: ideal for fixing technical nodes where the change of name will not break anything
    WARNING: this is not suitable for scenes where objects need to keep their names, do not use in Rig scenes
    
    @param skipObjects: list(str), list of objects which should be skipped, their short name will be used
    @param keepSuffix: bool, place number before suffix if object has it in the name
    @param clashingSuffix: str, string to be placed at the end of name in case there is no suffix or keepSuffix is False
    @return: list(str), list of new object names from renamed clashing objects
    """
    
    clashing = getClashingNamesInScene()
    clashingUuids = [ mc.ls( c, uuid = True )[0] for c in clashing ]
    
    
    # convert skip names to short names
    skipObjectsShort = []
    
    if skipObjects:
        
        skipObjectsShort = [ short( n ) for n in skipObjects ]
    
    newNames = []
    
    
    for cid in clashingUuids:
        
        obj = mc.ls( cid, uuid = True )[0]
        shortname = short( obj )
        
        if shortname in skipObjectsShort:
            
            continue
        
        newname = newNameWithNextIncrement( shortname, autofindSuffix = True, suffix = '' )
        
        if keepSuffix:
            
            nosuffix = removeSuffix( shortname )
            
            if not nosuffix == shortname:
                
                suffix = getSuffix( shortname )
                increm = findNextIncrement( nosuffix, suffix = '_' + suffix )
                newname = nosuffix + increm + '_' + suffix
            
            
        
        mc.lockNode( obj, lock = False )
        renamedObj = mc.rename( obj, newname )
        newNames.append( renamedObj )
    
    return newNames

    
    
