'''
shape module at rigit.utils
@category Rigging @subcategory Utils
@tags shape curve mesh nurbs polygon geometry
'''

import maya.cmds as mc

from . import name


def getShapes( object, useLongName = False, noIntermediate = True, shapeTypes = [] ):
    
    '''
    get shapes from passed object, return shape if shape was already passed
    
    @param object: object to search for shape
    @type object: str
    @param useLongName: if true, function will return long shape name (shapes are more prone to clashing names)
    @type useLongName: bool
    @param shapeTypes: list(str), types of shapes, if list is empty then all shapes will be returned
    @return: list(str), names of shapes, return [] if no shape was found
    '''
    
    # return object if it is a shape type
    
    objecttype = mc.nodeType( object )
    returnname = object
    returnnameLong = mc.ls( object, l = 1 )[0]
    
    if 'shape' in mc.nodeType( object, inherited = 1 ):
        
        if useLongName: return [returnnameLong]
        else: return [returnname]
    
    # return object shape if it was not shape itself
    
    if not shapeTypes:
        
        shapesfoundLong = mc.listRelatives( object, s = 1, f = 1, noIntermediate = noIntermediate )
    
    else:
        
        shapesfoundLong = mc.listRelatives( object, s = 1, f = 1, noIntermediate = noIntermediate, type = shapeTypes )
    
    if not shapesfoundLong: return []
    
    # get shape short names
    
    shapesfound = [ s.split( '|' )[-1] for s in shapesfoundLong ]
    
    
    if useLongName: return shapesfoundLong
    else: return shapesfound
    
def getType( object ):
    
    '''
    get shape type of provided object. In case transform was provided, get the type for its first shape
    
    @param object: object which can be transform or shape to get its shape type from
    @type object: str
    @return: str, name of shape type, if no shape found, return None 
    '''
    
    shapes = getShapes( object, useLongName = True )
    if not shapes: return None
    
    shapetype = mc.nodeType( shapes[0] )
    
    return shapetype

    
def scale( shapeobj = '', scale = [1, 1, 1], componentsList = [], translateRefObject = None, rotateRefObject = None ):
  
    '''
    scale shape components along its local pivot or provided reference objects
    for position and orientation
    
    @param shapeobj: object with shape or the shape itself, it doesnt have to be provided if componentList is provided
    @type shapeobj: str
    @param scale: scale to apply on the shape
    @type scale: list(float x, float y, float z), 3 values XYZ
    @param componentsList: optional, list of components to scale instead of shape object, this will override shape object
    @type componentsList: list(str), example: ['pSphere1.vtx[32:33]','pSphere1.vtx[51:54]']
    @param translateRefObject: optional, translate reference object
    @type translateRefObject: str
    @param rotateRefObject: optional, rotate reference object for scale axis orientation
    @type rotateRefObject: str
    @return: None
    '''
    
    # get shape from component list if provided
    
    if componentsList:
        
        shapeobj = mc.ls( componentsList[0], objectsOnly = True )[0]
    
    
    # or use shape parameter value
    
    else:
        
        compat = getComponentAttribute( shapeobj )
        componentsList = shapeobj + '.%s[*]' % compat
    
    
    # get translation
    
    pivotpos = [0, 0, 0]
    xformobj = shapeobj
    
    if not 'transform' in mc.nodeType( shapeobj, inherited = 1 ):
            
            xformobj = mc.listRelatives( shapeobj, p = 1 )[0]
    
    pivotpos = mc.xform( xformobj, q = 1, t = 1, ws = 1 )
    
    if translateRefObject:
        
        pivotpos = mc.xform( translateRefObject, q = 1, t = 1, ws = 1 )
    
    
    # get rotation
    
    rotationOffset = [0, 0, 0]
    
    if rotateRefObject:
        
        tempRotationGrp = mc.group( n = 'shapeScaleTempRot_grp', em = 1 )
        mc.delete( mc.parentConstraint( rotateRefObject, tempRotationGrp ) )
        rotationOffset = mc.getAttr( tempRotationGrp + '.r' )[0]
        mc.delete( tempRotationGrp )
    
    mc.scaleComponents( scale[0], scale[1], scale[2], componentsList, pivot = pivotpos, rotation = rotationOffset )
    

def getComponentAttribute( shapeobj ):
    
    '''
    return component attribute, for example 'cv' for nurbsCurve or 'vtx' for mesh
    
    @param shapeobj: object with shape or the shape itself
    @type shapeobj: str
    @return: str, name of component attribute
    '''
    
    compmaskdr = { 'mesh':'vtx', 'nurbsCurve':'cv', 'nurbsSurface':'cv', 'lattice':'pt' }
    shapenode = getShapes( shapeobj )[0]
    shapetype = mc.nodeType( shapenode )
    
    if not shapetype in list(compmaskdr.keys()): raise Exception( 'node type not supported' )
    
    compmask = compmaskdr[ shapetype ]
    
    return compmask


def getOrigShape( object, useLongName = False ):
    
    '''
    find orig shape under the object transform, those connected to deformers are preferred
    
    @param object: object to search for shape
    @type object: str
    @param useLongName: if true, function will return long shape name (shapes are more prone to clashing names)
    @type useLongName: bool
    @return: str, name of orig shape
    '''
    
    # check if this is shape
    
    nodeType = mc.nodeType( object )
    
    if not nodeType == 'transform':
        
        object = mc.listRelatives( object, p = 1 )[0]
    
    # get shapes under the transform
    
    shapesFound = mc.listRelatives( object, s = 1, fullPath = useLongName )
    
    if not shapesFound:
        
        return
    
    origShapes = [ s for s in shapesFound if mc.getAttr( s + '.intermediateObject' ) ]
    
    if not origShapes:
        
        return
    
    # now pick up orig shape connected to deformers
    # if no such found, just return any orig shape 
    
    deformOrigShapes = []
    
    for s in origShapes:
        
        h = mc.listHistory( s, future = 1 )
        if mc.ls( h, type = 'geometryFilter' ):
            
            deformOrigShapes.append( s )
    
    returnShape = origShapes[0]
    
    if deformOrigShapes:
        
        returnShape = deformOrigShapes[0]
    
    # get short name if requested
    
    if not useLongName:
        
        returnShape = name.short( returnShape )
    
    return returnShape

def fixNames( shapeTransform ):
    
    '''
    fix shape names, so they are named based on its parent
    as Maya often doesn`t name shapes correctly, e.g. when making/renaming a curve
        
    @param shapeTransform: name of transform object with shape(s)
    @type shapeTransform: str
    @return: list(str), list of new shape names
    '''
    
    shapesFound = mc.listRelatives( shapeTransform, s = 1, f = 1 )
    if not shapesFound:
        return []
    
    # first rename to something else
    # in case the prefix was already used on some shapes
    
    neutralPrefix = 'fixShapeNamesTemporaryName#'
    
    for s in shapesFound: mc.rename( s, neutralPrefix )
    shapesFound = mc.listRelatives( shapeTransform, s = 1, f = 1 )
    
    
    # rename shapes to temporary names first
    
    newShapeNames = []
    
    for i, s in enumerate( shapesFound ):
        
        renamed = mc.rename( s, 'tempShapeName#' )
    
    # rename shapes in the same way as Maya does it
    
    shapesFound = mc.listRelatives( shapeTransform, s = 1, f = 1 )
    
    for i, s in enumerate( shapesFound ):
        
        shapeTransformShort = name.short( shapeTransform )
        newName = shapeTransformShort + 'Shape'
        if i > 0: newName = shapeTransformShort + 'Shape' + str( i )
        renamed = mc.rename( s, newName )
        newShapeNames.append( renamed )
    
    return newShapeNames

def fixNamesAllScene( verbose= True ):

    """
    fix all shape names in scene  to be unique and avoid clashing (duplicate names)
    """

    allTransforms = mc.ls(type='transform')

    rangeX = range(3)
    for i in rangeX:
        for t in allTransforms:
            try: fixNames( t )
            except:
                if verbose:
                    print('# skipped %s - shape probably locked' % t)
                pass
    
def getAllGeosUnderObject( topObject, geoTypes = ['mesh'], longNames = False ):

    """
    get all geometry transforms under object, usually group, with selected types(s)

    @param topObject: str, name of parent object or group
    @param geoType: list(str), types of geometry shape to list
    @param longNames: bool, list objects with long names for scenes where short names would be clashing
    @return: list(str), list of model objects
    """

    assetModelGroup = topObject
    
    allGeos = [ mc.listRelatives(m, p=True, fullPath= longNames)[0] for m in mc.listRelatives( assetModelGroup, ad=True, type= geoTypes, fullPath= longNames ) ]
    allGeos = list( set(allGeos) )

    return allGeos





