'''
functions supporting skinCluster at rigitAnimTools.utils
@category Rigging @subcategory Utils
@tags skin skinning skinCluster sc copy weights
@author jakub
'''

import maya.cmds as mc
import sys
import os.path
from . import blendShape
from . import name
from . import mesh
from .. import maya_version_modules

def _loadSkinSmoothPlugin():

    """
    load tbSmooth plugin
    """

    modulesDir = os.path.split(maya_version_modules.__file__)[0]
    
    mayaVer = mc.about(v=True)
    pythonModulesDt = {'2020':'python2_07', '2022':'python3_07', '2023':'python3_09', '2024':'python3_10'}
    numpyFolder = '{}/python_modules/{}'.format(modulesDir, pythonModulesDt[mayaVer])
    pluginFolder = '{}/plugins/maya{}'.format(modulesDir, mayaVer)

    # add numpy folder to sys path
    if not numpyFolder in sys.path:
        sys.path.append(numpyFolder)
    
    tbSmoothPluginFile = 'thomasBittnerSmoothSkin.py'
    tbSmoothPluginFilepath = os.path.join(pluginFolder, tbSmoothPluginFile)
    if not mc.pluginInfo(tbSmoothPluginFile, q=1, l=1):
        mc.loadPlugin(tbSmoothPluginFilepath)



def getRelated( meshObj ):
    
    '''
    find related skinCluster if object is skinned
    
    @param shapeObj: str or list(str), shape or transform with shape or list with shapes from the same object
    @return: (str) skinCluster, if not found return None
    '''
    
    # get object shapes in case there was transform
    objShapes = mc.listRelatives( meshObj, s = 1, f = 1, noIntermediate = True )
    historyRes = mc.listHistory( objShapes, pruneDagObjects = True )
    sClusters = mc.ls( historyRes, type = 'skinCluster' )
    
    if sClusters:
        return sClusters[0]
    
def copySubdiv_delnondf( sourceMesh, subdivLevel= 1, subdUvBoundary= 2, verbose= False ):

    """
    simple method using Delete-Non-Deformers, but makes skinning spikes in some cases (mouth usually)

    @param sourceMesh: str, name of source mesh for transfering skinning and blendshapes on subdived copy
    @param subdivLevel: int, level of subdivision, default 1 is lowest subdivide level
    @param subdUvBoundary: int, openSubdiv UV boundary mode, 1-based, default 2: edges + corners
    @param verbose: int, print various function updates or warnings
    """

    forcedGenericUVs = False

    sourceSc = getRelated( sourceMesh )
    if not sourceSc:
        if verbose:
            print('# rigitAnimTools skincluster copySubdiv() - no skinCluster found on %s, skipping' % sourceMesh)
        return
    
    makeNewUVs = True # subdivUvMapMode == 2
    
    # subdiv value
    if subdivLevel < 1:
        subdivLevel = 1
    subdivVal = subdivLevel

    # subdiv UV boundary
    subdUvBoundaryVal = subdUvBoundary - 1
    
    # subdivide
    mc.polySmooth( sourceMesh, osdFvarBoundary=subdUvBoundaryVal, divisions=subdivVal, method=0, subdivisionType=2, osdVertBoundary=1, osdFvarPropagateCorners=0, osdSmoothTriangles=0, osdCreaseMethod=0, ch=True)
    
    # bake subdivisions and optional blendshapes - also post smoothing skin weights spikes
    # mc.bakePartialHistory(sourceMesh, prePostDeformers=True, postSmooth=True)
    # if name.short(sourceMesh) == 'body_geo':
    #     stop()
    mc.select(cl=True)
    # mc.refresh()
    mc.bakePartialHistory(sourceMesh, prePostDeformers=True, postSmooth=True)
    
    return [sourceMesh, forcedGenericUVs]


def copySubdiv_uvCopy( sourceMesh, subdivMeshPrefix = 'subdiv', replaceOrig = True, subdivLevel= 1, subdUvBoundary= 2, subdivUvMapMode= 1, verbose= False ):

    """
    subdivide duplicated mesh and copy skin weights using UV map 
    and also first blendShape and its target it finds
    NOTE: sourceMesh needs to have skinning but not UVs - these will be made on duplicate

    @param sourceMesh: str, name of source mesh for transfering skinning and blendshapes on subdived copy
    @param subdivMeshPrefix: str, prefix for new objects, '_' is also added between
    @param replaceOrig: int, replace original sourceMesh with new subdived hires mesh
    @param subdivLevel: int, level of subdivision, default 1 is lowest subdivide level
    @param subdUvBoundary: int, openSubdiv UV boundary mode, 1-based, default 2: edges + corners
    @param subdivUvMapMode: int, choice of UV map source to copy skin weights on subdivided mesh
    @param verbose: int, print various function updates or warnings
    """

    sourceSc = getRelated( sourceMesh )
    if not sourceSc:
        if verbose:
            print('# rigitAnimTools skincluster copySubdiv() - no skinCluster found on %s, skipping' % sourceMesh)
        return
    
    # load skin smooth plugin
    _loadSkinSmoothPlugin()

    makeNewUVs = True # subdivUvMapMode: 2
    forcedGenericUVs = False
    if subdivUvMapMode == 1:
        makeNewUVs = False

        # check UV map overlapping
        overlapRes = mesh.checkUVsOverlap( sourceMesh )
        if overlapRes:

            makeNewUVs = True
            forcedGenericUVs = True
    
    # subdiv value
    if subdivLevel < 1:
        subdivLevel = 1
    subdivVal = subdivLevel

    # subdiv UV boundary
    subdUvBoundaryVal = subdUvBoundary - 1
    
    # extract hires blendshape shapes
    bsnodeList = blendShape.getBlendShapeNodes( sourceMesh )
    sourceBsNode = None
    if bsnodeList:

        sourceBsNode = bsnodeList[0]
        parentgroup, snapshotObjectList, shapeDriverPlugs = blendShape.simpleExtract( 
                                                            bsnode= sourceBsNode, 
                                                            parentgroup = 'extractedShapes_grp', 
                                                            hideParentGrp = True, 
                                                            zeroAllBefore= True )
        for shapeObj in snapshotObjectList:
            mc.polySmooth( shapeObj, osdFvarBoundary=subdUvBoundaryVal, divisions=subdivVal, method=0, subdivisionType=2, osdVertBoundary=1, osdFvarPropagateCorners=0, osdSmoothTriangles=0, osdCreaseMethod=0, ch=False)

    # new UV map
    baseMesh = mc.duplicate( sourceMesh, n= 'temp_baseMesh_msh' )[0]
    baseMeshUVset = mc.polyUVSet(baseMesh, q=True, currentUVSet=True)[0]
    skinningUVset = baseMeshUVset

    if makeNewUVs:
        # >> NOTE - add new non-default map set for skinning copy, to keep existing mesh UVs
        skinningUVset = 'skincopy'
        mc.polyUVSet(baseMesh, create=True, uvSet= skinningUVset)
        mc.polyUVSet(baseMesh, currentLastUVSet=True)
        mc.polyAutoProjection(baseMesh, ch=0, lm=1, pb=0, ibd=1, cm=0, l=2, sc=2, o=0, p=12, ps=0.1, ws=False)
    
    # subdivide
    hiresMesh = mc.duplicate( baseMesh, n= '{}_{}'.format( subdivMeshPrefix, sourceMesh ) )[0]
    mc.polySmooth( hiresMesh, osdFvarBoundary=subdUvBoundaryVal, divisions=subdivVal, method=0, subdivisionType=2, osdVertBoundary=1, osdFvarPropagateCorners=0, osdSmoothTriangles=0, osdCreaseMethod=0, ch=False)

    # make hires blendshape
    if sourceBsNode:
        
        hiresBs = mc.blendShape( snapshotObjectList, hiresMesh, n= '%s_bls' % hiresMesh )[0]
        for plug, target in zip( shapeDriverPlugs, snapshotObjectList ):

            if plug:
                mc.connectAttr( plug, hiresBs + '.' + target )
    
    # skinning
    infs = mc.skinCluster(sourceSc, q=True, weightedInfluence= True) # get only influences with non-zero weights
    baseMeshSc = mc.skinCluster( baseMesh, infs, tsb= True )[0]
    hiresMeshSc = mc.skinCluster( hiresMesh, infs, tsb= True )[0]
    copySmooth = True
    mc.copySkinWeights(ss= sourceSc, ds= baseMeshSc, smooth= copySmooth, noMirror=True, surfaceAssociation= 'closestPoint', ia= ['label', 'name', 'oneToOne'], normalize=True)
    mc.copySkinWeights(ss= baseMeshSc, ds= hiresMeshSc, smooth= copySmooth, noMirror=True, surfaceAssociation= 'closestPoint', uvSpace= [skinningUVset, skinningUVset], ia= ['label', 'name', 'oneToOne'], normalize=True)

    # smooth skin weights on hires mesh
    origSel = mc.ls(sl=True)
    mc.select(hiresMesh + '.vtx[*]')
    mc.thomasBittnerSmoothSkin(iterations=2, ignoreJointLocks=False )
    mc.select(origSel)

    # cleanup
    mc.delete( baseMesh )
    if makeNewUVs:
        mc.polyUVSet( hiresMesh, delete=True, uvSet= skinningUVset)
    mc.polyUVSet( hiresMesh, currentUVSet=True,  uvSet= baseMeshUVset)

    if sourceBsNode:
        mc.delete(parentgroup)
    
    if replaceOrig:

        mc.delete(sourceMesh)
        sourceMeshShort = name.short(sourceMesh)
        hiresMesh = mc.rename( hiresMesh, sourceMeshShort )
    
    return [hiresMesh, forcedGenericUVs]








