'''
functions for working with surfaces (nurbsSurface) at rigit.utils
@category Rigging @subcategory Utils
@tags surfaceObj nurbs
@author jakub
'''

import maya.cmds as mc

from . import shape


def getClosestPointOnSurface( object, surfaceObj, normalizeCoords = False ):
    
    '''
    get closest point on surfaceObj, return both parameter and world position
    
    @param object: name of transform object
    @type object: str
    @param surfaceObj: name of nurbsSurface object
    @type surfaceObj: str
    @return: list( list( float, float ), list( float, float, float ) ) - first list has UV parameters and second has world position
    '''
    
    tempGrp = mc.createNode( 'transform', n = 'TEMP_position_query_grp_' + object )
    mc.pointConstraint( object, tempGrp )
    objectWorldPos = mc.xform( tempGrp, q = 1, t = 1, ws = 1 )
    mc.delete( tempGrp )
    
    closestParams, closestPos = getClosestPointOnSurfaceFromPosition( objectWorldPos, surfaceObj, normalizeCoords )
    
    return [ closestParams, closestPos ]
    

def getClosestPointOnSurfaceFromPosition( worldPosition, surfaceObj, normalizeCoords = False ):
    
    '''
    get closest point on surfaceObj, return both parameter and world position
    
    @param object: name of transform object
    @type object: str
    @param surfaceObj: name of nurbsSurface object
    @type surfaceObj: str
    @return: list( list( float, float ), list( float, float, float ) ) - first list has UV parameters and second has world position
    '''
    
    prefix = 'getClosestPointOnSurf'
    
    closestNode = mc.createNode( 'closestPointOnSurface', n = prefix + '_cps' )
    surfaceShape = shape.getShapes( surfaceObj )[0]
    
    mc.connectAttr( surfaceShape + '.worldSpace[0]', closestNode + '.inputSurface' )
    
    mc.setAttr( closestNode + '.inPosition', worldPosition[0], worldPosition[1], worldPosition[2] )
    
    closestPos = mc.getAttr( closestNode + '.position' )[0]
    closestParamU = mc.getAttr( closestNode + '.parameterU' )
    closestParamV = mc.getAttr( closestNode + '.parameterV' )
    closestParams = [ closestParamU, closestParamV ]
    
    mc.delete( closestNode )
    
    if normalizeCoords:
        
        closestParams = getNormalizedCoordinates( surfaceObj, closestParams )
    
    return [ closestParams, closestPos ]


def getNormalizedCoordinates( surfaceObj, coordinates ):
    
    """
    get normalized coordinates based on given surfaceObj
    surfaceObj coordinates are often out of range 0-1
    
    @param surfaceObj: name of nurbsSurface object
    @type surfaceObj: str
    @param coordinates: coordinates from the given surfaceObj to be normalized
    @type coordinates: list( float, float )
    @return: list( float, float )
    """
    
    # get ranges from surfaceObj
    
    uRanges = mc.getAttr( surfaceObj + '.minMaxRangeU' )[0]
    vRanges = mc.getAttr( surfaceObj + '.minMaxRangeV' )[0]
    
    # compute normalized coordinates
    
    normalizedCoords = []
    
    for coord, ranges in zip( coordinates, [ uRanges, vRanges ] ):
        
        clampedVal = coord - ranges[0]
        clampedUpperRange = ranges[1] - ranges[0]
        rangeMulti = 1.0 / clampedUpperRange
        normalizedCoord = clampedVal * rangeMulti
        normalizedCoords.append( normalizedCoord )
    
    return normalizedCoords




