'''
transform module at rigit.utils
@category Rigging @subcategory Utils
@tags distance measure transform
@author jakub
'''

import maya.cmds as mc

from . import vector
from . import name


def measureDistanceBetween2Points( pointA, pointB ):
    
    '''
    @param objectA: first point
    @type objectA: list(3 floats)
    @param objectB: second point
    @type objectB: list(3 floats)
    @return: float, distance between given points in same space
    '''
    
    vectA = vector.makeMVector( [ pointA[0], pointA[1], pointA[2] ] )
    vectB = vector.makeMVector( [ pointB[0], pointB[1], pointB[2] ] )
    pointDif = vectB - vectA
    
    return pointDif.length()
    

def measureDistanceBetweenMultiplePositions( positionList ):
    
    '''
    return sum of distances between chain of positions
    This depends on order of objects as they were passed to function,
    ideal for measuring length of geometry CVs perimeter
    
    @param objectList: list of transform names
    @type objectList: list(str)
    @return: float, distance sum
    '''
    
    distancesum = 0.0
    
    for i in range( 1, len( positionList ) ):
        
        distancesum += measureDistanceBetween2Points( positionList[i - 1], positionList[i] )
    
    return distancesum

def getParentList( obj, ascending = True, verbose = False, longNames = False ):
    
    '''
    get list of object parents
    
    @param obj: str, name of object to get its parent list from
    @param ascending: bool, if true parent list is ascending, if false list is descending
    @param fullNames: bool, return parent full names, will return short names by default
    @param verbose: bool, prints function results
    @param longNames: bool, return long names
    @return: list(str), list of parents in order starting from closest parent to top parent
    '''
    
    parents = []
    parentList = mc.listRelatives( obj, f = True, p = True )
    
    
    # check if parent is the world
    if not parentList:
        
        if verbose:
            
            print('# %s parent is world' % obj)
            
        return []
    
    
    # split str and delete first empty item    
    parentsRes = parentList[0].split( '|' )[1:]
    
    # get long names
    if longNames:
        
        parentsLongName = []
        for i, p in enumerate( parentsRes ):
            
            parentItems = parentsRes[:i + 1]
            pLongName = '|'.join( parentItems )
            pLongName = '|' + pLongName
            parentsLongName.append( pLongName )
        
        parentsRes = parentsLongName
    
    parentsRes.reverse()    
    
    return parentsRes

def getTopNodes( objects, returnShortNames = True ):
    
    """
    similar to sortFromLowest() function - returns top nodes which have no parent
    
    @param: objects: list( str ), list of objects to get sorted by hierarchy from highest to lowest
    @param: bool, returnShortNames: return short names
    @return: list( str ), list of object names 
    """
    
    longNames = [ mc.ls( o, l = 1 )[0] for o in objects ]
    objects = longNames
    
    topNodes = []
    for o in objects:
        
        if o.count( '|' ) == 1:
            
            topNodes.append( o )
    
    if returnShortNames:
        
        shortNames = [ mc.ls( o )[0] for o in topNodes ]
        topNodes = shortNames
        
    
    return topNodes

def getTopNode( obj, returnShortNames = True ):

    """
    returns highest parent of object
    """

    objParents = getParentList( obj, ascending = True, longNames = True )
    objParents.append(obj)
    topObj = getTopNodes( objParents, returnShortNames = returnShortNames )[0]
    
    return topObj

def getLongNameInHierarchy( obj, topObject ):

    """
    get long name of object short name from given hierarchy
    NOTE : if object not found it will return None
    NOTE 2 : if multiple objects match the name they will be all returned in list
    NOTE 3 : function works on shapes too

    @param shortNameObj: str, short name of object to get long name from hierarchy
    @param topObject: str, name of top hierarchy / parent object
    @return: list(str), return list of short name matching object long names, if nothing found return None
    """

    # make sure we have short name
    # - due to reusing same name in variable where it gets long at some point if function is used multiple times
    shortNameObj = name.short(obj)

    matches = []
    allObjsLongNames = mc.listRelatives( topObject, ad=True, fullPath=True )

    for ln in allObjsLongNames:

        objShort = name.short( ln )
        if shortNameObj == objShort:
            matches.append( ln )
    
    return matches
