'''
vector in rigit.utils
@category Rigging @subcategory Utils
@tags vector math
@author jakub
'''

import maya.OpenMaya as om
import maya.cmds as mc


def makeMVector( values = [0.0, 0.0, 0.0] ):
    
    '''
    @summary: wrapper to make MVector instance from list of floats
    @param values: 3 numbers to make vector
    @type values: list of floats     
    
    @return: OpenMaya.MVector
    '''
    
    return om.MVector( values[0], values[1], values[2] )
    

def from2Objects( objA, objB, normalized = True ):
    
    '''
    @param objA: str, first object
    @param objB: str, second object
    @param normalized: bool, optional, normalize resulting vector
    @return: list of 3 floats
    
    make vector using world positions of 2 given objects
    NOTE: order of passed objects matters - object B is the one "more far" or "later in time"
    '''
    
    avect = makeMVector( _getPositionFromObj( objA ) )
    bvect = makeMVector( _getPositionFromObj( objB ) )
    
    resvect = bvect - avect
    
    if normalized:
        
        resvect.normalize()
    
    
    return resvect

def _getPositionFromObj( obj ):
    
    """
    get world position from reference object
    
    @param obj: str, name of reference object
    @return: str, name of group
    """
    
    grp = mc.group( n = obj + '_ref_grp', em = True )
    mc.delete( mc.parentConstraint( obj, grp ) )
    pos = mc.xform( grp, q = 1, t = 1, ws = 1 )
    mc.delete( grp )
    
    return pos

    
    

def getAxisNameFromVector( values = [0, 0, 0] ):
    
    '''
    get axis name as lower case letter based on 3 floating values, NOTE: values are taken as local
    
    example: [0,1,0] would result as 'y'
    
    @param values: 3 floating values
    @type values: list( float, float, float )
    @return: str, name of axis, valid values: 'x', 'y', 'z'
    '''
    
    maxIndex = 0
    max = values[maxIndex]
    
    for i in range( 3 ):
        
        if values[i] > max:
            
            maxIndex = i
    
    axisNames = ['x', 'y', 'z']
    
    return axisNames[maxIndex]


def setupDotProduct( transformA, transformB, prefix = 'dotProductSetup', comparedAxis = 'x' ):
    
    '''
    create vector dot product output node from 2 transforms (groups, joints etc.)
    
    @param transformA: first transform
    @type transformA: str
    @param transformB: second transform
    @type transformB: str
    @param prefix: prefix to name new objects
    @type prefix: str
    @param comparedAxis: name of axis to compare from both transforms in dot product
    @return: str, dot product value output plug ("NODE.ATTRIBUTE")
    '''
    
    # make passed axis is lower case
    comparedAxis = comparedAxis.lower()
    
    poseVector = {'x':[1, 0, 0], 'y':[0, 1, 0], 'z':[0, 0, 1]}[comparedAxis]
    
    # - Pose Transform Vector
    vectorNodeA = mc.createNode( 'vectorProduct', n = prefix + 'VectorA_vtp' )
    mc.setAttr( vectorNodeA + '.operation', 3 )  # Vector Matric Product
    mc.setAttr( vectorNodeA + '.input1', poseVector[0], poseVector[1], poseVector[2] )
    mc.setAttr( vectorNodeA + '.normalizeOutput', 1 )
    mc.connectAttr( transformA + '.worldMatrix[0]', vectorNodeA + '.matrix', f = True )
    
    # - Pose Reference Vector
    vectorNodeB = mc.createNode( 'vectorProduct', n = prefix + 'VectorB_vtp' )
    mc.setAttr( vectorNodeB + '.operation', 3 )  # Vector Matric Product
    mc.setAttr( vectorNodeB + '.input1', poseVector[0], poseVector[1], poseVector[2] )
    mc.setAttr( vectorNodeB + '.normalizeOutput', 1 )
    mc.connectAttr( transformB + '.worldMatrix[0]', vectorNodeB + '.matrix', f = True )
    
    # - Pose Vector Comparison
    dotProductNode = mc.createNode( 'vectorProduct', n = prefix + 'DotProduct_vtp' )
    mc.setAttr( dotProductNode + '.operation', 1 )  # Dot Product
    mc.setAttr( dotProductNode + '.normalizeOutput', 1 )
    mc.connectAttr( vectorNodeA + '.output', dotProductNode + '.input1' )
    mc.connectAttr( vectorNodeB + '.output', dotProductNode + '.input2' )
    
    # return output plug
    
    outputPlug = dotProductNode + '.outputX'
    
    return outputPlug
    
